#include "defs.h"
#include <unistd.h>

char autorp_recv_buf[MAX_IP_PACKET_LEN];	/* output packet buffer        */

struct g_rplist *auto_rp_set;
struct rp_afilter *rp_discovery_filter;

struct sockaddr_in autorp_src;
int autorp_sock = -1;
int autorp_ifindex;
__u32 autorp_dst;

int autorp_announce;

static void autorp_read(int f, fd_set *rfd);

struct autorphdr
{
	__u8	verstype;
	__u8	rpcount;
	__u16	holdtime;
	__u32	resv1;
};

struct autorp
{
	__u32	router;
	__u8	pimvers;
	__u8	grpcount;
};

static void accept_autorp(int recvlen);

void init_autorp(void)
{
	struct sockaddr_in sin;
	int on = 1;
	unsigned char ucval;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(496);

	autorp_sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (autorp_sock < 0) {
		log(LOG_ERR, errno, "autorp_socket");
		return;
	}
	if (bind(autorp_sock, (struct sockaddr*)&sin, sizeof(sin)) < 0) {
		log(LOG_ERR, errno, "bind autorp_socket");
		close(autorp_sock);
		autorp_sock = -1;
		return;
	}
	if (setsockopt(autorp_sock, IPPROTO_IP, IP_PKTINFO,
		       (char *) &on, sizeof(on)) < 0)
		log(LOG_ERR, errno, "setsockopt IP_PKTINFO %u", on);

	ucval = 0;
	if (setsockopt(autorp_sock, IPPROTO_IP, IP_MULTICAST_LOOP,
		       (char *) &ucval, sizeof(ucval)) < 0)
		log(LOG_ERR, errno, "setsockopt IP_MULTICAST_LOOP %u", ucval);

	ucval = 64;
	if (setsockopt(autorp_sock, IPPROTO_IP, IP_MULTICAST_TTL,
		       (char *) &ucval, sizeof(ucval)) < 0)
		log(LOG_ERR, errno, "setsockopt IP_MULTICAST_TTL %u", ucval);

	if (register_input_handler(autorp_sock, autorp_read) < 0)
		log(LOG_WARNING, 0, "Couldn't register autorp_read as an input handler");
}

void autorp_vif(int ifindex, int on)
{
	struct ip_mreqn imr;	/* multicast request struct to join the group */

	imr.imr_multiaddr.s_addr = __constant_htonl(0xE0000128UL);
	imr.imr_address.s_addr = 0;
	imr.imr_ifindex = ifindex;

	if (setsockopt(autorp_sock, SOL_IP, on ? IP_ADD_MEMBERSHIP : IP_DROP_MEMBERSHIP,
		       &imr, sizeof(imr)) < 0) {
		log(LOG_ERR, errno, "can't join autorp on if%d", ifindex);
	} else {
		log(LOG_INFO, 0, "listen AUTORP on if%d", ifindex);
	}
}

void add_to_dr_rpset(__u32 rp, __u32 grp, int masklen, int prio, u_long tmo, int vers)
{
	struct g_rplist *p, **pp, *oldp;

	if (masklen < 4 || masklen > 32)
		return;

	if (autorp_src.sin_addr.s_addr) {
		struct rp_afilter *a;

		for (a=rp_discovery_filter; a; a = a->next) {
			if (!((a->rpaddr^rp)&a->rpmask) &&
			    !((a->oaddr^autorp_src.sin_addr.s_addr)&a->omask) &&
			    masklen >= a->grplen &&
			    !((a->group^grp)&a->grpmask)) {
				if (a->flag)
					return;
				else
					break;
			}
		}
	}

	for (pp = &auto_rp_set; (p=*pp) != NULL; pp = &p->next) {
		if (p->group == grp && p->grpmasklen == masklen) {
			if (p->override)
				return;
			if (p->rpaddr == rp) {
				if (p->expires != 0xFFFFFFFF) {
					p->expires = tmo;
					if (p->info_source != autorp_src.sin_addr.s_addr)
						p->info_source = autorp_src.sin_addr.s_addr;
					if (p->pimversion != vers) {
						p->pimversion = vers;
						dirty_bits |= 4;
					}
				}
				return;
			}
			if (htonl(p->rpaddr) < htonl(rp))
				break;
		}
		if (p->grpmasklen < masklen)
			break;
	}

	oldp = p;
	p = (struct g_rplist *) alloc_mem(sizeof(struct g_rplist));
	p->rpaddr = rp;
	p->group = grp;
	p->grpmasklen = masklen;
	p->grpmask = htonl((0xFFFFFFFF<<(32-masklen)) & 0xFFFFFFFF);
	p->timer = 0;
	p->announce = 0;
	p->override = 0;
	p->scope = 0;
	p->dense = 0;
	p->pimversion = vers;
	p->info_source = autorp_src.sin_addr.s_addr;
	p->expires = tmo;
	p->next = oldp;
	*pp = p;
	dirty_bits |= 4;
}

static void accept_autorp(int recvlen)
{
	struct autorphdr *ah;
	struct autorp *rp;
	u_long tmo;
	int i;

	if (recvlen < sizeof(struct autorphdr)) {
		log(LOG_WARNING, 0, "packet too short (%u bytes) for AUTORP header",
		    recvlen);
		return;
	}

	ah = (struct autorphdr *) (autorp_recv_buf);

	tracef(TRF_AUTORP, "RX AUTORP: from %s v/t=%02x count=%d holdtime=%u resv=%08x",
	       inet_fmt(autorp_src.sin_addr.s_addr, s1),
	       ah->verstype, ah->rpcount, ntohs(ah->holdtime),
	       ah->resv1);

	if ((ah->verstype & 0xF0) != 0x10) {
		tracef(TRF_AUTORP|TRF_UNUSUAL, "autorp: wrong version %02x", ah->verstype);
		return;
	}
	if (autorp_dst != __constant_htonl(0xE0000128UL)) {
		tracef(TRF_AUTORP|TRF_UNUSUAL, "autorp: wrong group %s", inet_fmt(autorp_dst, s1));
		return;
	}
	if ((ah->verstype & 0x0F) != 0x2) {
		tracef(TRF_AUTORP|TRF_UNUSUAL, "autorp: wrong type %02x", ah->verstype);
		return;
	}
	tmo = ntohs(ah->holdtime);
	if (tmo == 0)
		tmo = 0xFFFFFFFE;
	else
		tmo += wall_time;
	rp = (struct autorp*)&ah[1];
	for (i = ah->rpcount-1; i>=0; i--) {
		int vers = rp->pimvers&3;
		__u32 rtr = rp->router;
		__u8 *cp;
		__u32 grp;
		int k;
		tracef(TRF_AUTORP, "RX AUTORP: RP %s v%d", inet_fmt(rtr, s1), vers);
		cp = ((__u8*)rp) + 6;
		for (k = rp->grpcount-1; k >=0; k--) {
			memcpy(&grp, cp+2, 4);
			tracef(TRF_AUTORP, "RX AUTORP: GROUP %s/%u", inet_fmt(grp, s1), cp[1]);
			add_to_dr_rpset(rtr, grp, cp[1], 1, tmo, vers);
			cp += 6;
		}
		rp = (struct autorp*)(((unsigned long)cp + 3) & ~3);
	}
	tracef_flush(TRF_AUTORP|TRF_UNUSUAL);
}


/* Read an IGMP message */
void autorp_read(f, rfd)
int f;
fd_set *rfd;
{
	int recvlen;
	struct iovec iov = { autorp_recv_buf, sizeof(autorp_recv_buf) };
	char cbuf[128];
	struct cmsghdr *cmsg;
	struct msghdr msg = { 
		&autorp_src, sizeof(autorp_src),
		&iov,	1,
		cbuf,	sizeof(cbuf),
		0
	};

	recvlen = recvmsg(autorp_sock, &msg, MSG_DONTWAIT);

	if (recvlen < 0) {
		if (errno != EINTR && errno != EAGAIN)
			log(LOG_WARNING, errno, "AURORP recvmsg");
		return;
	}
	if (msg.msg_flags&MSG_TRUNC) {
		log(LOG_WARNING, 0, "AUTORP recvmsg: truncated message");
		return;
	}
	if (msg.msg_flags&MSG_CTRUNC) {
		log(LOG_WARNING, 0, "AUTORP recvmsg: truncated control message");
		return;
	}
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_IP ||
		    cmsg->cmsg_type != IP_PKTINFO ||
		    cmsg->cmsg_len < CMSG_LEN(sizeof(struct in_pktinfo))) {
			log(LOG_WARNING, 0, "AUTORP recvmsg: too short in_pktinfo");
			return;
		}
		autorp_ifindex = ((struct in_pktinfo*)CMSG_DATA(cmsg))->ipi_ifindex;
		autorp_dst = ((struct in_pktinfo*)CMSG_DATA(cmsg))->ipi_addr.s_addr;

		accept_autorp(recvlen);
		return;
	}
	log(LOG_WARNING, 0, "AUTORP recvmsg: no in_pktinfo");
}

void autorp_send_announce(void)
{
	char buf[4096];
	struct autorphdr *ah;
	struct autorp *ap;
	struct sockaddr_in sin;
	struct g_rplist *r;
	static u_long ticks;
	int ttl;
	int next_ttl;

	if (ticks > wall_time)
		return;

	ticks = wall_time + 60;

	next_ttl = 0;

next_scope:
	ttl = next_ttl;
	next_ttl = 256;
	ah = (void*)buf;
	ap = (void*)&ah[1];

        ah->verstype = 0x11;
        ah->rpcount = 0;
        ah->holdtime = htons(180);
        ah->resv1 = 0; 

	for (r = auto_rp_set; r; r = r->next) {
		__u8 *cp;
		if (!r->announce || r->dense)
			continue;
		if (ttl != r->scope) {
			if (ttl < r->scope && r->scope<next_ttl)
				next_ttl = r->scope;
			continue;
		}
		ah->rpcount++;
		ap->router = r->rpaddr;
		ap->pimvers = r->pimversion;
		ap->grpcount = 1;
		cp = ((__u8*)ap) + 6;
		cp[0] = 0;
		cp[1] = r->grpmasklen;
		memcpy(cp+2, &r->group, 4);
		tracef(TRF_AUTORP, "TX AUTORP: RP %s v%d",
		       inet_fmt(ap->router, s1), ap->pimvers);
		tracef(TRF_AUTORP, "TX AUTORP: GROUP %s/%u",
		       inet_fmt(r->group, s1), cp[1]);
		ap = (void*)(cp+6);
	}
	if (ah->rpcount) {
		tracef(TRF_AUTORP, "TX AUTORP: scope=%d v/t=%02x count=%d holdtime=%u resv=%08x",
		       ttl, ah->verstype, ah->rpcount, ntohs(ah->holdtime),
		       ah->resv1);

		if (ttl) {
			memset(&sin, 0, sizeof(sin));
			sin.sin_family = AF_INET;
			sin.sin_port = __constant_htons(496);
			sin.sin_addr.s_addr = __constant_htonl(0xE0000127);
			if (setsockopt(autorp_sock, IPPROTO_IP, IP_MULTICAST_TTL,
				       (char *) &ttl, sizeof(ttl)) < 0) {
				log(LOG_ERR, errno, "setsockopt IP_MULTICAST_TTL %d", ttl);
			}
			if (sendto(autorp_sock, buf, (__u8*)ap - (__u8*)buf, 0, (struct sockaddr*)&sin,
				   sizeof(sin)) < 0) {
				log(LOG_WARNING, errno, "autorp_announce sendto");
			}
		}
	}
	if (next_ttl < 256)
		goto next_scope;
}

void dump_auto_rp_set(FILE *fp)
{
	struct g_rplist *r;
	fprintf(fp, "\nAuto-RP Set\n");
	fprintf(fp, "(R,G)                             Vers  InfoSource      Expires\n\n");

	for (r = auto_rp_set; r; r = r->next) {
		char buf[64];

		if (!r->dense)
			sprintf(buf, "(%s,%s/%d)",
				inet_fmt(r->rpaddr, s1),
				inet_fmt(r->group, s2), r->grpmasklen);
		else
			sprintf(buf, "(dense,%s/%d)",
				inet_fmt(r->group, s2), r->grpmasklen);

		fprintf(fp, "%-33s", buf);

		switch (r->pimversion) {
		case 0:	
			fprintf(fp, " None ");
			break;
		case 1:	
			fprintf(fp, " PIMv1");
			break;
		case 2:	
			fprintf(fp, " PIMv2");
			break;
		case 3:	
			fprintf(fp, " Dual ");
			break;
		default:	
			fprintf(fp, " %-4u ", r->pimversion);
			break;
		}
		if (r->info_source)
			fprintf(fp, " %-15s", inet_fmt(r->info_source, s1));
		else
			fprintf(fp, " %-15s", "localhost");
		
		if (r->expires == 0xFFFFFFFF)
			fprintf(fp, " %-9s", "static");
		else if (r->expires == 0xFFFFFFFE)
			fprintf(fp, " %-9s", "never");
		else if (r->expires < wall_time)
			fprintf(fp, " %-9s", "expired");
		else
			fprintf(fp, " %-9s", inet_time(r->expires, s2));

		if (r->announce)
			fprintf(fp, " announce %d", r->scope);
		if (r->override)
			fprintf(fp, " override");
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n");
}

