/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */

#include "defs.h"

u_short c_rp_holdtime = 3 * C_RP_Adv_INTERVAL;
/* the hold time for the candidate rp advertisements */
u_int C_RP_Adv_timer = 0;
u_char HashMaskLen = HASHMASKLEN;	/* start with default. Maybe overriden by conf */

u_int C_RP_Timeout = RPHoldtime;

u_char AxptAnyRPSet = 0;	/* a state flag for the RP-Set mechanism */
u_char ElectedBSR = 0;		/* a state flag for the BSR election mech */
u_int RPSetTimer = BootstrapTimeout;	/* RP Set timer initialized to max thresh */
u_char LclBSRPrio = 0;		/* local BSR priority init'ed to 0 */

static struct g_rplist *c_rp_set = NULL;	/* initializing Candidate RP Set, kept at 
					 * the Bootstrap-RP 
					 */

static struct g_rplist *DR_c_rp_set = NULL;	/* kept at DRs */

/* to be used with the hash function in mrt.c */
/* initialize the hash mask to mask the least 2 bits */

__u32 hashMask = 0xFFFFFFFC;

static int search_c_rp_set(struct g_rplist *setPtr, __u32 address, struct g_rplist **Ptr);
static void destroyList(struct g_rplist **list);

/* function name: initBSR
   input: none
   output: int (TRUE if successful, FALSE o.w.)
   operation: if this router is a candidate BSR
   assign the configured address, o.w. empty values
   comments: called by init_vifs
 */
int initBSR(void)
{
	/* can use just an rp address, but just in case we need it later */

	if (!CBSRFlag) {	/* I am not a Candidate bootstrap RP */
		BSR.rpaddr = INADDR_ANY;
		return (FALSE);
	}
	BSR.rpaddr = confBSR->rpaddr;
	BSR.timer = 0;
	LclBSRPrio = confBSRPrio;

	if (debug > 3)
		log(LOG_DEBUG, 0, "Boostrap Router is %s",
		    inet_fmt(BSR.rpaddr, s1));

	return (TRUE);
}



static int dumpRPSet(struct g_rplist * rpset)
{
	struct g_rplist *r, *prev_r;

	log(LOG_DEBUG, 0, "*** Dumping RP-Set ***");

	for (prev_r = NULL, r = rpset; r;) {
		log(LOG_DEBUG, 0, "\nRP: %s, timer:%u",
		    inet_fmt(r->rpaddr, s1), r->timer);

		if (local_address(r->rpaddr) != MAXVIFS)
			log(LOG_DEBUG, 0, ",  local address");
		prev_r = r;
		r = r->next;
	}
	return (TRUE);
}


/* function name: send_C_RP_Adv
   input: none
   output: int (TRUE if successful, FALSE o.w.)
   operation
   comments: called by periodic_C_RP_Adv
 */
static int send_C_RP_Adv(__u32 bsradd)
{
	char *pkt;
	int datalen = 0;
	u_long candidacy_group = allrouters_group;
	u_char candidacy_mask = 4;	/* currently mask all */
	/* i.e. currently use a wild card candidacy */
	u_long src;
	u_long dst;

	src = CRP->rpaddr;

	pkt = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
	pkt[0] = 1;		/* prefixCnt; *//* the prefix count set to 1 for now */
	pkt[1] = 0;
	pkt += 2;
	datalen += 2;
	bcopy((char *) &(c_rp_holdtime), pkt, 2);
	pkt += 2;
	datalen += 2;
	bcopy((char *) &(CRP->rpaddr), pkt, ADDR_LEN);
	pkt += ADDR_LEN;
	datalen += ADDR_LEN;
	datalen += pim2_encode_group(&pkt, candidacy_group, candidacy_mask);

	dst = bsradd;

	if (debug > 2)
		log(LOG_DEBUG, 0, "\n**Sending C-RP-ADVERTISEMENT from %s to %s\n",
		    inet_fmt(src, s1), inet_fmt(dst, s2));

	send_pim(pim_send_buf, src, dst,
		 PIM_CANDIDATE_RP_ADVERTISEMENT, datalen, 0, MAXTTL);
	return (TRUE);
}

/* function name: periodic_C_RP_Adv
   input: none
   output: int (TRUE if a message is sent, o.w. FALSE)
   operation: increments and checks the Candidate RP Advertisement
   (C-RP-Adv) timer.
   comments: called by age_routes
 */
int periodic_C_RP_Adv(void)
{
	u_long bsraddr;

	C_RP_Adv_timer += TIMER_INTERVAL;
	if (C_RP_Adv_timer < c_rp_adv_interval)
		return (FALSE);

	bsraddr = BSR.rpaddr;

	/* if I have no bsr or I am the BSR, return */
	if (!bsraddr || (local_address(bsraddr) != MAXVIFS))
		return (FALSE);

	if (send_C_RP_Adv(bsraddr))
		/* we may fail if the mapping for the C-RP-Adv group is not
		   available, in which case we don't want to reset the timer,
		   so that we try again next time */
	{
		C_RP_Adv_timer = 0;
		return (TRUE);
	}
	return (FALSE);
}

/* function name: construct_c_rp_set
   input: pointer to the pkt (char *)
   output: int returns the datalength.
   operation: from the c_rp_set, this function constructs the C-RP-Set
   comments: called by periodic_c_rp_set
 */
int construct_c_rp_set(char *pkt)
{
	static u_short FragmentTag = 0;		/* fragment tag number */

	struct g_rplist *r;
	u_char count;
	char *pktPtr;
	int datalen = 0;

	if (debug > 2)
		log(LOG_DEBUG, 0, "\nConstructing RPSET");

	pktPtr = pkt;

	FragmentTag++;		/* for now, every RPSet, gets a new tag */
	if (debug > 2)
		log(LOG_DEBUG, 0, "\nFragTag: %u", FragmentTag);
	bcopy((char *) &FragmentTag, pktPtr, sizeof(u_short));
	pktPtr += sizeof(u_short);
	datalen += sizeof(u_short);
	pktPtr[0] = HashMaskLen;
	pktPtr[1] = LclBSRPrio;
	pktPtr += sizeof(u_short);
	datalen += sizeof(u_short);
	bcopy((char *) &BSR.rpaddr, pktPtr, ADDR_LEN);
	pktPtr += ADDR_LEN;
	datalen += ADDR_LEN;

	/* the encoded groups are only 1 for now to cover all grps.. 
	 * i.e. 224.0.0.0/4
	 */
	datalen += pim2_encode_group(&pktPtr, allpimrouters_group, 4);
	/* skip RPCnt and FragRPCnt */
	pktPtr += sizeof(u_short);
	datalen += sizeof(u_short);

	bzero(pktPtr, sizeof(u_short));
	pktPtr += sizeof(u_short);
	datalen += sizeof(u_short);
	for (count = 0, r = DR_c_rp_set; r; r = r->next, count++) {
		if (r->grpmasklen != 4 ||
		    r->group != __constant_htonl(0xE000000))
			continue;
		bcopy((char *) &r->rpaddr, pktPtr, ADDR_LEN);
		pktPtr += ADDR_LEN;
		datalen += ADDR_LEN;
	}
	pkt[14] = count;
	pkt[15] = count;

	if (debug > 2)
		log(LOG_DEBUG, 0, "RPCnt in C-RP-Set is %u", count);

	return (datalen);	/* count ? datalen : FALSE); */
}

/* function name: send_c_rp_set
   input: packet pointer (char *) and datalength
   output: int
   operation: sends the c_rp_set packet to the well known group
   comments: called by periodic_c_rp_set
 */
int send_c_rp_set(char *packet, int datalen, __u32 src, __u32 dst)
{
	char *pktTest;
	u_char count;
	u_long address;

	if (debug > 3) {	/* show the contents of the RP-Set */
		pktTest = packet + MIN_IP_HEADER_LEN + PIM_MINLEN;
		count = pktTest[15];
		pktTest += 18;
		log(LOG_DEBUG, 0, "in Send C-RP-Set Count is %u", count);
		for (; count; count--) {
			bcopy(pktTest, (char *) &address, 4);
			log(LOG_DEBUG, 0, "RP %u is %s", count,
			    inet_fmt(address, s1));
			pktTest += 4;
		}
	}
	if (dst == allpimrouters_group)
		/* sending periodic RP-Set to all pim routers grp */
		send_pim(packet, src, dst, PIM_BOOTSTRAP, datalen, 0, 1);
	else
		/* triggering a unicast RP-Set to a new pim nbr */
		send_pim(packet, src, dst, PIM_BOOTSTRAP, datalen, 0, MAXTTL);
	return (TRUE);
}

/* function name: frwdRPSet
   input: pkt and iif
   output: int (TRUE if successful, FALSE o.w.)
   operation: checks the bootstrap RP address then calls appropriate 
   parsing and update functions to process the C-RP-Set message
   comments: called by periodic_c_rp_set and accept_c_rp_set
   the BSR should call this function with non valid iif,
   since it originates the message.
 */
static int frwdRPSet(char *pkt, int datalen, vifi_t iif)
{
	vifi_t vifi;
	struct vifentry *v;

	if (debug > 3)
		log(LOG_DEBUG, 0, "\nForwarding RPSet,iif is %u", iif);

	/* browse thru the interfaces except iif and send RPSet */
	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		if (!v->neighbors || v->v1neighnum == v->neighnum)
			continue;
		/* don't send message on the iif */
		if (vifi == iif)
			continue;
		send_c_rp_set(pkt, datalen, v->prefsrc, allpimrouters_group);
	}
	return (TRUE);
}


/* function name: periodic_c_rp_set
   input: none
   output: int (TRUE if successful)
   operation: constructs a C-RP-Set message from the c_rp_set rplist
   and mcasts it to the well-known C-RP-Set_Group
   comments: called by originatePeriodicRPSet
   calls frwdRPSet
 */
static int periodic_c_rp_set(void)
{
	char *pkt;
	int datalen;

	if (debug > 2)
		log(LOG_DEBUG, 0, "\nPeriodic C-RP-Set");

	pkt = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
	if ((datalen = construct_c_rp_set(pkt)) != 0)
		frwdRPSet(pim_send_buf, datalen, MAXVIFS);
	return (TRUE);
}

/* function name: timeout_c_rps
   input: none
   output: int (TRUE if successful)
   operation: times out the c_rp_set entries for the C-RPs
   comments: called by originatePeriodicRPSet
 */
static int timeout_c_rps(void)
{
	struct g_rplist *r, *prev_r, *temp;
	int change = 0;

	if (debug > 2)
		log(LOG_DEBUG, 0, "I am b-RP timing out C-RP-Set");

	for (prev_r = NULL, r = c_rp_set; r;) {
		/* don't age a local address */
		if (local_address(r->rpaddr) != MAXVIFS) {
			if (debug > 2)
				log(LOG_DEBUG, 0, "\n%s is local address",
				    inet_fmt(r->rpaddr, s1));
			prev_r = r;
			r = r->next;
			continue;
		}
		if ((r->timer += TIMER_INTERVAL) > C_RP_Timeout) {
			change = 1;
			temp = r;
			if (prev_r)
				prev_r->next = r->next;
			else	/* we timed out the first entry */
				c_rp_set = r->next;

			r = r->next;
			if (debug > 2)
				log(LOG_DEBUG, 0, "C-RP %s timed out",
				    inet_fmt(temp->rpaddr, s1));
			free((char *) temp);
		} else {
			if (debug > 2)
				log(LOG_DEBUG, 0, "\nC-RP %s, not timed out ",
				    inet_fmt(r->rpaddr, s1));
			prev_r = r;
			r = r->next;
		}
	}
	if (debug > 2)
		log(LOG_DEBUG, 0, "\nleaving timeout\n");
	return (change);
}

/* function name: duplicateList
   input: rplist pointer
   output: rplist pointer for the newly created rplist
   operation: copies the given list into a newly created list
   and returns a pointer to it.
   comments: called by originatePeriodicRPSet
 */
struct g_rplist * duplicateList(struct g_rplist *listToCopy)
{
	struct g_rplist *list, *prev, *listHead;

	list = NULL;		/* initialize all to NULL */
	prev = NULL;
	listHead = NULL;

	if (debug > 3)
		log(LOG_DEBUG, 0, "duplicateList");

	for (; listToCopy; listToCopy = listToCopy->next) {
		list = (struct g_rplist *) alloc_mem(sizeof(struct rplist));
		list->rpaddr = listToCopy->rpaddr;
		/* all other fields are by default set to 0's */

		if (!prev)
			listHead = list;	/* first element in the new list */
		else
			prev->next = list;
		prev = list;
		list = list->next;
	}
	return (listHead);
}

/* function name: addToList
   input: address and rplist pointer
   output: int
   operation: adds the given address into the given list
   comments: called by originatePeriodicRPSet
 */
static int addToList(__u32 add, struct g_rplist **list)
{
	struct g_rplist *nuelem;

	nuelem = (struct g_rplist *) alloc_mem(sizeof(struct rplist));
	nuelem->rpaddr = add;
	nuelem->next = (*list);	/* put nuelem at list head for now */
	(*list) = nuelem;
	/* may sort accdg to mask groups.. so on */

	return (TRUE);
}

/* function name: periodic_c_rp_set_handling
   input: none
   output: int (TRUE if successful, False o.w.)
   operation: times out the candidate RPs on the C-RP-Set, then sends
   out periodic C-RP-Set messages.
   comments: called by ageRPSetTimer
 */
static int originatePeriodicRPSet(void)
{
	struct g_rplist *dummy;

	/* check if I am Elected as BSR for the first time */
	if (!ElectedBSR) {
		if (!confBSR)
			return (FALSE);		/* error */

		if (debug > 2)
			log(LOG_DEBUG, 0, "\nFirst time elected, prev bsr is %s, prio %x",
			    inet_fmt(BSR.rpaddr, s1), LclBSRPrio);

		BSR.rpaddr = confBSR->rpaddr;
		LclBSRPrio = confBSRPrio;
		/* now I am the elected BSR */

		if (debug > 3)
			log(LOG_DEBUG, 0, "\nset to new bsr %s, prio %x",
			    inet_fmt(BSR.rpaddr, s1), LclBSRPrio);

		/* this is first time as elected BSR, so
		 * send the latest RP-Set I have [i.e. DRRPSet]
		 */
		destroyList(&c_rp_set);
		if (CRPFlag)	/* if I am a C-RP, add me to the list */
			if (!search_c_rp_set(DR_c_rp_set, CRP->rpaddr, &dummy))
				addToList(CRP->rpaddr, &DR_c_rp_set);

		c_rp_set = duplicateList(DR_c_rp_set);
		if (debug > 3) {
			log(LOG_DEBUG, 0, "\nDumping C-RP-Set");
			dumpRPSet(c_rp_set);
		}
	}
	/* this is not my first time as elected BSR, so use the 
	   constructed RP-Set */
	else {
		/* currently; for simplicity, destroy old set and create new and
		 * then link;
		 */
		timeout_c_rps();
		destroyList(&DR_c_rp_set);

		DR_c_rp_set = duplicateList(c_rp_set);
	}

	if (debug > 2)
		log(LOG_DEBUG, 0, "\nsending periodic RPSet");
	periodic_c_rp_set();
	return TRUE;
}

/* function name: ageRPSetTimer
   input: none
   output: int (TRUE if RP-Set timer expired, False o.w.)
   operation: decreases the RP Set timer, and implements the BSR election/
   RP-Set distribution mechanism as in the spec., accdg to the state
   of the router [i.e. the timer and the CBSRFlag..]
   comments: called by age_routes
 */
int ageRPSetTimer()
{
	/* decrease the timer and check if it expired */
	if ((RPSetTimer -= TIMER_INTERVAL) != 0)
		return (FALSE);

	if (debug > 2)
		log(LOG_DEBUG, 0, "\nRP Set timer timed out");

	if (!CBSRFlag) {	/* if I am not a CBSR *//* change your state to accept any following RP-Set */
		if (debug > 2)
			log(LOG_DEBUG, 0, "\nsetting Axpt Any");
		AxptAnyRPSet = 1;
		return (TRUE);
	}
	/* else: I am a CBSR */
	if (debug > 2)
		log(LOG_DEBUG, 0, "\nI am the elected BSR");
	RPSetTimer = BootstrapPeriod;	/* restart RP-Set timer at min thresh */
	originatePeriodicRPSet();
	ElectedBSR = 1;		/* change my state to elected BSR */
	return TRUE;
}

/**************** Handling Candidate RP Set *******************************/

/* function name: search_c_rp_set
   input: RP address, and a pointer to the list to search
   output: int (TRUE if found, FALSE o.w.)
   operation: browses thru the list [assumed ordered list], and 
   compares the RP address passed with the existing
   RP addresses, if found returns TRUE and sets the
   pointer to the element found, else returns
   False and sets the pointer to the point of insertion.
   comments: called by find_c_rp
 */
static int search_c_rp_set(struct g_rplist *setPtr, __u32 address, struct g_rplist **Ptr)
{
	struct g_rplist *p;	/* keep track of prev pointer */

	for (p = NULL, *Ptr = setPtr; *Ptr; p = *Ptr, *Ptr = (*Ptr)->next) {
		if (ntohl(address) > (*Ptr)->rpaddr)
			continue;
		if (ntohl(address) == (*Ptr)->rpaddr) {
			/* found */
			return (TRUE);
		}
		*Ptr = p;
		return (FALSE);
	}
	*Ptr = p;
	return (FALSE);
}

/* function name: create_c_rp_set_entry
   input: address of rp, and the insertion pointer
   output: int (TRUE if successful, False o.w.)
   operation: allocates memory for the new entry and links it to the 
   list thru the given insertion pointer
   comments: called by find_c_rp
 */
static int create_c_rp_set_entry(struct g_rplist **setPtr,
				 __u32 address,
				 struct g_rplist *insertionPtr)
{
	struct g_rplist *newEntry;

	newEntry = (struct g_rplist *) alloc_mem(sizeof(struct g_rplist));
	newEntry->rpaddr = address;
	newEntry->timer = 0;	/* initialize timer */

	if (!insertionPtr) {	/* first element or insertion at head */
		newEntry->next = *setPtr;
		*setPtr = newEntry;
	} else {
		newEntry->next = insertionPtr->next;
		insertionPtr->next = newEntry;
	}

	return (TRUE);
}

/* function name: find_c_rp
   input: source address, and create flag
   output: int; TRUE if successful, FALSE if not.
   operation: searches for the given source address in the c_rp-set,
   if found returns true; o.w. if the create
   is set it creates an entry. Else, returns FALSE.
   comments: is called by update_rp_set
   calls search_c_rp_set and create_c_rp_set_entry
 */
int find_c_rp(__u32 source, int create)
{
	struct g_rplist *c_rp_set_ptr;

	if (debug > 5)
		log(LOG_DEBUG, 0, "In find c_rp");

	if (search_c_rp_set(c_rp_set, source, &c_rp_set_ptr)) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "found %s in C-RP-Set", inet_fmt(source, s1));
		c_rp_set_ptr->timer = 0;	/* reset the timer */
		return (TRUE);
	}
	if (!create) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "NOT found %s in C-RP-Set", inet_fmt(source, s1));
		return (FALSE);
	}
	if (debug > 3)
		log(LOG_DEBUG, 0, "NOT found %s in C-RP-Set, Creating.. ",
		    inet_fmt(source, s1));

	if (create_c_rp_set_entry(&c_rp_set, source, c_rp_set_ptr))
		return (TRUE);

	return (FALSE);
}

/* function name: update_c_rp_set
   input: source address
   output: int (TRUE if added a new rp to the C-RP-Set, FALSE o.w.)
   operation: browses thru the c_rp_set list and searches for the source
   address supplied, if not found, then add it and initialize the 
   timers.. so on,.. if found then just reset the timers.. 
   comments: called by parse_rp_adv
 */
static int update_c_rp_set(__u32 source)
{
	return (find_c_rp(source, TRUE));
}


/* function name: parse_rp_adv
   input: src address and pkt pointer
   output: int
   operation: parses the C-RP-Adv message and calls the processing modules.
   Supporting group prefixes is not supported yet.
   comments: called by accept_rp_adv
   calls update_c_rp_set.
 */
int parse_rp_adv(__u32 src, char *pkt)
{
	u_char prefixCount;
	u_short holdTime;
	__u32 CRP_addr;

	prefixCount = pkt[0];
	holdTime = *((__u16 *) (pkt + 2));
	pkt += 4;
	if (pim2_decode_unicast(&pkt, &CRP_addr) == FALSE)
		return FALSE;
	if (debug > 2)
		log(LOG_DEBUG, 0, "\nC-RP Add is %s", inet_fmt(CRP_addr, s1));
	/* then comes the group mask .. */
	return (update_c_rp_set(CRP_addr));
}

/* function name: accept_rp_adv()
   input: src, dst, grp, pkt, datalen
   output: int (TRUE if accepted, FALSE o.w.)
   operation: processes the candidate-rp-advertisement message
   comments: called by accept_igmp in igmp.c when the PIM code
   is CANDIDATE_RP_ADVERTISEMENT
 */
int accept_rp_adv_v2(__u32 src, __u32 dst, char *p, int datalen)
{
	/* if I am not the bootstrap RP, then do not accept it */
	if (local_address(BSR.rpaddr) == MAXVIFS) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "I am not the bootstrap RP and getting C-RP-Advs... !! Strange.. !!");
		return (FALSE);
	}
	if (debug > 3)
		log(LOG_DEBUG, 0, "Parsing the received C-RP-Adv from %s",
		    inet_fmt(src, s1));

	parse_rp_adv(src, p);

	/* dump the C-RP-Set */
	if (debug > 5)
		dumpRPSet(c_rp_set);

	return (TRUE);
}

static void destroyList(struct g_rplist **list)
{
	struct g_rplist *temp;

	for (; *list;) {
		temp = *list;
		*list = (*list)->next;
		free((char *) temp);
	}
	*list = NULL;
}


/* function name: parse_c_rp_set
   input: packet pointer (char *)
   output: rplist pointer (NULL if failure)
   operation: parses the C-RP-Set and construst an rplist
   comments: called by accept_c_rp_set
 */
static struct g_rplist * parse_c_rp_set(char *packet)
{
	struct g_rplist *list, *prev, *listHead;
	u_char RPCnt, fragRPCnt;
	u_char hashMaskLen, priority, grpMskLen;
	u_short fragTag;	/* the fragment tag, starting byte 5 */
	u_long BSRAddr, grpAddr, hashMsk;
	list = NULL;		/* initialize all to NULL */
	prev = NULL;
	listHead = NULL;

	if (debug > 3)
		log(LOG_DEBUG, 0, "\nParsing RPSet");

	fragTag = *((__u16 *) packet);
	log(LOG_DEBUG, 0, "\nFragtag: %04x", fragTag);
	hashMaskLen = packet[2];
	priority = packet[3];
	packet += 4;
	if (pim2_decode_unicast(&packet, &BSRAddr) == FALSE ||
	    pim2_decode_group(&packet, &grpAddr, &grpMskLen) == FALSE)
		return FALSE;
	RPCnt = packet[0];
	fragRPCnt = packet[1];
	packet += 4;

	hashMsk = (0xffffffff << (32 - hashMaskLen));

	if (debug > 4) {
		log(LOG_DEBUG, 0, "\nFragTag: %u,Msklen: %u,Msk: %x, prio: %u",
		    fragTag, hashMaskLen, hashMsk, priority);
		log(LOG_DEBUG, 0, "BSRAdd: %s,grpAddr: %s,grpMsklen:%u,RPCnt: %u,fragRPCnt %u",
		    inet_fmt(BSRAddr, s1), inet_fmt(grpAddr, s2), grpMskLen, RPCnt, fragRPCnt);
	}

	for (; fragRPCnt; fragRPCnt--) {
		list = (struct g_rplist *) alloc_mem(sizeof(struct rplist));
		if (pim2_decode_unicast(&packet, &list->rpaddr) == FALSE)
			return FALSE;
		packet += 4;
		list->timer = 0;
		if (!prev)
			listHead = list;	/* first element in the new list */
		else
			prev->next = list;
		prev = list;
		list = list->next;
	}
	return (listHead);
}

/* function name: update_dr_c_rp_set
   input: rplist struct; the new list to assign
   output: int (TRUE if successful)
   operation: currently; simply assign the passed list to c_rp_set
   comments: called by accept_c_rp_set
 */
int update_dr_c_rp_set(struct g_rplist *list)
{
	if (!list)
		return (FALSE);

	/* currently; for simplicity, destroy old set and create new and 
	 * then link; 
	 */
	destroyList(&DR_c_rp_set);

	DR_c_rp_set = list;
	return (TRUE);
}

/* function name: storeBSR
   input: BSRPriority, and BSRAddr 
   output: int
   operation: updates BSR and RP-Set timer
   comments: called by accept_c_rp_set
 */
int storeBSR(u_char BSRPriority, __u32 BSRAddress)
{
	LclBSRPrio = BSRPriority;	/* update the BSR prio and add */
	BSR.rpaddr = BSRAddress;
	RPSetTimer = BootstrapTimeout;	/* restart the RP-Set timer */
	return TRUE;
}


/* function name: storeRPSet
   input: BSRPriority, BSRAddr and RP-Set message pointer
   output: int 
   operation: calls procedures to parse and update BSR, RP-Set info
   and RP-Set timer
   comments: called by accept_c_rp_set
 */

int storeRPSet(char *pkt)
{
	struct g_rplist *list;

	if (!(list = parse_c_rp_set(pkt))) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "can't Parse C-RP-Set message.. ");
		return (FALSE);
	}
	return (update_dr_c_rp_set(list));
}

/* function name: prefBSR
   input: BSRPriority and address
   output: int
   operation: compares the priorities [if the given is lower than local,
   returns false, if the given is higher than local, returns true,
   else [i.e. the priorities are equal] compares the addresses,
   if the given address is lower than local returns false, o.w.
   returns true.
   comments: called by accept_c_rp_set
 */
static int prefBSR(u_char priority, __u32 address)
{
	if (priority < LclBSRPrio)
		return (FALSE);

	if (priority > LclBSRPrio)
		return (TRUE);

	if (address < BSR.rpaddr)
		return (FALSE);

	return (TRUE);
}

/* function name: rpf_check
   input: address to check, address of the next hop, and vifi
   output: int (TRUE if the next hop and iif, match those given,
   FALSE o.w.)
   operation: looks up the unicast routing tables and gets the 
   nxt hop and iif for the given address, then checks
   against those given.
   comments: called by accept_c_rp_set
   calls k_req_incoming
 */
static int rpf_check(__u32 address, __u32 next_hop, vifi_t iif)
{
	struct srcentry *s = find_source_r(address, TRUE);

	if (s) {
		if (debug > 5)
			log(LOG_DEBUG, 0, "RPF to %s, nxt hp is %s, iif is %u",
			    inet_fmt(address, s1),
			    s->upstream ? inet_fmt(s->upstream->address, s2) : "direct",
			    s->incoming);
		if (s->incoming == iif &&
		    (!s->upstream || next_hop == s->upstream->address))
			return (TRUE);
	}
	return (FALSE);
}


/* function name: accept_c_rp_set
   input: src, dst, pkt ptr, and datalength
   output: int (TRUE if successful, FALSE o.w.)
   operation: checks the bootstrap RP address then calls appropriate 
   parsing and update functions to process the C-RP-Set message
   comments:
 */
int accept_c_rp_set_v2(__u32 src, __u32 dst, char *pkt, int datalen)
{
	u_char BSRPriority = 0;
	u_long BSRAddr = 0;
	vifi_t vifi;
	int forceAccept = 0;
	char *p = pkt;

	/* to improve convergence forward then process */
	/* copy the RP-Set message to be forwarded as is */
	bcopy(p, pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN, datalen);

	/* get the interface which received the message */
	if ((vifi = vifi_by_index(pim_ifindex)) == MAXVIFS) {
		/* if I sourced the message or non-directly connected, return */
		if (debug > 4)
			log(LOG_INFO, 0, "ignoring RP-Set from router %s",
			    inet_fmt(src, s1));
		return FALSE;
	}
	/* if not sent to all pim routers and I don't have RP-Set info */
	if ((ntohl(dst) != allpimrouters_group) && (!c_rp_set)) {
		/* force acceptance */
		if (debug > 3)
			log(LOG_DEBUG, 0, "###### Unicast RP-Set.... Forcing Accept");
		forceAccept = 1;
	}

	/* parse the message and get the included BSR address and priority */
	BSRPriority = p[3];	/* the BSR priority field */
	if (pim2_decode_unicast(&p, &BSRAddr) == FALSE)
		return FALSE;

	if (local_address(BSRAddr) != MAXVIFS) {
		if (debug > 2)
			log(LOG_DEBUG, 0, "\nRPSet w/ local BSR..!!");
		return (FALSE);
	}
	if (forceAccept) {
		if (prefBSR(BSRPriority, BSRAddr))
			storeBSR(BSRPriority, BSRAddr);
		return (storeRPSet(pkt));
	}
	/* then rpf_check */
	if (!rpf_check(BSRAddr, src, vifi)) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "\nrpf check failed towards %s from %s on i/f %x\n",
			 inet_fmt(BSRAddr, s1), inet_fmt(src, s2), vifi);
		return (FALSE);
	}
	if (!AxptAnyRPSet && !prefBSR(BSRPriority, BSRAddr))
		return (FALSE);

	/* forward the RP-Set msg */
	frwdRPSet(pim_send_buf, datalen, vifi);
	storeBSR(BSRPriority, BSRAddr);
	storeRPSet(pkt);

	/* change the state flags accordingly */
	if (AxptAnyRPSet)
		AxptAnyRPSet = 0;	/* I am not in AxptAny state */
	if (ElectedBSR)
		ElectedBSR = 0;	/* I am not the ElectedBSR anymore */
	return TRUE;
}
