/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: igmp.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"
#include <asm/byteorder.h>
#include <sys/time.h>
#include <unistd.h>

/*
 * DVMRP message types (carried in the "code" field of an IGMP header)
 */
#define DVMRP_PROBE             1	/* for finding neighbors             */
#define DVMRP_REPORT            2	/* for reporting some or all routes  */
#define DVMRP_ASK_NEIGHBORS     3	/* sent by mapper, asking for a list */
					/* of this router's neighbors. */
#define DVMRP_NEIGHBORS         4	/* response to such a request */
#define DVMRP_ASK_NEIGHBORS2    5	/* as above, want new format reply */
#define DVMRP_NEIGHBORS2        6
#define DVMRP_PRUNE             7	/* prune message */
#define DVMRP_GRAFT             8	/* graft message */
#define DVMRP_GRAFT_ACK         9	/* graft acknowledgement */
#define DVMRP_INFO_REQUEST      10	/* information request */
#define DVMRP_INFO_REPLY        11	/* information reply */


/*
 * Mrouted imported stuff for mtrace
 * credits to Bill Fenner / Xerox
 */


/*
 * The packet format for a traceroute request
 */

struct tr_query {
	__u32 tr_src;		/* traceroute source */
	__u32 tr_dst;		/* traceroute destination */
	__u32 tr_raddr;	/* traceroute response address */
#if defined(__LITTLE_ENDIAN_BITFIELD)
	struct {
		__u32 qid:24,	/* traceroute query id */
		       ttl:8;	/* traceroute response ttl */
	} q;
#elif defined(__BIG_ENDIAN_BITFIELD)
	struct {
		__u32 ttl:8,	/* traceroute response ttl */
		      qid:24;	/* traceroute query id */
	} q;
#else
#error "Please fix <asm/byteorder.h>"
#endif				/* BYTE_ORDER */
};

#define tr_rttl q.ttl
#define tr_qid  q.qid

/*
 * Traceroute response format.  A traceroute response has a tr_query at the
 * beginning, followed by one tr_resp for each hop taken.
 */
struct tr_resp {
	__u32 tr_qarr;		/* query arrival time */
	__u32 tr_inaddr;	/* incoming interface address */
	__u32 tr_outaddr;	/* outgoing interface address */
	__u32 tr_rmtaddr;	/* parent address in source tree */
	__u32 tr_vifin;	/* input packet count on interface */
	__u32 tr_vifout;	/* output packet count on interface */
	__u32 tr_pktcnt;	/* total incoming packets for src-grp */
	u_char tr_rproto;	/* routing protocol deployed on router */
	u_char tr_fttl;		/* ttl required to forward on outvif */
	u_char tr_smask;	/* subnet mask for src addr */
	u_char tr_rflags;	/* forwarding error codes */
};

/* defs within mtrace */
#define QUERY	1
#define RESP	2
#define QLEN	sizeof(struct tr_query)
#define RLEN	sizeof(struct tr_resp)

/* fields for tr_rflags (forwarding error codes) */
#define TR_NO_ERR	0
#define TR_WRONG_IF	1
#define TR_PRUNED	2
#define TR_OPRUNED	3
#define TR_SCOPED	4
#define TR_NO_RTE	5
#define TR_NO_FWD	7
#define TR_NO_SPACE	0x81
#define TR_OLD_ROUTER	0x82

/* fields for tr_rproto (routing protocol) */
#define PROTO_DVMRP	1
#define PROTO_MOSPF	2
#define PROTO_PIM	3
#define PROTO_CBT	4

/*
 * 'flags' byte values in DVMRP_NEIGHBORS2 reply.
 */
#define DVMRP_NF_TUNNEL         0x01	/* neighbors reached via tunnel */
#define DVMRP_NF_SRCRT          0x02	/* tunnel uses IP source routing */
#define DVMRP_NF_PIM            0x04	/* neighbor is a PIM neighbor */
#define DVMRP_NF_DOWN           0x10	/* kernel state of interface */
#define DVMRP_NF_DISABLED       0x20	/* administratively disabled */
#define DVMRP_NF_QUERIER        0x40	/* I am the subnet's querier */
#define DVMRP_NF_LEAF           0x80	/* Neighbor reports that it is a leaf */


static void send_dvmrp(char *buf, __u32 src, __u32 dst, int type, int datalen,
		       int ifindex, int mc_loop, __u8 ttl)
{
	struct igmp *igmp = (struct igmp *)(buf + MIN_IP_HEADER_LEN);

	igmp->igmp_group.s_addr = __constant_htonl(MROUTED_LEVEL);
	tracef(TRF_IGMP, "TX DVMRP: from %s to %s type %d len=%d via %d",
	       inet_fmt(src, s1), inet_fmt(dst, s2), 
	       type, datalen, ifindex);
	send_igmp((char *) buf, src, dst, IGMP_DVMRP, type, datalen, ifindex,
		  mc_loop, ttl);
}


void accept_mtrace_request(u_long src, u_long dst, u_long group, char *data, u_int no, int datalen)
{
	u_char type;
	struct grpentry *gt;
	struct mrtentry *r;
	struct srcentry *src_ent;
	struct tr_query *qry;
	struct tr_resp *resp;
	int vifi;
	char *p;
	int rcount;
	int errcode = TR_NO_ERR;
	int resptype;
	struct timeval tp;
	struct sioc_vif_req v_req;
	struct sioc_sg_req sg_req;

	/* Remember qid across invocations */
	static __u32 oqid = 0;

	/* timestamp the request/response */
	gettimeofday(&tp, 0);

	/*
	 * Check if it is a query or a response
	 */
	if (datalen == QLEN) {
		type = QUERY;
		log(LOG_DEBUG, 0, "Initial traceroute query rcvd from %s to %s",
		    inet_fmt(src, s1), inet_fmt(dst, s2));
	} else if ((datalen - QLEN) % RLEN == 0) {
		type = RESP;
		log(LOG_DEBUG, 0, "In-transit traceroute query rcvd from %s to %s",
		    inet_fmt(src, s1), inet_fmt(dst, s2));
		if (IN_MULTICAST(ntohl(dst))) {
			log(LOG_DEBUG, 0, "Dropping multicast response");
			return;
		}
	} else {
		log(LOG_WARNING, 0, "%s from %s to %s",
		    "Non decipherable traceroute request recieved",
		    inet_fmt(src, s1), inet_fmt(dst, s2));
		return;
	}

	qry = (struct tr_query *) data;

	/*
	 * if it is a packet with all reports filled, drop it
	 */
	if (((rcount = (datalen - QLEN) / RLEN) == no) & (no > 0)) {
		log(LOG_DEBUG, 0, "packet with all reports filled in, rcount=%d no=%d",
		    rcount,
		    no);
		return;
	}
	log(LOG_DEBUG, 0, "s: %s g: %s d: %s ",
	    inet_fmt(qry->tr_src, s1),
	    inet_fmt(group, s2),
	    inet_fmt(qry->tr_dst, s3));
	log(LOG_DEBUG, 0, "rttl: %d rd: %s", qry->tr_rttl,
	    inet_fmt(qry->tr_raddr, s1));
	log(LOG_DEBUG, 0, "rcount:%d, qid:%06x", rcount, qry->tr_qid);

	/* determine the routing table entry for this traceroute */

	if ((src_ent = find_source_r(qry->tr_src, TRUE)) != NULL) {
		log(LOG_DEBUG, 0, "rt parent vif: %d rtr: %s metric: %d",
		    src_ent->incoming,
		    inet_fmt(src_ent->upstream ? src_ent->upstream->address : src_ent->source, s1),
		    src_ent->metric);
	} else
		log(LOG_DEBUG, 0, "...no route");

	if (!(r = longest_match_r(qry->tr_src, group)))
		log(LOG_DEBUG, 0, "mtrace: No match for (%s,%s)",
		    inet_fmt(src, s1),
		    inet_fmt(group, s2));

	/*
	 * Query type packet - check if rte exists 
	 * Check if the query destination is a vif connected to me.
	 * and if so, whether I should start response back
	 */
	if (type == QUERY) {
		if (oqid == qry->tr_qid) {
			/*
			 * If the multicast router is a member of the group being
			 * queried, and the query is multicasted, then the router can
			 * recieve multiple copies of the same query.  If we have already
			 * replied to this traceroute, just ignore it this time.
			 *
			 * This is not a total solution, but since if this fails you
			 * only get N copies, N <= the number of interfaces on the router,
			 * it is not fatal.
			 */
			log(LOG_DEBUG, 0, "ignoring duplicate traceroute packet");
			return;
		}
		if (r == NULL) {
			log(LOG_DEBUG, 0, "Mcast traceroute: no route entry %s",
			    inet_fmt(qry->tr_src, s1));
			if (IN_MULTICAST(ntohl(dst)))
				return;
		}
		src_ent = find_source_r(qry->tr_dst, TRUE);
		if (src_ent == NULL || (vifi = src_ent->incoming) == MAXVIFS) {
			/* The traceroute destination is not on one of my subnet vifs. */
			log(LOG_DEBUG, 0, "Destination %s not an interface",
			    inet_fmt(qry->tr_dst, s1));
			if (IN_MULTICAST(ntohl(dst)))
				return;
			errcode = TR_WRONG_IF;
		} else if (r != NULL && !VIFM_ISSET(vifi, r->outgoing)) {
			log(LOG_DEBUG, 0, "Destination %s not on forwarding tree for src %s",
			    inet_fmt(qry->tr_dst, s1), inet_fmt(qry->tr_src, s2));
			if (IN_MULTICAST(ntohl(dst)))
				return;
			errcode = TR_WRONG_IF;
		}
	} else {		/* not QUERY */
		/*
		 * determine which interface the packet came in on
		 * RESP packets travel hop-by-hop so this either traversed
		 * a tunnel or came from a directly attached mrouter.
		 */
		if ((vifi = vifi_by_index(igmp_ifindex)) == MAXVIFS ||
		    neigh_lookup(vifi, src, 3) == NULL) {
			log(LOG_DEBUG, 0, "Wrong interface for packet");
			errcode = TR_WRONG_IF;
		}
	}

	/* Now that we've decided to send a response, save the qid */
	oqid = qry->tr_qid;

	log(LOG_DEBUG, 0, "Sending traceroute response");

	/* copy the packet to the sending buffer */
	p = dvmrp_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN;

	bcopy(data, p, datalen);

	p += datalen;

	/*
	 * If there is no room to insert our reply, coopt the previous hop
	 * error indication to relay this fact.
	 */
	if (p + sizeof(struct tr_resp) > dvmrp_send_buf + MAX_IP_PACKET_LEN) {
		resp = (struct tr_resp *) p - 1;
		resp->tr_rflags = TR_NO_SPACE;
		goto sendit;
	}
	/*
	 * fill in initial response fields
	 */
	resp = (struct tr_resp *) p;
	bzero(resp, sizeof(struct tr_resp));
	datalen += RLEN;

	resp->tr_qarr = htonl((tp.tv_sec + JAN_1970) << 16) +
		((tp.tv_usec >> 4) & 0xffff);

	resp->tr_rproto = PROTO_PIM;
	if (errcode != TR_NO_ERR) {
		resp->tr_rflags = errcode;
		goto sendit;
	}
	resp->tr_outaddr = viflist[vifi].prefsrc;
	resp->tr_fttl = viflist[vifi].threshold;
	resp->tr_rflags = TR_NO_ERR;

	/*
	 * obtain # of packets out on interface
	 */
	v_req.vifi = vifi;
	if (ioctl(igmp_socket, SIOCGETVIFCNT, (char *) &v_req) >= 0)
		resp->tr_vifout = htonl(v_req.ocount);

	/*
	 * fill in scoping & pruning information
	 */
	if (r) {
		for (gt = r->group; gt; gt = gt->next) {
			if (gt->group >= group)
				break;
		}
	} else
		gt = NULL;

	if (gt && gt->group == group) {
		sg_req.src.s_addr = qry->tr_src;
		sg_req.grp.s_addr = group;
		if (ioctl(igmp_socket, SIOCGETSGCNT, (char *) &sg_req) >= 0)
			resp->tr_pktcnt = htonl(sg_req.pktcnt);

		if (VIFM_ISSET(vifi, gt->leaves))
			resp->tr_rflags = TR_SCOPED;
		else
			resp->tr_rflags = TR_NO_FWD;
#if 0
                else if (gt->gt_prsent_timer)
			resp->tr_rflags = TR_PRUNED;
		else if (!VIFM_ISSET(vifi, gt->gt_grpmems))
			if (VIFM_ISSET(vifi, rt->rt_children) &&
			    !VIFM_ISSET(vifi, rt->rt_leaves))
				resp->tr_rflags = TR_OPRUNED;
			else
				resp->tr_rflags = TR_NO_FWD;
#endif
	} else {
#if 0
		if (scoped_addr(vifi, group))
			resp->tr_rflags = TR_SCOPED;
		else if (rt && !VIFM_ISSET(vifi, rt->rt_children))
#endif
		resp->tr_rflags = TR_NO_FWD;
	}

	/*
	 *  if no rte exists, set NO_RTE error
	 */
	if (r == NULL) {
		src = dst;	/* the dst address of resp. pkt */
		resp->tr_inaddr = 0;
		resp->tr_rflags = TR_NO_RTE;
		resp->tr_rmtaddr = 0;
	} else {
		/* get # of packets in on interface */
		v_req.vifi = r->source->incoming;
		if (ioctl(igmp_socket, SIOCGETVIFCNT, (char *) &v_req) >= 0)
			resp->tr_vifin = htonl(v_req.icount);

		/* MASK_TO_VAL(rt->rt_originmask, resp->tr_smask); */
		resp->tr_smask = 32;
		src = viflist[r->source->incoming].prefsrc;
		resp->tr_inaddr = src;
		resp->tr_rmtaddr = viflist[vifi].prefsrc;
		if (!VIFM_ISSET(vifi, r->outgoing)) {
			log(LOG_DEBUG, 0, "Destination %s not on forwarding tree for src %s",
			    inet_fmt(qry->tr_dst, s1), inet_fmt(qry->tr_src, s2));
			resp->tr_rflags = TR_WRONG_IF;
		}
		if (r->source->metric >= UNREACHABLE) {
			resp->tr_rflags = TR_NO_RTE;
		}
	}

sendit:
	/*
	 * if metric is 1 or no. of reports is 1, send response to requestor
	 * else send to upstream router.  If the upstream router can't handle
	 * mtrace, set an error code and send to requestor anyway.
	 */
	log(LOG_DEBUG, 0, "rcount:%d, no:%d", rcount, no);

	if ((rcount + 1 == no) || r == NULL || r->source->upstream == NULL
	    || r->source->direct) {
		resptype = IGMP_MTRACE_RESP;
		dst = qry->tr_raddr;
	} else {
		if (r->upstream)
			dst = r->upstream->address;
		else
			dst = r->source->upstream->address;
		resptype = IGMP_MTRACE;
	}

	if (IN_MULTICAST(ntohl(dst))) {
		/*
		 * Send the reply on a known multicast capable vif.
		 * If we don't have one, we can't source any multicasts anyway.
		 */

		if (r && r->source) {
			log(LOG_DEBUG, 0, "Sending reply to %s from %s via %s",
			    inet_fmt(dst, s1),
			    inet_fmt(viflist[r->source->incoming].prefsrc, s2),
			    viflist[r->source->incoming].name);
			send_igmp(dvmrp_send_buf,
				  viflist[r->source->incoming].prefsrc,
				  dst,
				  resptype,
				  no,
				  datalen,
				  viflist[r->source->incoming].ifindex,
				  0, qry->tr_rttl);
		} else {
			log(LOG_DEBUG, 0, "Sending reply to %s",
			    inet_fmt(dst, s1));
			send_igmp(dvmrp_send_buf,
				  0,
				  dst,
				  resptype,
				  no,
				  datalen,
				  0,
				  0, qry->tr_rttl);
		}
	} else {
		log(LOG_DEBUG, 0, "Sending %s to %s from %s",
		    resptype == IGMP_MTRACE_RESP ? "reply" : "request on",
		    inet_fmt(dst, s1), inet_fmt(src, s2));

		send_igmp(dvmrp_send_buf, src, dst,
			  resptype, no,
			  datalen, 0, -1, MAXTTL);
	}
	return;
}


/* 
 *Support for DVMRP ASK NEIGHBORS
 *
 */

static void accept_neighbor_request(__u32 src, __u32 dst)
{
	vifi_t vifi;
	struct vifentry *v;
	u_char *p, *ncount;
	struct nbrentry *la;
	int datalen;
	unsigned long temp_addr, us, them = src;

	/* Determine which of our addresses to use as the source of our response
	 * to this query.
	 */
	if (IN_MULTICAST(ntohl(dst))) {		/* query sent to a multicast group */
		int udp;	/* find best interface to reply on */
		struct sockaddr_in addr;
		int addrlen = sizeof(addr);

		addr.sin_family = AF_INET;
#if (defined(BSD) && (BSD >= 199103))
		addr.sin_len = sizeof addr;
#endif
		addr.sin_addr.s_addr = dst;
		addr.sin_port = htons(2000);	/* any port over 1024 will do... */
		if ((udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0
		    || connect(udp, (struct sockaddr *) &addr, sizeof(addr)) < 0
		    || getsockname(udp, (struct sockaddr *) &addr, &addrlen) < 0) {
			log(LOG_WARNING, errno, "Determining local address");
			close(udp);
			return;
		}
		close(udp);
		us = addr.sin_addr.s_addr;
	} else			/* query sent to us alone */
		us = dst;

#define PUT_ADDR(a)     temp_addr = ntohl(a); \
                        *p++ = temp_addr >> 24; \
                        *p++ = (temp_addr >> 16) & 0xFF; \
                        *p++ = (temp_addr >> 8) & 0xFF; \
                        *p++ = temp_addr & 0xFF;

	p = (u_char *) (dvmrp_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
	datalen = 0;

	for (vifi = 0, v = viflist; vifi < numvifs; vifi++, v++) {
		if (v->flags & VIFF_DISABLED)
			continue;


		ncount = 0;

		for (la = v->neighbors; la; la = la->next) {
			/* Make sure that there's room for this neighbor.. */
			if ((la->address == 0x0) && !(v->flags & VIFF_TUNNEL))
				continue;
			if (datalen + (ncount == 0 ? 4 + 3 + 4 : 4) > MAX_DVMRP_DATA_LEN) {
				send_dvmrp(dvmrp_send_buf, us, them, DVMRP_NEIGHBORS,
					   datalen, 0, 0, MAXTTL);
				p = (u_char *) (dvmrp_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
				datalen = 0;
				ncount = 0;
			}
			/* Put out the header for this neighbor list... */

			if (ncount == 0) {
				PUT_ADDR(v->prefsrc);
				*p++ = 0x1;	/* v->metric; */
				*p++ = v->threshold;
				ncount = p;
				*p++ = 0;
				datalen += 4 + 3;
			}
			PUT_ADDR(la->address);
			datalen += 4;
			(*ncount)++;
		}
	}

	if (datalen != 0)
		send_dvmrp(dvmrp_send_buf, us, them, DVMRP_NEIGHBORS,
			   datalen, 0, 0, MAXTTL);
}

/*
 * Send a list of all of our neighbors to the requestor, `src'.
 */
static void accept_neighbor_request2(__u32 src, __u32 dst)
{
	vifi_t vifi;
	struct vifentry *v;
	u_char *p, *ncount;
	struct nbrentry *la;
	int datalen;
	unsigned long us, them = src;

	/* Determine which of our addresses to use as the source of our response
	 * to this query.
	 */
	if (IN_MULTICAST(ntohl(dst))) {		/* query sent to a multicast group */
		int udp;	/* find best interface to reply on */
		struct sockaddr_in addr;
		int addrlen = sizeof(addr);

		addr.sin_family = AF_INET;
#if (defined(BSD) && (BSD >= 199103))
		addr.sin_len = sizeof addr;
#endif
		addr.sin_addr.s_addr = dst;
		addr.sin_port = htons(2000);	/* any port over 1024 will do... */
		if ((udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0
		    || connect(udp, (struct sockaddr *) &addr, sizeof(addr)) < 0
		    || getsockname(udp, (struct sockaddr *) &addr, &addrlen) < 0) {
			log(LOG_WARNING, errno, "Determining local address");
			close(udp);
			return;
		}
		close(udp);
		us = addr.sin_addr.s_addr;
	} else			/* query sent to us alone */
		us = dst;

	p = (u_char *) (dvmrp_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
	datalen = 0;

	for (vifi = 0, v = viflist; vifi < numvifs; vifi++, v++) {
		u_short vflags = v->flags;


		register u_char rflags = 0;
		if (vflags & VIFF_TUNNEL)
			rflags |= DVMRP_NF_TUNNEL;
		if (vflags & VIFF_SRCRT)
			rflags |= DVMRP_NF_SRCRT;
		if (vflags & VIFF_DOWN)
			rflags |= DVMRP_NF_DOWN;
		if (vflags & VIFF_DISABLED)
			rflags |= DVMRP_NF_DISABLED;
		if (vflags & VIFF_DR)
			rflags |= DVMRP_NF_QUERIER;
		if (vflags & VIFF_NOMRT)
			rflags |= DVMRP_NF_LEAF;
		ncount = 0;
		la = v->neighbors;
		if (la == NULL) {
			/*
			 * include down & disabled interfaces and interfaces on
			 * leaf nets.
			 */
			if (rflags & DVMRP_NF_TUNNEL)
				rflags |= DVMRP_NF_DOWN;
			if (datalen > MAX_DVMRP_DATA_LEN - 12) {
				send_dvmrp(dvmrp_send_buf, us, them, DVMRP_NEIGHBORS2,
					   datalen, 0, -1, MAXTTL);
				p = (u_char *) (dvmrp_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
				datalen = 0;
			}
			*(u_int *) p = v->prefsrc;
			p += 4;
			*p++ = 1;	/* v->metric; */
			*p++ = v->threshold;
			*p++ = rflags;
			*p++ = 1;
			*(__u32 *) p = v->prefsrc;
			p += 4;
			datalen += 12;
		} else {
			for (; la; la = la->next) {
				if ((la->address == 0x0) && !(v->flags & VIFF_TUNNEL))
					continue;
				/* Make sure that there's room for this neighbor... */
				if (datalen + (ncount == 0 ? 4 + 4 + 4 : 4) > MAX_DVMRP_DATA_LEN) {
					send_dvmrp(dvmrp_send_buf, us, them, DVMRP_NEIGHBORS2,
					  datalen, 0, 0, MAXTTL);
					p = (u_char *) (dvmrp_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
					datalen = 0;
					ncount = 0;
				}
				/* Put out the header for this neighbor list... */
				if (ncount == 0) {
					*(u_int *) p = v->prefsrc;
					p += 4;
					*p++ = 1;	/* v->metric; */
					*p++ = v->threshold;
					*p++ = rflags;
					ncount = p;
					*p++ = 0;
					datalen += 4 + 4;
				}
				*(u_int *) p = la->address;
				p += 4;
				datalen += 4;
				(*ncount)++;
			}
		}
	}
	if (datalen != 0)
		send_dvmrp(dvmrp_send_buf, us, them, DVMRP_NEIGHBORS2,
			   datalen, 0, -1, MAXTTL);

}

void accept_dvmrp(int code, __u32 src, __u32 dst)
{
	switch (code) {
	case DVMRP_ASK_NEIGHBORS:
		accept_neighbor_request(src, dst);
		break;
	case DVMRP_ASK_NEIGHBORS2:
		accept_neighbor_request2(src, dst);
		break;
	}
}
