/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: igmp.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"
#ifndef NO_ANK_FIX
#include <sys/time.h>
#endif

static char igmp_recv_buf[2 * MAX_IP_PACKET_LEN]; /* input packet buffer         */
char pimv1_send_buf[2 * MAX_IP_PACKET_LEN];
char igmp_send_buf[128];
char dvmrp_send_buf[MAX_IP_PACKET_LEN];
int igmp_socket;

u_long allhosts_group;		/* allhosts  addr in net order */
u_long allrouters_group;
char bufftime[32];

int igmp_ifindex;

void init_igmp(void)
{
	struct ip *ip;

	if ((igmp_socket = socket(AF_INET, SOCK_RAW, IPPROTO_IGMP)) < 0)
		log(LOG_ERR, errno, "IGMP socket");


	k_igmp_hdr_include(TRUE);	/* include IP header when sending */
#ifdef NO_ANK_FIX
	k_igmp_set_rcvbuf(2 * 1024 * 1024);	/* increased input buffering even more */
	k_igmp_set_ttl(MINTTL);	/* restrict multicasts to one hop */
#else
	k_igmp_set_rcvbuf(BUFSIZE);
#endif
	k_igmp_set_loop(FALSE);	/* disable multicast loopback     */
	k_igmp_set_rxinfo(TRUE);

	ip = (struct ip *) pimv1_send_buf;
	ip->ip_hl = sizeof(struct ip) >> 2;
	ip->ip_v = IPVERSION;
	ip->ip_tos = 0xc0;
	ip->ip_off = 0L;
	ip->ip_p = IPPROTO_IGMP;
	ip->ip_ttl = MINTTL;


	ip = (struct ip *) igmp_send_buf;
	ip->ip_hl = sizeof(struct ip) >> 2;
	ip->ip_v = IPVERSION;
	ip->ip_tos = 0xC0;
	ip->ip_off = 0L;
	ip->ip_p = IPPROTO_IGMP;
	ip->ip_ttl = MINTTL;

	ip = (struct ip *) dvmrp_send_buf;
	ip->ip_hl = sizeof(struct ip) >> 2;
	ip->ip_v = IPVERSION;
	ip->ip_tos = 0xC0;
	ip->ip_off = 0L;
	ip->ip_p = IPPROTO_IGMP;
	ip->ip_ttl = MAXTTL;	/* applies to unicasts only ??? */


	allhosts_group = htonl(INADDR_ALLHOSTS_GROUP);
	allrouters_group = htonl(INADDR_ALLROUTERS_GROUP);

	if (register_input_handler(igmp_socket, igmp_read) < 0)
		log(LOG_WARNING, 0, "Couldn't register igmp_read as an input handler");
}


/* Read an IGMP message */
void igmp_read(int f, fd_set *rfd)
{
	int igmp_recvlen;
	struct iovec iov = { igmp_recv_buf, sizeof(igmp_recv_buf) };
	char cbuf[128];
	struct cmsghdr *cmsg;
	struct msghdr msg = { 
		NULL, 0,
		&iov,	1,
		cbuf,	sizeof(cbuf),
		0
	};

	igmp_recvlen = recvmsg(igmp_socket, &msg, MSG_DONTWAIT);

	if (igmp_recvlen < 0) {
		if (errno != EINTR && errno != EAGAIN)
			log(LOG_ERR, errno, "IGMP recvmsg");
		return;
	}
	if (msg.msg_flags&MSG_TRUNC) {
		log(LOG_WARNING, 0, "IGMP recvmsg: truncated message");
		return;
	}
	if (msg.msg_flags&MSG_CTRUNC) {
		log(LOG_WARNING, 0, "IGMP recvmsg: truncated control message");
		return;
	}
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_IP ||
		    cmsg->cmsg_type != IP_PKTINFO ||
		    cmsg->cmsg_len < CMSG_LEN(sizeof(struct in_pktinfo))) {
			log(LOG_WARNING, 0, "IGMP recvmsg: too short in_pktinfo");
			return;
		}
		igmp_ifindex = ((struct in_pktinfo*)CMSG_DATA(cmsg))->ipi_ifindex;

		accept_igmp(igmp_recvlen);
		return;
	}
	log(LOG_WARNING, 0, "IGMP recvmsg: no in_pktinfo");
}

static char * igmp_packet_kind(u_char type, u_char code)
{
	switch (type) {
	case IGMPMSG_NOCACHE:
		return "Kernel cache miss ";
	case IGMPMSG_WRONGVIF:
		return "Kernel wrongvif ";
	case IGMPMSG_WHOLEPKT:
		return "Kernel wholepkt ";
	case IGMP_HOST_MEMBERSHIP_QUERY:
		return "membership query ";
	case IGMP_HOST_MEMBERSHIP_REPORT:
		return "membership report ";
	case IGMP_HOST_NEW_MEMBERSHIP_REPORT:
		return "new membership report ";
	case IGMP_DVMRP:
		return "DVMRP message ";
	case IGMP_PIM:
		switch (code) {
		case PIMv1_QUERY:
			return "PIMv1 query ";
		case PIMv1_REGISTER:
			return "PIMv1 register ";
		case PIMv1_REGISTER_STOP:
			return "PIMv1 register_stop ";
		case PIMv1_JOIN_PRUNE:
			return "PIMv1 Join/Prune ";
		case PIMv1_RPREACHABILITY:
			return "PIMv1 RP-Reachability ";
		case PIMv1_ASSERT:
			return "PIMv1 Assert ";
		case PIMv1_GRAFT:
			return "PIMv1 graft ";
		case PIMv1_GRAFT_ACK:
			return "PIMv1 graft ack ";
		default:
			return "unknown PIM msg ";
		}
	case IGMP_MTRACE_RESP:
		return "mtraceroute response";
	case IGMP_MTRACE:
		return "mtraceroute request";
	default:
		return "unknown IGMP msg ";
	}
}

void accept_pim_v1(__u8 code, __u32 src, __u32 dst, int pimlen, void *payload)
{
	tracef(TRF_PIM1, "RX PIMv1: %s from %s to %s",
	       igmp_packet_kind(IGMP_PIM, code),
	       inet_fmt(src, s1), inet_fmt(dst, s2));

	switch (code) {
	case PIMv1_QUERY:
		accept_query_v1(src, dst, payload, pimlen);
		break;

	case PIMv1_REGISTER:
		accept_register_v1(src, dst, payload, pimlen);
		break;

	case PIMv1_REGISTER_STOP:
		accept_register_stop_v1(src, dst, payload, pimlen);
		break;

	case PIMv1_JOIN_PRUNE:
		accept_join_prune_v1(src, dst, payload, pimlen);
		break;

	case PIMv1_RPREACHABILITY:
		accept_rpreach_v1(src, dst, payload, pimlen);
		break;

	case PIMv1_ASSERT:
		accept_assert_v1(src, dst, payload, pimlen);
		break;

	case PIMv1_GRAFT:
		log(LOG_INFO, 0, "ignore PIM GRAFT from %s to %s",
		    inet_fmt(src, s1), inet_fmt(dst, s2));
		break;

	case PIMv1_GRAFT_ACK:
		log(LOG_INFO, 0, "ignore PIM GRAFT-ACK from %s to %s",
		    inet_fmt(src, s1), inet_fmt(dst, s2));
		break;

	default:
		tracef(TRF_PIM1|TRF_UNUSUAL, "RX PIMv1: ignore unknown PIM message code %u from %s to %s",
		       code, inet_fmt(src, s1), inet_fmt(dst, s2));
		break;
	}
}

void accept_igmp(int recvlen)
{
	__u32 src, dst, group;
	register struct ip *ip;
	register struct igmp *igmp;
	int ipdatalen, iphdrlen, pimlen, igmpdatalen;

	if (recvlen < sizeof(struct ip)) {
		log(LOG_WARNING, 0, "packet too short (%u bytes) for IP header",
		    recvlen);
		return;
	}
	ip = (struct ip *) igmp_recv_buf;
	src = ip->ip_src.s_addr;
	dst = ip->ip_dst.s_addr;
	iphdrlen = ip->ip_hl << 2;

#ifndef BROKEN_RAW_SOCKET
	ipdatalen = ntohs(ip->ip_len) - iphdrlen;
#else
	ipdatalen = ip->ip_len;
#endif
	/* pkts sent up from kernelCalls to daemon ip->ip_p = 0 */
	if (ip->ip_p == 0) {
		if (src == 0 || dst == 0) {
			tracef(TRF_MRTSOCK|TRF_IGMP|TRF_UNUSUAL,
			       "kernel request not accurate, %s->%s",
			       inet_fmt(src, s1),
			       inet_fmt(dst, s2));
		} else
			process_kernelCall(igmp_recv_buf);

		return;
	}
	igmp = (struct igmp *) (igmp_recv_buf + iphdrlen);
	group = igmp->igmp_group.s_addr;
	igmpdatalen = pimlen = ipdatalen - IGMP_MINLEN;
	if (pimlen < 0) {
		log(LOG_WARNING, 0,
		    "IGMP:IP data field too short (%u bytes) for IGMP, from %s",
		    ipdatalen, inet_fmt(src, s1));
		return;
	}

	tracef(TRF_IGMP, "RX IGMP: %s from %s to %s via %d",
	       igmp_packet_kind(igmp->igmp_type, igmp->igmp_code),
	       inet_fmt(src, s1),
	       inet_fmt(dst, s2), igmp_ifindex);

	if (inet_cksum((u_short *) igmp, IGMP_MINLEN + igmpdatalen)) {
		log(LOG_WARNING, 0,
		    "IGMP: wrong IGMP cksum, from %s", inet_fmt(src, s1));
		return;
	}

	switch (igmp->igmp_type) {
	case IGMP_DVMRP:
		accept_dvmrp(igmp->igmp_code, src, dst);
		break;

	case IGMP_PIM:
		if (igmp->igmp_group.s_addr != __constant_htonl(0x10000000)) {
			log(LOG_WARNING, 0,
			    "IGMP: PIM version is %08x from %s",
			    igmp->igmp_group.s_addr, inet_fmt(src, s1));
			break;
		}
		accept_pim_v1(igmp->igmp_code, src, dst, pimlen,
			      (void*)(igmp + 1));
		break;
	case IGMP_MTRACE:
		accept_mtrace_request((u_long) src, dst, group, (char *) (igmp + 1),
				      igmp->igmp_code, igmpdatalen);
		break;
	case IGMP_MTRACE_RESP:
		break;

	case IGMP_HOST_MEMBERSHIP_QUERY:
		break;
	case IGMP_HOST_MEMBERSHIP_REPORT:
	case IGMP_HOST_NEW_MEMBERSHIP_REPORT:
		accept_group_report(src, dst, group);
		break;
	case IGMP_HOST_LEAVE_MESSAGE:
		accept_leave_message(src, dst);
		break;

	default:
		tracef(TRF_IGMP|TRF_UNUSUAL,
		       "ignoring unknown IGMP message type %u from %s to %s",
		       igmp->igmp_type,
		       inet_fmt(src, s1),
		       inet_fmt(dst, s2));
		break;
	}
}

void accept_leave_message(u_long src, u_long dst)
{

};

void send_igmp(char *buf,
	       __u32 src,
	       __u32 dst,
	       int type,
	       int code,
	       int datalen,
	       int ifindex,
	       int mc_loop,
	       __u8 ttl)
{
	static struct sockaddr_in sockdst = {AF_INET};
	struct ip *ip;
	struct msghdr *msg;
	struct cmsghdr *cmsg;
	struct iovec iovector;
	struct igmp *igmp;
	char msgbuf[256];
	struct in_pktinfo *pktinfo;

	ip = (struct ip *) buf;

	ip->ip_src.s_addr = src;
	ip->ip_dst.s_addr = dst;
	ip->ip_ttl = ttl;

	igmp = (struct igmp *) (buf + MIN_IP_HEADER_LEN);
	igmp->igmp_type = type;
	igmp->igmp_code = code;
	igmp->igmp_cksum = 0;
	igmp->igmp_cksum = inet_cksum((u_short *) igmp, IGMP_MINLEN + datalen);

	if (!ifindex) {
		if (IN_MULTICAST(ntohl(dst)) || debug > 4)
			log(LOG_WARNING, 0, "send_igmp: unknown iface.");
	}

	if (mc_loop >= 0)
		k_igmp_set_loop(mc_loop);

	sockdst.sin_addr.s_addr = dst;

	tracef(TRF_IGMP, "sending %s from %-15s to %s over if:%d",
	       igmp_packet_kind(type, code),
	       inet_fmt(src, s1),
	       inet_fmt(dst, s2),
	       ifindex);

	cmsg = (struct cmsghdr *) (msgbuf + sizeof(struct msghdr));
	cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(struct in_pktinfo);
	cmsg->cmsg_level = SOL_IP;
	cmsg->cmsg_type = IP_PKTINFO;
	pktinfo = (struct in_pktinfo *) CMSG_DATA(cmsg);
	pktinfo->ipi_ifindex = ifindex;
	pktinfo->ipi_spec_dst.s_addr = src;
	pktinfo->ipi_addr.s_addr = src;

	iovector.iov_base = (void *) buf;
	iovector.iov_len = MIN_IP_HEADER_LEN + IGMP_MINLEN + datalen;
	msg = (struct msghdr *) msgbuf;
	msg->msg_name = &sockdst;
	msg->msg_namelen = sizeof(sockdst);
	msg->msg_iov = &iovector;
	msg->msg_iovlen = 1;
	msg->msg_control = cmsg;
	msg->msg_controllen = cmsg->cmsg_len;

	if (sendmsg(igmp_socket, msg, 0) < 0) {
		log(LOG_WARNING, errno, "sendmsg on %s to %s",
			    inet_fmt(src, s1),
			    inet_fmt(dst, s2));
	}
}


void send_pim_v1(char *buf, __u32 src, __u32 dst, int type, int datalen,
		 int ifindex, int mc_loop, __u8 ttl)
{
	struct igmp *igmp = (struct igmp *)(buf + MIN_IP_HEADER_LEN);

	/* Fill PIM version field. 
	   All the rest send_igmp will make.
	 */

	igmp->igmp_group.s_addr = __constant_htonl(0x10000000);
	tracef(TRF_PIM1, "TX PIMv1: from %s to %s type %d len=%d via %d",
	       inet_fmt(src, s1), inet_fmt(dst, s2), 
	       type, datalen, ifindex);
	send_igmp((char *) buf, src, dst, IGMP_PIM, type, datalen, ifindex,
		  mc_loop, ttl);
}
