/* 
   * Copyright (c) 1996 University of Southern California
   * All rights reserved.
   *
   * Do NOT redistribute.
   *
 */
/*
   * Procedures for the Daemon part of PIM-SMv2.0,
   * Protocol Independent Multicast, Sparse-Mode version 2.0
   *
   * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: main.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>

static char pidfilename[] = "/var/run/pimd.pid";
static char dumpfilename[] = "/var/tmp/pimd.dump";
int debug = 0;
u_long wall_time;
static int timer_ticked;
unsigned trace_flags = TRF_UNUSUAL;
int dirty_bits;

static void start_timer(void);
static void timer_event(void);
static void timer(int);
static void hup(int);
static void fdump(int);
static void dump(int);

#define NHANDLERS       3

static struct ihandler {
	int fd;			/* File descriptor               */
	ihfunc_t func;		/* Function to call with &fd_set */
} ihandlers[NHANDLERS];
static int nhandlers = 0;

int register_input_handler(int fd, ihfunc_t func)
{
	if (nhandlers >= NHANDLERS)
		return -1;

	ihandlers[nhandlers].fd = fd;
	ihandlers[nhandlers++].func = func;

	return 0;
}

static void set_signal(int signo, void (*handler)(int))
{
	struct sigaction sa;
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;
	sa.sa_flags = SA_INTERRUPT;
	sigaction(signo, &sa, NULL);	
} 

static void usage(void)
{
	fprintf(stderr, "usage: pimd [-c configfile] [-d [debug_level]] [-t trace_flags]\n");
	exit(1);
}

static void dump_info(void)
{
	if (trace_flags&TRF_VIFS && dirty_bits&1)
		dump_vifs(stderr);
	if (trace_flags&TRF_MRTABLE && dirty_bits&2)
		dump_routes(stderr);
	if (trace_flags&TRF_RPSET && dirty_bits&4)
		dump_auto_rp_set(stderr);
	dirty_bits = 0;
}

int main(int argc, char *argv[])
{
	int t;
	FILE *fp;

	fd_set rfds, readers;
	int nfds, n, i;

	setlinebuf(stderr);

	if (geteuid() != 0) {
		fprintf(stderr, "must be root\n");
		exit(1);
	}
	argv++;
	argc--;
	while (argc > 0 && *argv[0] == '-') {
		if (strcmp(*argv, "-d") == 0) {
			if (argc > 1 && isdigit(*(argv + 1)[0])) {
				argv++;
				argc--;
				debug = atoi(*argv);
			} else
				debug = DEBUG;
		} else if (strcmp(*argv, "-t") == 0) {
			if (--argc > 0) {
				int negate = 0;
				unsigned msk = 0;
				++argv;
				if (*argv[0] == '!') {
					negate = 1;
					(*argv)++;
				}
				if (strcmp(*argv, "unusual") == 0)
					msk = TRF_UNUSUAL;
				else if (strcmp(*argv, "igmp") == 0)
					msk = TRF_IGMP;
				else if (strcmp(*argv, "mrtsock") == 0)
					msk = TRF_MRTSOCK;
				else if (strcmp(*argv, "pim") == 0)
					msk = TRF_PIM;
				else if (strcmp(*argv, "pim1") == 0)
					msk = TRF_PIM1;
				else if (strcmp(*argv, "pim2") == 0)
					msk = TRF_PIM2;
				else if (strcmp(*argv, "autorp") == 0)
					msk = TRF_AUTORP;
				else if (strcmp(*argv, "netlink") == 0)
					msk = TRF_NETLINK;
				else if (strcmp(*argv, "mrtable") == 0)
					msk = TRF_MRTABLE;
				else if (strcmp(*argv, "vifs") == 0)
					msk = TRF_VIFS;
				else if (strcmp(*argv, "rpset") == 0)
					msk = TRF_RPSET;
				else if (strcmp(*argv, "all") == 0)
					msk = ~0;
				else
					usage();
				if (negate)
					trace_flags &= ~msk;
				else
					trace_flags |= msk;
			} else
				usage();
		} else if (strcmp(*argv, "-c") == 0) {
			if (argc > 1) {
				argv++;
				argc--;
				configfilename = *argv;
			} else {
				fprintf(stderr, "usage: pimd [-c configfile] [-d [debug_level]]\nSome recomended debug levels:\n1: basic PIM messages\n2: plus routing decisions\n5: a lot of stuff");
				exit(1);
			}
		} else {
			fprintf(stderr, "usage: pimd [-c configfile] [-d [debug_level]]\n");
			exit(1);
		}
		argv++;
		argc--;
	}

	if (argc > 0) {
		fprintf(stderr, "usage: pimd [-c configfile] [-d [debug_level]]\n");
		exit(1);
	}
	if (debug == 0) {
		if (fork())
			exit(0);
		(void) close(0);
		(void) close(1);
		(void) close(2);
		(void) open("/dev/null", 0);
		(void) dup2(0, 1);
		(void) dup2(0, 2);
		t = open("/dev/tty", 2);
		if (t >= 0) {
			(void) ioctl(t, TIOCNOTTY, (char *) 0);
			(void) close(t);
		}
	}
#ifdef LOG_DAEMON
	(void) openlog("pimd", LOG_PID, LOG_DAEMON);
	(void) setlogmask(LOG_UPTO(LOG_NOTICE));
#else
	(void) openlog("pimd", LOG_PID);
#endif				/* LOG_DAEMON */

	fp = fopen(pidfilename, "w");
	if (fp != NULL) {
		fprintf(fp, "%d\n", getpid());
		(void) fclose(fp);
	}
	srandom(gethostid());

	init_igmp();
	init_pim();
	init_routesock();
	init_autorp();
	init_vifs();
	log(LOG_NOTICE, 0, "initialized vifs");

	if (debug >= 2)
		dump(0);

	set_signal(SIGALRM, timer);
	set_signal(SIGHUP, hup);
	set_signal(SIGTERM, hup);
	set_signal(SIGINT, hup);
	set_signal(SIGUSR1, fdump);
	if (debug != 0)
		(void) set_signal(SIGQUIT, dump);

	nfds = 1;
	FD_ZERO(&readers);
	for (i = 0; i < nhandlers; i++) {
		FD_SET(ihandlers[i].fd, &readers);
		if (ihandlers[i].fd >= nfds)
			nfds = ihandlers[i].fd + 1;
	}

	start_timer();

	for (;;) {
		bcopy((char *) &readers, (char *) &rfds, sizeof(rfds));

		if ((n = select(nfds, &rfds, NULL, NULL, NULL)) > 0) {
			wall_time = time(NULL);
			for (i = 0; i < nhandlers; i++) {
				if (FD_ISSET(ihandlers[i].fd, &rfds)) {
					(*ihandlers[i].func) (ihandlers[i].fd, &rfds);
					if (debug && dirty_bits)
						dump_info();
				}
			}
		} else if (n < 0 && errno != EINTR)
			log(LOG_ERR, errno, "select failed");

		if (timer_ticked) {
			wall_time = time(NULL);
			timer_event();
			if (debug && dirty_bits)
				dump_info();
		}
	}
}


static void timer(int signo)
{
	timer_ticked = 1;
}

static void timer_event()
{
	extern int autorp_announce;

	age_vifs();		/* Advance the timers for neighbors and groups */
	age_routes();		/* Advance the timers in the route entries     */

	if (autorp_announce)
		autorp_send_announce();

	timer_ticked = 0;
}

static void start_timer(void)
{
	struct itimerval it;
	memset(&it, 0, sizeof(it));
	it.it_value.tv_sec = 1;
	it.it_interval.tv_sec = TIMER_INTERVAL;
	setitimer(ITIMER_REAL, &it, NULL);
}

static void hup(int signo)
{
	log(LOG_INFO, 0, "hup");
	dump(signo);
	exit(1);
}


static void dump(int signo)
{
	dump_vifs(stderr);
	dump_routes(stderr);
	dump_auto_rp_set(stderr);
}


static void fdump(int signo)
{
	FILE *fp;

	fp = fopen(dumpfilename, "w");
	if (fp != NULL) {
		dump_vifs(fp);
		dump_routes(fp);
		dump_auto_rp_set(fp);
		(void) fclose(fp);
	}
}


void log(int severity, int syserr, char *format, ...)
{
	va_list args;
	char fmt[1024];

	va_start(args, format);

	switch (debug) {
	case 0:
		break;
	case 1:
		if (severity > LOG_NOTICE)
			break;
	case 2:
		if (severity > LOG_INFO)
			break;
	default:
		fmt[0] = '\0';
		if (severity == LOG_WARNING)
			strcat(fmt, "warning - ");
		strncat(fmt, format, 80);
		vfprintf(stderr, fmt, args);
		if (syserr == 0)
			fprintf(stderr, "\n");
		else if (syserr < sys_nerr)
			fprintf(stderr, ": %s\n", sys_errlist[syserr]);
		else
			fprintf(stderr, ": errno %d\n", syserr);
	}

	if (severity <= LOG_NOTICE) {
		fmt[0] = '\0';
		if (severity == LOG_WARNING)
			strcat(fmt, "warning - ");
		strncat(fmt, format, 80);
		if (syserr != 0) {
			strcat(fmt, ": %m");
			errno = syserr;
		}
		vsyslog(severity, fmt, args);

		if (severity <= LOG_ERR)
			exit(-1);
	}
	va_end(args);
}


void do_tracef (char* fmt, ...)
{
	va_list args;
	va_start(args, fmt);

	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");

	va_end(args);
} 



void * alloc_mem(int size)
{
	void *c;

	if ((c = malloc(size)) == NULL)
		log(LOG_ERR, 0, "ran out of memory");

	return (memset(c, 0, size));
}

