/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: mrt.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"


/************************ Source List manipulations ***********************/

struct srcentry *srchash[256];


/* function name: set_src_iif
   input: srcentry
   output: int (TRUE if the nbr [nxt hop] was found, FALSE o.w.)
   operation: calls k_req_incoming (the kernel interface to get the iif)
   gets the iif and the next hop info. and process them to see if the
   next hop is a PIM nbr on the iif; if not, assigns null to upstream
   comments: called by create_source
   calls k_req_incoming
 */
int set_src_iif(struct srcentry *src_ent)
{
	struct rpfctl rpfc;
	__u32 nbr;
	register struct vifentry *v;
	register struct nbrentry *n;
	__u32 source = src_ent->source;
	int directlyConnected = 0;

	src_ent->metric = 0xFFFFFFFF;
	src_ent->preference = 0x7FFFFFFF;
	src_ent->asserted = 0;
	src_ent->direct = 0;

	if (debug > 4)
		log(LOG_DEBUG, 0, "set_src_iif: for %s", inet_fmt(source, s1));

	/* check if the add is local or directly connected .. */
	if ((rpfc.iif = local_address(source)) != MAXVIFS) {	/* if local address */
		src_ent->incoming = rpfc.iif;
		src_ent->upstream = NULL;
		src_ent->metric = 0;
		src_ent->preference = 0;
		src_ent->asserted = 1;
		src_ent->direct = 2;
		return (TRUE);
	}

	rpfc.source.s_addr = source;
	if (!k_req_incoming(&rpfc) || rpfc.rpfneighbor.s_addr == 0) {
		log(LOG_DEBUG, 0, "No such route");
		return FALSE;
	}

	src_ent->incoming = rpfc.iif;
	nbr = rpfc.rpfneighbor.s_addr;
	if (debug > 3)
		log(LOG_DEBUG, 0, "set_src_iif:Looking for neighbor %s", inet_fmt(nbr, s1));
	v = &viflist[src_ent->incoming];

	if (nbr == source) {
		src_ent->metric = v->metric;
		src_ent->preference = 0;
		src_ent->asserted = 1;
		src_ent->direct = 1;
		directlyConnected = 1;
	}

	n = neigh_lookup(rpfc.iif, nbr, 3);
	if (n) {
		if (!directlyConnected) {
			src_ent->metric = v->metric + 1;
			src_ent->preference = 0x7FFFFFFE;
		}
		src_ent->upstream = n;
		return TRUE;
	}

	if (debug > 4)
		log(LOG_INFO, 0, "set_src_iif:neighbor not found, setting nexthop to NULL");

	src_ent->upstream = NULL;

	/* to allow registering for
	 * directly connected hosts.
	 */
	if (directlyConnected) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "set_src_iif: Directly connected!");
		src_ent->asserted = 1;
		return (TRUE);
	}
	log(LOG_DEBUG, 0, "Failed to route %s via %d %s",
	    inet_fmt(src_ent->source, s1),
	    rpfc.iif,
	    inet_fmt(nbr, s2));
	return (FALSE);
}

static struct srcentry * create_source(unsigned hidx, __u32 source)
{
	struct srcentry *s;

	s = (struct srcentry *) alloc_mem(sizeof(struct srcentry));
	s->source = source;

	/* free if not using */
	if (set_src_iif(s) == FALSE) {
		free(s);
		if (debug)
			log(LOG_DEBUG, 0, "create_source: srcentry is not created! (%s)", inet_fmt(source, s1));
		return NULL;
	}

	/* initialize mrtlink to NULL, assigned later in find_route */
	s->mrtlink = NULL;
	s->uptime = wall_time;

	s->next = srchash[hidx];
	srchash[hidx] = s;
	s->hash = hidx;
	return s;
}

struct srcentry * find_source_r(__u32 source, char create)
{
	struct srcentry *s;
	unsigned hidx = source;

	hidx ^= (hidx>>16);
	hidx ^= (hidx>>8);
	hidx &= 0xFF;

	for (s = srchash[hidx]; s; s = s->next) {
		if (s->source == source)
			return s;
	}

	if (!create) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "Source %s is not found, exiting find source",
			    inet_fmt(source, s1));
		return (FALSE);
	}

	if (source==0 || IN_MULTICAST(ntohl(source))) {
		log(LOG_WARNING, 0, "find_source_r: invalid source %s", inet_fmt(source, s1));
		return NULL;
	}

	return create_source(hidx, source);
}

/**************** Group List manipulations *********************************/

struct grpentry *grphash[256];


/* function name: set_grp_rpmap
   input: group entry pointer
   output: int (TRUE if the map was found and assigned appropriately)
   operation: use the hash function and the DR C-RP-Set to get RP to Grp
   mapping for the given group entry.
   comments: called by create_group
 */
static int set_grp_rpmap(struct grpentry *grp_ent)
{
	struct g_rplist *r;
#if 0
	int i, rp_count = 0;
	u_long maxWeight = 0, weight = 0;
#endif
	u_long RPToMap = 0;
	u_long group;
	int glen = 0;
	int pvers = 0;

	/* if I don't have a C-RP-Set then return error */
	if (!auto_rp_set) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "I have No DR C RP Set !!!");
		return (FALSE);
	}
	/* use the hash function to map the group to one C-RP from 
	 * the C-RP-Set 
	 */
	group = (grp_ent->group & hashMask);

	for (r = auto_rp_set; r; r = r->next) {
		if ((r->group^grp_ent->group)&r->grpmask)
			continue;
		if (r->expires < wall_time)
			continue;
#if 0
		srand(group);
		srand(rand() ^ r->rpaddr);
		weight = rand();
		if (weight >= maxWeight) {
			maxWeight = weight;
			RPToMap = r->rpaddr;
		}
#else
		if (r->grpmasklen < glen)
			break;
		glen = r->grpmasklen;
		if (ntohl(RPToMap) < ntohl(r->rpaddr)) {
			RPToMap = r->rpaddr;
			pvers = r->pimversion;
		}
#endif
	}

	if (debug > 5)
		log(LOG_DEBUG, 0, "For Grp %s, the chosen rp is %s",
		    inet_fmt(group, s2), inet_fmt(RPToMap, s1));

	if (!RPToMap)
		return (FALSE);

	if (!grp_ent->active_rp)	/* if the entry does not exist
					   then allocate memory for it */
		grp_ent->active_rp = (struct rplist *)
			alloc_mem(sizeof(struct rplist));

	grp_ent->active_rp->rpaddr = RPToMap;
	if (pvers == 0)
		pvers = 1;
	grp_ent->active_rp->pimversion = pvers;
	return (TRUE);
}

/* function name: create_group
   input: group address, and the insertion pointer.
   output: int (TRUE if created correctly)
   operation: allocates memory for the new group entry, and inserts it
   after the insertion point passed.
   comments: called by find_group
   calls set_grp_rpmap
 */
static struct grpentry * create_group(unsigned hidx, __u32 group)
{
	struct grpentry *g;

	g = (struct grpentry *) alloc_mem(sizeof(struct grpentry) + sizeof(u_char) * numvifs);
	g->flags = 0;
	g->leaves = 0;
	g->group = group;
	g->mrtlink = NULL;
	g->uptime = wall_time;

	/* if I can't find the mapping, then don't link group; rather destroy */
	if (!set_grp_rpmap(g)) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "Couldn't get mapping deleting new entry");
		free((char *) g);
		return NULL;
	}
	g->next = grphash[hidx];
	grphash[hidx] = g;
	g->hash = hidx;
	return g;
}

struct grpentry * find_group_r(__u32 group, char create)
{
	struct grpentry *g;
	unsigned hidx = group;

	hidx ^= (hidx>>16);
	hidx ^= (hidx>>8);
	hidx &= 0xFF;

	for (g = grphash[hidx]; g; g = g->next) {
		if (g->group == group)
			return g;
	}

	if (!create) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "Not found, exiting find Group");
		return NULL;
	}

	if (!IN_MULTICAST(ntohl(group))) {
		log(LOG_WARNING, 0, "find_group_r: invalid group %s", inet_fmt(group, s1));
		return NULL;
	}

	return create_group(hidx, group);
}


/****************** Multicast Routing Table (MRT) manipulations ************/

struct mrtentry *mrthash[256];

/* function name: change_flags
   input: pointer to the mrt entry, and the flags
   output: int, TRUE if the flags were changed, FALSE o.w.
   operation: checks the change in flags, if a change happened,
   then updates the flags giving special care to RP and WC 
   updates.
   comments: called by create_mrt
 */
static int change_flags(struct mrtentry * mrtPtr, u_short flags)
{
	if (mrtPtr->flags == flags)
		return (FALSE);

	if ((flags & MRTF_WC) && (flags & MRTF_RP))
		mrtPtr->group->rp_entry = mrtPtr;

	mrtPtr->flags |= flags;
	dirty_bits |= 2;
	return (TRUE);
}

/* function name : alloc_route
   input: grp entry, src entry, and flags.
   output: mrtentry 
   operation: allocates memory for an mrt entry to hold the source and
   group entries passed, with the flags.  
   comments: called by find_route
 */
static struct mrtentry *
alloc_route(struct srcentry *s, struct grpentry *g, u_short flags)
{
	struct mrtentry *r;

	if ((r = alloc_mem(sizeof(struct mrtentry) + sizeof(u_char) * numvifs)) == NULL)
		log(LOG_ERR, 0, "ran out of memory");

	r->group = g;
	r->source = s;
	r->flags = flags | MRTF_CRT;
	return (r);
}

static struct mrtentry * create_mrt(unsigned hidx, __u32 source, __u32 group, u_short flags)
{
	struct grpentry *g;
	struct srcentry *s;
	struct mrtentry *r;

	s = find_source_r(source, TRUE);
	if (!s) {
		if (debug > 2)
			log(LOG_DEBUG, 0, "create_mrt (%s,%s) flags %04x failed. Bad source.",
			    inet_fmt(source, s1),
			    inet_fmt(group, s2),
			    flags);
		return NULL;
	}

	g = find_group_r(group, TRUE);
	if (!g) {
		if (debug > 2)
			log(LOG_DEBUG, 0, "create_mrt (%s,%s) flags %04x failed. Bad group.",
			    inet_fmt(source, s1),
			    inet_fmt(group, s2),
			    flags);
		return NULL;
	}

	r = alloc_route(s, g, flags);
	if (!r) {
		if (debug > 2)
			log(LOG_DEBUG, 0, "create_mrt (%s,%s) flags %04x failed. Route alloc.",
			    inet_fmt(source, s1),
			    inet_fmt(group, s2),
			    flags);
		return NULL;
	}

	r->next = mrthash[hidx];
	mrthash[hidx] = r;
	r->hash = hidx;

	r->grpnext = g->mrtlink;
	g->mrtlink = r;

	r->srcnext = s->mrtlink;
	s->mrtlink = r;

	change_flags(r, flags);

	dirty_bits |= 2;

	if (s->asserted == 0 && s->upstream) {
		if ((r->flags & MRTF_RP) == 0) {
			tracef(TRF_PIM, "asserting %s%s", inet_fmt(s->source, s1), r->flags&MRTF_RP ? " RPT" : "");
			send_assert_v1(r, &viflist[s->upstream->vif]);
		}
	}

	return r;
}

void unlink_route(struct mrtentry *r)
{
	struct mrtentry *r1, **r1p;

	for (r1p = &mrthash[r->hash]; (r1 = *r1p) != NULL; r1p = &r1->next) {
		if (r1 == r) {
			*r1p = r1->next;
			break;
		}
	}
	for (r1p = &r->source->mrtlink; (r1 = *r1p) != NULL; r1p = &r1->srcnext) {
		if (r1 == r) {
			*r1p = r1->srcnext;
			break;
		}
	}
	for (r1p = &r->group->mrtlink; (r1 = *r1p) != NULL; r1p = &r1->grpnext) {
		if (r1 == r) {
			*r1p = r1->grpnext;
			break;
		}
	}
}

struct mrtentry * find_route_r(__u32 source, __u32 group, u_short flags, char create)
{
	register struct mrtentry *r;
	unsigned hidx = source;
	
	hidx ^= ntohl(group)<<4;
	hidx ^= (hidx>>16);	
	hidx ^= (hidx>>8);
	hidx &= 0xFF;

	for (r = mrthash[hidx]; r; r = r->next) {
		if (r->source->source == source && r->group->group == group) {
			if (!create) {
				if (!flags || (r->flags&flags))
					return r;
				if (debug > 3)
					log(LOG_DEBUG, 0, "(%s,%s) with flags %04x is not found. Real flags are %04x",
					    inet_fmt(source, s1),
					    inet_fmt(group, s2), flags, r->flags);
				return NULL;
			}
			change_flags(r, flags);
			return r;
		}
	}

	if (!create) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "(%s,%s) is not found", inet_fmt(source, s1),
			    inet_fmt(group, s2));
		return NULL;
	}

	return create_mrt(hidx, source, group, flags);
}

