/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: pim.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"

char pim_recv_buf[MAX_IP_PACKET_LEN];	/* input packet buffer   */
char pim_send_buf[MAX_IP_PACKET_LEN];	/* output packet buffer  */
__u32 allpimrouters_group;	/* ALL_PIM_ROUTERS group */
int pim_socket;			/* socket for PIM control msgs */

int pim_ifindex;
__u8 pim_recvttl;

static void pim_read(int f, fd_set *rfd);
static void accept_pim(int recvlen);

void init_pim()
{
	struct ip *ip;

	if ((pim_socket = socket(AF_INET, SOCK_RAW, IPPROTO_PIM)) < 0) {
		log(LOG_ERR, errno, "PIM socket");
		abort(0);
	}
	k_pim_hdr_include(TRUE);	/* include IP header when sending */
	k_pim_set_rcvbuf(BUFSIZE);	/* lots of input buffering        */
	k_pim_set_loop(FALSE);		/* disable multicast loopback     */
	k_pim_set_rxinfo(TRUE);

	allpimrouters_group = htonl(INADDR_ALL_PIM_ROUTERS);

	ip = (struct ip *) pim_send_buf;
	ip->ip_hl = sizeof(struct ip) >> 2;
	ip->ip_v = IPVERSION;
	ip->ip_tos = 0xc0;
	ip->ip_off = 0L;
	ip->ip_p = IPPROTO_PIM;
	ip->ip_ttl = MINTTL;

	if (register_input_handler(pim_socket, pim_read) < 0)
		log(LOG_WARNING, errno, "Couldn't register RSRR as an input handler");
}

/* Read an PIM message */
static void pim_read(int f, fd_set *rfd)
{
	int recvlen;
	struct iovec iov = { pim_recv_buf, sizeof(pim_recv_buf) };
	char cbuf[128];
	struct cmsghdr *cmsg;
	struct msghdr msg = { 
		NULL, 0,
		&iov,	1,
		cbuf,	sizeof(cbuf),
		0
	};

	recvlen = recvmsg(pim_socket, &msg, MSG_DONTWAIT);

	if (recvlen < 0) {
		if (errno != EINTR && errno != EAGAIN)
			log(LOG_ERR, errno, "PIM recvmsg");
		return;
	}
	if (msg.msg_flags&MSG_TRUNC) {
		log(LOG_WARNING, 0, "PIM recvmsg: truncated message");
		return;
	}
	if (msg.msg_flags&MSG_CTRUNC) {
		log(LOG_WARNING, 0, "PIM recvmsg: truncated control message");
		return;
	}
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_IP ||
		    cmsg->cmsg_type != IP_PKTINFO ||
		    cmsg->cmsg_len < CMSG_LEN(sizeof(struct in_pktinfo))) {
			log(LOG_WARNING, 0, "PIM recvmsg: too short in_pktinfo");
			return;
		}
		pim_ifindex = ((struct in_pktinfo*)CMSG_DATA(cmsg))->ipi_ifindex;

		accept_pim(recvlen);
		return;
	}
	log(LOG_WARNING, 0, "PIM recvmsg: no in_pktinfo");
}

static char * pim_packet_kind(u_char type)
{
	switch (type) {
	case PIM_HELLO:
		return "PIMv2 hello ";
	case PIM_REGISTER:
		return "PIMv2 register ";
	case PIM_REGISTER_STOP:
		return "PIMv2 register_stop ";
	case PIM_JOIN_PRUNE:
		return "PIMv2 Join/Prune ";
	case PIM_BOOTSTRAP:
		return "PIMv2 bootstrap ";
	case PIM_ASSERT:
		return "PIMv2 assert ";
	case PIM_GRAFT:
		return "PIMv2 graft ";
	case PIM_GRAFT_ACK:
		return "PIMv2 graft ack ";
	case PIM_CANDIDATE_RP_ADVERTISEMENT:
		return "PIMv2 candidate RP advertisement ";
	default:
		return "unknown PIMv2 msg ";
	}
}


static void accept_pim(int recvlen)
{
	__u32 src, dst;
	struct ip *ip;
	struct pimhead *pim;
	int ipdatalen, iphdrlen, pimlen;

	if (recvlen < sizeof(struct ip)) {
		log(LOG_WARNING, 0, "packet too short (%u bytes) for IP header",
		    recvlen);
		return;
	}
	ip = (struct ip *) pim_recv_buf;
	src = ip->ip_src.s_addr;
	dst = ip->ip_dst.s_addr;
	pim_recvttl = ip->ip_ttl;
	iphdrlen = ip->ip_hl << 2;
#ifndef BROKEN_RAW_SOCKET
	ipdatalen = ntohs(ip->ip_len) - iphdrlen;
#else
	ipdatalen = ip->ip_len;
#endif

	if (iphdrlen + ipdatalen != recvlen) {
		log(LOG_WARNING, 0,
		"packet (%u bytes) no match with hdr+data length (%u+%u)",
		    recvlen, iphdrlen, ipdatalen);
		return;
	}
	pim = (struct pimhead *) (pim_recv_buf + iphdrlen);
	pimlen = ipdatalen - PIM_MINLEN;
	if (pimlen < 0) {
		log(LOG_WARNING, 0,
		    "PIM:IP data field too short (%u bytes) for PIM, from %s",
		    ipdatalen, inet_fmt(src, s1));
		return;
	}
	tracef(TRF_PIM2, "RX PIMv2: %s (vers %d) from %s to %s",
	       pim_packet_kind(pim->pim_vertype&0xF), pim->pim_vertype>>4,
	       inet_fmt(src, s1), inet_fmt(dst, s2));
	if ((pim->pim_vertype>>4) != 2) {
		log(LOG_WARNING, 0,
		    "PIM: wrong PIM version %d, from %s", pim->pim_vertype>>4,
		    inet_fmt(src, s1));
		return;
	}
	if (inet_cksum((u_short *) pim, PIM_MINLEN + pimlen)) {
		log(LOG_WARNING, 0,
		    "PIM: wrong PIM cksum, from %s", inet_fmt(src, s1));
		return;
	}

	switch (pim->pim_vertype & 0xF) {
	case PIM_HELLO:
		accept_query_v2(src, dst, (char *) (pim+1), pimlen);
		break;

	case PIM_REGISTER:
		accept_register_v2(src, dst, (char *) (pim+1), pimlen);
		break;

	case PIM_REGISTER_STOP:
		accept_register_stop_v2(src, dst, (char *) (pim+1), pimlen);
		break;

	case PIM_JOIN_PRUNE:
		accept_join_prune_v2(src, dst, (char *) (pim+1), pimlen);
		break;

	case PIM_BOOTSTRAP:
		accept_c_rp_set_v2(src, dst, (char *) (pim+1), pimlen);
		break;

	case PIM_ASSERT:
		accept_assert_v2(src, dst, (char *) (pim+1), pimlen);
		break;

	case PIM_CANDIDATE_RP_ADVERTISEMENT:
		accept_rp_adv_v2(src, dst, (char *) (pim+1), pimlen);
		break;


	default:
		log(LOG_INFO, 0, "ignore unknown PIMv2 message code %u from %s to %s",
		    pim->pim_vertype&0xF, inet_fmt(src, s1), inet_fmt(dst, s2));
		break;
	}
}

void send_pim(char *buf, __u32 src, __u32 dst, int type, int datalen, int ifindex, __u8 ttl)
{
	static struct sockaddr_in sockdst = {AF_INET};
	struct ip *ip;
	register struct pimhead *pim;

	struct msghdr *msg;
	struct cmsghdr *cmsg;
	struct iovec iovector;
	char msgbuf[256];
	struct in_pktinfo *pktinfo;

	ip = (struct ip *) buf;
	ip->ip_src.s_addr = src;
	ip->ip_dst.s_addr = dst;
	ip->ip_ttl = ttl;

	pim = (struct pimhead *) (buf + MIN_IP_HEADER_LEN);
	pim->pim_vertype = (2<<4)|type;
	pim->pim_reserved = 0;
	pim->pim_cksum = 0;
	pim->pim_cksum = inet_cksum((u_short *) pim, PIM_MINLEN + datalen);

	if (!ifindex) {
		if (IN_MULTICAST(ntohl(dst)) || debug > 4)
			log(LOG_WARNING, 0, "send_pim: unknown iface.");
	}

	sockdst.sin_addr.s_addr = dst;

	tracef(TRF_PIM2, "sending %s from %-15s to %s over if:%d",
	       pim_packet_kind(type),
	       inet_fmt(src, s1),
	       inet_fmt(dst, s2),
	       ifindex);

	cmsg = (struct cmsghdr *) (msgbuf + sizeof(struct msghdr));
	cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(struct in_pktinfo);
	cmsg->cmsg_level = SOL_IP;
	cmsg->cmsg_type = IP_PKTINFO;
	pktinfo = (struct in_pktinfo *) CMSG_DATA(cmsg);
	pktinfo->ipi_ifindex = ifindex;
	pktinfo->ipi_spec_dst.s_addr = src;
	pktinfo->ipi_addr.s_addr = src;

	iovector.iov_base = (void *) buf;
	iovector.iov_len = MIN_IP_HEADER_LEN + PIM_MINLEN + datalen;
	msg = (struct msghdr *) msgbuf;
	msg->msg_name = &sockdst;
	msg->msg_namelen = sizeof(sockdst);
	msg->msg_iov = &iovector;
	msg->msg_iovlen = 1;
	msg->msg_control = cmsg;
	msg->msg_controllen = cmsg->cmsg_len;

	if (sendmsg(pim_socket, msg, 0) < 0) {
		log(LOG_WARNING, errno, "sendmsg on %s to %s",
		    inet_fmt(src, s1),
		    inet_fmt(dst, s2));
	}

	tracef(TRF_PIM2, "TX PIM: sending %s from %-15s to %s",
	       pim_packet_kind(type), inet_fmt(src, s1), inet_fmt(dst, s2));
}
