#include "defs.h"


int pim2_decode_group(char **pp, void *res, u_char *masklen)
{
	__u8 *p = *pp;

	if (p[0] != 1) {
		log(LOG_DEBUG, 0, "pim2_decode_group: cannot handle af=%d", p[0]);
		return FALSE;
	}
	if (p[1] != 0) {
		log(LOG_DEBUG, 0, "pim2_decode_group: cannot handle not native IP encoding %d", p[1]);
		return FALSE;
	}
	if (masklen == NULL) {
		if (p[3] != 32) {
			log(LOG_DEBUG, 0, "pim2_decode_group: cannot handle not native IP masklen %d", p[3]);
			return FALSE;
		}
	} else {
		if (p[3] > 32) {
			log(LOG_DEBUG, 0, "pim2_decode_group: cannot handle IP masklen %d", p[3]);
			return FALSE;
		}
		*masklen = p[3];
	}
	memcpy(&res, p+4, 4);
	*pp += 8;
	return TRUE;
}

int pim2_decode_unicast(char **pp, void *res)
{
	__u8 *p = *pp;

	if (p[0] != 1) {
		log(LOG_DEBUG, 0, "pim2_decode_unicast: cannot handle af=%d", p[0]);
		return FALSE;
	}
	if (p[1] != 0) {
		log(LOG_DEBUG, 0, "pim2_decode_unicast: cannot handle not native IP encoding %d", p[1]);
		return FALSE;
	}
	memcpy(&res, p+2, 4);
	*pp += 6;
	return TRUE;
}


int pim2_decode_source(char **pp, void *res, u_short *flags_p)
{
	__u8 *p = *pp;

	if (p[0] != 1) {
		log(LOG_DEBUG, 0, "pim2_decode_source: cannot handle af=%d", p[0]);
		return FALSE;
	}
	if (p[1] != 0) {
		log(LOG_DEBUG, 0, "pim2_decode_source: cannot handle not native IP encoding %d", p[1]);
		return FALSE;
	}
	if (p[3] != 32) {
		log(LOG_DEBUG, 0, "pim2_decode_source: cannot handle not native IP masklen %d", p[3]);
		return FALSE;
	}

	*flags_p = (((((p[2]) & (1 << 1)) ? MRTF_WC : 0) |
		     ((p[2] & 1) ? MRTF_RP : 0))
		    & (0x00ff));
	memcpy(&res, p+4, 4);
	*pp += 8;
	return TRUE;
}


/* function name: encode_group
   input: buffer address, the group address, and the group mask length
   output: the number of bytes taken for the encoded group
   operation: constructs an encoded group address for the given group
   using the maskLen, in the buffer pointer referred to by pktAdd,
   the buffer pointer is advanced automatically, and the total
   datalen is returned.[check spec for format]
   comments: called by send_C_RP_Adv, periodic_join_prune
 */
int pim2_encode_group(char **pktAdd, __u32 group, u_char maskLen)
{
	__u8 *p = *pktAdd;

	p[0] = 1;
	p[1] = 0;
	p[2] = 0;
	p[3] = maskLen;
	memcpy(p+4, &group, 4);
	*pktAdd = p;
	return 8;
}

int pim2_encode_unicast(char **pktAdd, __u32 addr)
{
	__u8 *p = *pktAdd;

	p[0] = 1;
	p[1] = 0;
	memcpy(p+2, &addr, 4);
	*pktAdd = p;
	return 6;
}


/* function name: encode_source
   input: buffer address, mrt entry pointer, and the source mask length
   output: the number of bytes taken for the encoded source
   operation: constructs an encoded source address for the given group
   using the rp_bit and the maskLen, in the buffer pointer referred 
   to by pktAdd, the buffer pointer is advanced automatically, and 
   the total datalen is returned.[check spec for format]
   comments: called by even_join_prune, and periodic_join_prune
 */
int pim2_encode_source(char **pktAdd, struct mrtentry *mrtPtr, u_int rp_bit, u_char maskLen)
{
	__u8 *p = *pktAdd;
	u_int flags;

	flags = p[2] =
		((1 << 2) |	/* The S bit is always set */
		 (((mrtPtr->flags & MRTF_WC) ? 1 : 0) << 1) |	/* The WC bit */
		 ((((mrtPtr->flags & MRTF_RP) | rp_bit) ? 1 : 0)));
	/* The RP bit */

	p[0] = 1;
	p[1] = 0;
	p[3] = maskLen;

	memcpy(p+4, &(mrtPtr->source->source), 4);
	return 8;
}

void query_neighbors_v2(struct vifentry *v)
{
	char *p;
	__u16 *ptr;
	struct vifaddr *vifa;

	p = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
	ptr = (__u16*)p;
	ptr[0] = htons(1);
	ptr[1] = htons(2);
	ptr[2] = htons(v->query_holdtime);
	v->timer = 0;
	for (vifa = v->ifa; vifa; vifa = vifa->next) {
		if (!vifa->primary)
			continue;
		send_pim(pim_send_buf, vifa->local, allpimrouters_group,
			 PIM_HELLO, 6, v->ifindex, 1);
	}
}


/* unicast register stop. Optional rate limiting is not implemented yet. */
static void send_reg_stop_v2(__u32 source, __u32 group, __u32 regsrc,
			     __u32 regdst, __u32 active_rp)
{
	char *buf;
	u_long datalen = 0;

	buf = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
	datalen += pim2_encode_group(&buf, group, GRP_MSKLEN);
	bcopy((char *) &source, buf, ADDR_LEN);
	datalen += ADDR_LEN;
	send_pim(pim_send_buf, regsrc, regdst, PIM_REGISTER_STOP, datalen, 0, MAXTTL);
}


int accept_register_v2(__u32 src, __u32 dst, char *pkt, int pimlen)
{
	return accept_register_gen(src, dst, pkt+4, pkt[0]&0x80, pkt[0]&0x40, send_reg_stop_v2);
}


/* function name: prune_rpt
   input: mrt entry pointer
   output: int (-1 if nbr is not PIM)
   operation:
   comments: called by process_wrongiif
 */
int prune_rpt_v2(struct vifentry *v, struct nbrentry *n, struct mrtentry *mrtPtr)
{
	char *buf, *bufPtr;
	int datalen;
	u_long num_joins, num_prunes;
	char join_buf[ADDRSIZE * MAXSrc], prune_buf[ADDRSIZE * MAXSrc],	*jptr, *pptr;
	u_char num_grps;
	u_char rp_count = 0;

	buf = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
	num_joins = 0;
	num_prunes = 1;
	num_grps = 1;

	bcopy(&(n->address), buf, 4);	/* The upstream neighbor Address */
	bzero(buf + 4, 1);	/* Reserved */
	buf[5] = num_grps;
	bcopy(jpholdtime, buf + 6, 2);
	datalen = 8;		/* Two quads just filled */

	bufPtr = &buf[datalen];
	datalen += pim2_encode_group(&bufPtr, mrtPtr->group->group, GRP_MSKLEN);

/* Copy the LS two bytes */
	buf[datalen++] = (char) (num_joins >> 8);
	buf[datalen++] = (char) num_joins;
/* Copy the LS two bytes */
	buf[datalen++] = (char) (num_prunes >> 8);
	buf[datalen++] = (char) num_prunes;
	buf[datalen++] = rp_count;	/* 0, as prune */

	jptr = join_buf;
	pptr = prune_buf;
	/* This is an RP bit set prune, WC,A, bit is not set */
	*pptr++ = ((1 << 0) | (1 << 2));
	/* Assuming that mask length is 32 */
	*pptr++ = 32;
	bcopy((char *) &(mrtPtr->source->source), pptr, 4);	/* Copy 4 byte IP address */
	bcopy(join_buf, buf + datalen, num_joins * ADDRSIZE);
	bcopy(prune_buf, buf + datalen + num_joins * ADDRSIZE, num_prunes * ADDRSIZE);

	datalen += (num_joins + num_prunes) * ADDRSIZE + rp_count * 4;
	send_pim(pim_send_buf, v->prefsrc, allpimrouters_group,
		 PIM_JOIN_PRUNE, datalen, v->ifindex, 1);
	return (TRUE);
}

/* function name: send_null_register
   input: mrtentry pointer, the entry on which behalf the null register is to 
   be sent.
   output: none
   operation: creates a Register message with the null bit set and a dummy ip 
   header with the src and grp for the mrt included therein, then
   unicasts it to the RP for the corresponding group.
   comments: called by ageRegisterBitTimer
 */
void send_null_register_v2(struct mrtentry *r)
{
	register struct ip *ip;
	char *buf;
	int pktlen = 0;
	__u32 dest;
	__u32 reg_head = __constant_htonl(0x40000000);

	buf = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;

	bcopy((char *) &reg_head, buf, 4);		/* set the null reg bit */
	pktlen += 4;
	buf += 4;

	ip = (struct ip *) (buf);
	ip->ip_src.s_addr = r->source->source;	/* set src/dst in dummy hdr */
	ip->ip_dst.s_addr = r->group->group;

	pktlen += MIN_IP_HEADER_LEN;	/* include the dummy ip header */

	dest = r->group->active_rp->rpaddr;

	send_pim(pim_send_buf, 0, dest, PIM_REGISTER, pktlen, 0, MAXTTL);
	return;
}

/* function name: accept_assert
   input: source, destinatoin, pkt pointer, and data length
   output: int (TRUE if this router won the assert [if this is 
   a downstream router], or changed the rpf nbr [if this 
   is an upstream router]. FALSE o.w.)
   operation: checks to see if the assert was received on oif,
   if so, then I am an upstream router, and perform
   a longest match. O.w., I am a downstream router,
   and perform an exact match.
   comments: called by accept_igmp [igmp.c]
   calls find_route, longest_match, and assert
 */
int accept_assert_v2(__u32 src, __u32 dst, char *p, int datalen)
{
	vifi_t vifi;
	struct nbrentry *n;
	__u32 group, source, metric, pref;

	if ((vifi = vifi_by_index(pim_ifindex)) == MAXVIFS)
		return FALSE;
	n = neigh_lookup(vifi, src, 1);
	if (n == NULL)
		return FALSE;

	if (pim2_decode_group(&p, &group, NULL) == FALSE ||
	    pim2_decode_unicast(&p, &source) == FALSE)
		return FALSE;

	bcopy(p, (char *) &pref, 4);
	bcopy(p + 4, (char *) &metric, 4);

	return accept_assert_gen(vifi, n, source, group, ntohl(pref), ntohl(metric));
}


/* function name: assert
   input: mrtentry pointer, and vifentry pointer 
   output: none
   operation: sends an Assert message for the given mrt, on the given
   vif.
   comments: called by trigger_assert
   calls send_pim and encode_group
 */
void send_assert_v2(struct mrtentry *mrtPtr, struct vifentry *v)
{
	char *p, *pktPtr;
	int datalen;
	__u32 pref = htonl(mrtPtr->source->preference);
	__u32 metric = htonl(mrtPtr->source->metric);

	pktPtr = p = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
	pim2_encode_group(&pktPtr, mrtPtr->group->group, GRP_MSKLEN);
	pim2_encode_unicast(&pktPtr, mrtPtr->source->source);
	memcpy(pktPtr, &pref, 4);
	/* If its RPbit entry set the first bit */
	*pktPtr |= ((mrtPtr->flags & MRTF_RP) ? 1 << 7 : 0);
	pktPtr += 4;
	memcpy(pktPtr, &metric, 4);
	pktPtr += 4;
	datalen = pktPtr - p;
	send_pim(pim_send_buf, v->prefsrc, allpimrouters_group,
		 PIM_ASSERT, datalen, v->ifindex, 1);
}


/* function name: accept_join_prune
   input: src and dst addresses, pkt pointer and data length
   output: none
   operation: processes incoming Join/Prune messages. Currently
   only messages destined to local address are processed 
   (i.e. the optional Join/Prune suppression is not implemented yet).
   The message is parsed, and interfaces are added/deleted accordingly.
   comments: called by accept_pim
   calls add_outgo, del_outgo, find_route, loacl_address, 
   find_vif, 
 */
void accept_join_prune_v2(__u32 src, __u32 dst, char *p, int datalen)
{
	u_short flags;
	u_long source, group;
	vifi_t vifi;
	struct vifentry *v;
	__u32 target;
	int len, num_joins, num_prunes;
	int i, j;
	u_char num_grps;
	struct nbrentry *n;

	if ((vifi = vifi_by_index(pim_ifindex)) == MAXVIFS ||
	    (n = neigh_lookup(vifi, src, 2)) == NULL) {
		log(LOG_DEBUG, 0, "Not directly cnctd, ignoring (%s)!\n", inet_fmt(src, s1));
		return;		/* if not directly cnctd */
	}

	if (debug > 3)
		log(LOG_DEBUG, 0, "\n accept join prune src %s, dst %s, on vifi %d",
		    inet_fmt(src, s1), inet_fmt(dst, s2), vifi);

	v = &viflist[vifi];

	if (pim2_decode_unicast(&p, &target) == FALSE)
		return;

	/* if I am not the targer of the join message */
	if (!is_local_on(target, v))
		/* May add code for joiner bit and LAN prune suppression, 
		   by sending join */
		return;

	/* I am the target of this join, so process the message */
	num_grps = p[1];
	/* copy the holdtime in the join, into the oif timeout of the
	 * corresponding vif
	 */
	bcopy((char *) &p[2], (char *) &oifTimeout[vifi], 2);
	p += 4;
	len = 0;

	for (i = 0; i < num_grps; i++) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "Num of Groups = %u", num_grps);

		if (pim2_decode_group(&p, &group, NULL) == FALSE)
			return;
		num_joins = (int) (p[1]) + (int) (p[0]) * 256;
		num_prunes = (int) (p[3]) + (int) (p[2]) * 256;
		p += 4;

		for (j = 0; j < num_joins; j++) {
			if (pim2_decode_source(&p, &source, &flags) == FALSE)
				return;

			do_accept_join(vifi, source, group, flags);
		}

		for (j = 0; j < num_prunes; j++) {
			if (debug > 3) {
				log(LOG_DEBUG, 0, "Num of Prunes = %u", num_prunes);
				if (num_joins > 5)
					return;
			}

			if (pim2_decode_source(&p, &source, &flags) == FALSE)
				return;

			do_accept_prune(vifi, source, group, flags);
		}
	}
}

/* function name: event_join_prune
   input: mrt entry pointer, and a character (join flag)
   output: int (FALSE if the join is not sent, o.w. TRUE)
   operation: Send join message triggered by some new mrt etc 
   comments: called by add_leaf, add_outgo, 
 */
int event_join_prune_v2(struct mrtentry *r, char join,
			struct nbrentry *n, struct grpentry *g,
			struct vifentry *v)
{
	char *buf, *bufPtr;
	int datalen;
	u_long num_joins, num_prunes;
	char join_buf[ADDRSIZE * MAXSrc], prune_buf[ADDRSIZE * MAXSrc];
	char *jptr, *pptr;
	u_char num_grps;

	buf = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
	bcopy(&(n->address), buf, 4);	/* The upstream neighbor Address */

	bzero(buf + 4, 1);	/* Reserved */
	bcopy(jpholdtime, buf + 6, 2);
	jptr = join_buf;
	pptr = prune_buf;
	datalen = 8;		/* Three quads just filled */
	num_joins = 0;
	num_prunes = 0;
	num_grps = 0;
	if ((r->flags & MRTF_RP) && !(r->flags & MRTF_WC) && (!join))
		/* RPbit Entry */
		pim2_encode_source(&pptr, r, TRUE, SRC_MSKLEN);
	else {
		if ((((r->outgoing | g->leaves) == EMPTY_VIF) && (!join))) {
			num_prunes++;
			pim2_encode_source(&pptr, r, FALSE, SRC_MSKLEN);
			if (debug > 4)
				log(LOG_DEBUG, 0, "pruning Src %s, Grp %s, entry flags %x",
				    inet_fmt(r->source->source, s1),
				    inet_fmt(r->group->group, s2),
				    r->flags);
		} else {	/* Right now it doesn't take care of Jbit */
			num_joins++;

			pim2_encode_source(&jptr, r, FALSE, SRC_MSKLEN);

			if (debug > 4)
				log(LOG_DEBUG, 0, "joining Src %s, Grp %s, entry flags %x",
				    inet_fmt(r->source->source, s1),
				    inet_fmt(r->group->group, s2),
				    r->flags);
		}
	}

	if (debug > 3)
		log(LOG_DEBUG, 0, "Num Joins %u, Num Prunes %u",
		    num_joins, num_prunes);

	/* if the total number is zero, don't send */
	if ((num_joins + num_prunes) == 0)
		return (FALSE);
	num_grps++;
	bufPtr = &buf[datalen];

	datalen += pim2_encode_group(&bufPtr, r->group->group, GRP_MSKLEN);

	/* Copy the LS two bytes */
	buf[datalen++] = (char) (num_joins >> 8);
	buf[datalen++] = (char) num_joins;
	/* Copy the LS two bytes */
	buf[datalen++] = (char) (num_prunes >> 8);
	buf[datalen++] = (char) num_prunes;

	bcopy(join_buf, buf + datalen, num_joins * ADDRSIZE);
	bcopy(prune_buf, buf + datalen + num_joins * ADDRSIZE, num_prunes * ADDRSIZE);
	datalen += (num_joins + num_prunes) * ADDRSIZE;

	buf[5] = num_grps;

	if (debug > 4)
		log(LOG_DEBUG, 0, "\nSending triggered Join/Prune to all pim routers");

	send_pim(pim_send_buf, v->prefsrc, allpimrouters_group, PIM_JOIN_PRUNE,
		 datalen, v->ifindex, 1);
	return (TRUE);
}

/* function name: periodic_join_prune
   input: a neighbor entry pointer, and a virtual interface
   output: int (the number of groups for which a Join/Prune 
   message has been sent)
   operation: Sends the periodic join/prune message to the 
   neighbor n, and returns number of groups for which 
   message has been sent.
   comments: called by periodic_vif_join_prune
   calls join_or_prune
 */
int periodic_join_prune_v2(struct nbrentry *n, vifi_t vifi)
{
	u_long num_joins, num_prunes;
	int datalen;
	__u32 group;
	register struct mrtentry *r, *r2;
	struct grpentry *g;
	char *jptr, *pptr;
	char join_buf[ADDRSIZE * MAXSrc], prune_buf[ADDRSIZE * MAXSrc];
	char *buf, *bufPtr;
	u_char num_grps;
	int j_or_p;

	num_grps = 0;
	buf = bufPtr = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;

	if (n != NULL) {
		pim2_encode_unicast(&bufPtr, n->address);
	} else
		/* if the upstream neighbor is not a PIM neighbor, don't
		   send any messages... return an error code (-1) */
		return (-1);

	bufPtr[0] = 0;
	bcopy(jpholdtime, bufPtr + 2, 2);
	bufPtr += 4;
	datalen = bufPtr - buf;

	forall_groups(g) {
		num_joins = 0;
		num_prunes = 0;
		group = g->group;
		r = g->mrtlink;
		jptr = join_buf;
		pptr = prune_buf;

		for (r2 = r; r2 != NULL; r2 = r2->grpnext) {
			/* Check if its time to send the periodic join_prune */

			if ((r2->join_prune_timer += TIMER_INTERVAL) < 15 /* JOIN_PRUNE_INTERVAL */ )
				continue;
			r2->join_prune_timer = 0;	/* Reset the timer */

			/* one of my address */
			if (local_address(r2->source->source) != MAXVIFS)
				continue;

			j_or_p = join_or_prune(n, vifi, r2);

			if (j_or_p == NONE)
				continue;

			if (j_or_p == PRUNE) {
				tracef(TRF_PIM2, "\nsending Prunes for (%s,%s)",
				       inet_fmt(r2->source->source, s1),
				       inet_fmt(r2->group->group, s2));
				num_prunes++;
				pim2_encode_source(&pptr, r2, TRUE, SRC_MSKLEN);
			} else {
				tracef(TRF_PIM2, "\nsending Join for (%s,%s)",
				       inet_fmt(r2->source->source, s1),
				       inet_fmt(r2->group->group, s2));

				num_joins++;
				pim2_encode_source(&jptr, r2, FALSE, SRC_MSKLEN);
			}
		}
		if ((num_joins + num_prunes) == 0)
			continue;
		num_grps++;

		bufPtr = &buf[datalen];
		datalen += pim2_encode_group(&bufPtr, group, GRP_MSKLEN);

		/* Copy the LS two bytes */
		buf[datalen++] = (char) (num_joins >> 8);
		buf[datalen++] = (char) num_joins;
		/* Copy the LS two bytes */
		buf[datalen++] = (char) (num_prunes >> 8);
		buf[datalen++] = (char) num_prunes;
		bcopy(join_buf, buf + datalen, num_joins * ADDRSIZE);
		bcopy(prune_buf, buf + datalen + num_joins * ADDRSIZE, num_prunes * ADDRSIZE);
		datalen += (num_joins + num_prunes) * ADDRSIZE;
		continue;
	} endforall_groups(g);
	if (num_grps == 0)
		return num_grps;
	buf[5] = num_grps;
	send_pim(pim_send_buf, viflist[vifi].prefsrc, allpimrouters_group,
		 PIM_JOIN_PRUNE, datalen, viflist[n->vif].ifindex, 1);
	return num_grps;
}

void send_register_v2(__u32 dest, struct ip *ip, int pktlen)
{
	char *buf;

	buf = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;

	bzero(buf, 4);		/* reset the border bit and the reserved field */
	buf += 4;
	pktlen += 4;

	/* Copy the data packet at the back of the register packet */
	pktlen += ntohs(ip->ip_len);
	bcopy((char *) ip, buf, pktlen);
	send_pim(pim_send_buf, 0, dest, PIM_REGISTER, pktlen, 0, MAXTTL);
}

