/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: route.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */

#include "defs.h"

u_short oifTimeout[MAXVIFS];

static int event_join_prune(struct mrtentry *r, char join);

static vifbitmap_t computeOifs(struct mrtentry *mrtPtr);


/*************** Longest Matches and other processing ********************/

struct mrtentry * longest_match_r(__u32 source, __u32 group)
{
	struct grpentry *g;
	struct mrtentry *r;

	if ((r = find_route_r(source, group, EMPTY_FLAG, FALSE)) != NULL)
		return r;

	if (!(g = find_group_r(group, FALSE)))
		return NULL;

	return g->rp_entry;
}

/* Checks if src is local address. If yes returns the vif index else MAXVIFS */
vifi_t local_address(__u32 src)
{
	int vifi;
	register struct vifentry *v;

	/* 
	 * Reversed lookup, to return ethers first  and not tunnels
	 */
	for (vifi = numvifs - 1, v = &viflist[vifi]; vifi >= 0; --vifi, --v) {
		struct vifaddr *vifa;
		if (v->flags & (VIFF_DOWN | VIFF_DISABLED))
			continue;
		for (vifa = v->ifa; vifa; vifa = vifa->next) {
			if (src == vifa->local)
				return vifi;
		}
	}
	/* Returning Maxvifs  means not a local address */
	return (MAXVIFS);
}

int is_local_on(__u32 src, struct vifentry *v)
{
	struct vifaddr *vifa;

	for (vifa = v->ifa; vifa; vifa = vifa->next) {
		if (src == vifa->local)
			return TRUE;
	}
	return FALSE;
}


/************ Leaves and Outgoing Interface triggered manipulations *******/

/* function name: update_kernel_leaves
   input: vifbitmap of old leaves, and a pointer to the group
   output: int (TRUE if kernel caches were found and updated, and
   FALSE o.w.).
   operation: browses the group mrtlink looking for kernel cache mirrors,
   when/if found any, checks the old (before adding the leaf)
   and new oifs (after adding the leaf), if there was a 
   change then the kernel is updated accordingly.
   comments: called by add_leaf
 */
int update_kernel_leaves(vifbitmap_t oldLeaves, struct grpentry *grpPtr)
{
	vifbitmap_t oldOif, newOif, tempOif;
	struct mrtentry *mrtptr, *rteptr;
	int updated = FALSE;
	if (debug > 4)
		log(LOG_DEBUG, 0, "update_kernel_leaves: G:%s",
		    inet_fmt(grpPtr->group, s1));

	if (!(rteptr = grpPtr->rp_entry)) {	/* if (*,G) does not exist */
		if (debug > 4)
			log(LOG_DEBUG, 0, "update_kernel_leaves: (*,G) does not exist!");
		return (FALSE);
	}
	rteptr->outgoing = rteptr->group->leaves;
	/* simplest is to browse thru the group list and get the entries */
	for (mrtptr = grpPtr->mrtlink; mrtptr; mrtptr = mrtptr->grpnext) {
		oldOif = 0;
		newOif = 0;
		tempOif = 0;

		/* oifs = oif(*,G) - pruned oifs + valid oif(S,G) */
		tempOif = (rteptr->outgoing & ~(mrtptr->vifmaskoff)) & ~nomrt_vif;
		tempOif |= (mrtptr->outgoing & ~nomrt_vif);

		oldOif = tempOif | oldLeaves;
		newOif = mrtptr->group->leaves | tempOif;

		if (oldOif != newOif) {
			updated = TRUE;
			/* tell the kernel. */
			if (newOif)
				mrtptr->flags &= ~MRTF_NULL_OIF;
			if (mrtptr->flags & MRTF_KERNEL_CACHE)
				k_chg_mrtf(mrtptr);
			else
				mrtptr->outgoing = newOif;
			dirty_bits |= 2;
		}		/* if changed */
	}
	return (updated);
}

/* function name: add_leaf
   input: vifi (the virtual interface on which the the group report
   was received), and the group address
   output: none.
   operation: find/create the group for which to add the leaf, then
   find/create the corresponding (*,G); if the entry
   is created, a join/prune is triggered for (*,G),
   and any S,G having previously null oifs. If the leaves have
   changed, call update_kernel_leaves.
   comments: called by accept_group_report [vif.c], 
   calls update_kernel_leaves, find_group, find_route
   FIX:      See if G is RP based or not, if not use Dense Mode
 */
void add_leaf(vifi_t vifi, __u32 group)
{
	struct mrtentry *mrtptr;
	struct grpentry *g;
	struct rplist *rp_ptr1;
	vifbitmap_t oldLeaves = 0; /* fake gcc */
	u_int changed_leaves;

	if (debug > 2)
		log(LOG_DEBUG, 0, "add_leaf:Adding leaf %d for group %s",
		    vifi, inet_fmt(group, s1));

	if (!(g = find_group_r(group, TRUE)))
		return;

	if (g->rp_entry != NULL) {
		/* there exists (*,G) entry */
		if (debug > 2)
			log(LOG_DEBUG, 0, "add_leaf:(*,%s) already exists!\n", inet_fmt(group, s1));
		mrtptr = g->rp_entry;
	} else {		/* this is the first leaf (i.e. (*,G) does not exist) */

		if (!(rp_ptr1 = g->active_rp))
			return;
		/* create (*,G) entry */
		if (!(mrtptr = find_route_r(rp_ptr1->rpaddr, group, MRTF_WC | MRTF_RP, TRUE))) {
			if (debug > 2)
				log(LOG_DEBUG, 0, "add_leaf:failed to create (*,G) entry");
			return;
		}
		if (debug > 2)
			log(LOG_DEBUG, 0, "add_leaf:(*,%s) created RP:%s !\n",
			    inet_fmt(group, s1),
			    inet_fmt(mrtptr->source->source, s2));
	}			/* if this is the first leaf */

	if (debug > 2)
		log(LOG_DEBUG, 0, "add_leaf:(*,%s) RP: %s  in: %d, outgoing: %08x inc vif: %d",
		    inet_fmt(mrtptr->group->group, s1),
		    inet_fmt(mrtptr->source->source, s2),
		    mrtptr->source->incoming,
		    mrtptr->outgoing,
		    vifi);


	if ((mrtptr->flags & MRTF_CRT) ||
	    ((mrtptr->outgoing | g->leaves) == EMPTY_VIF)) {
		event_join_prune(mrtptr, TRUE);
		mrtptr->flags &= ~MRTF_CRT;
		dirty_bits |= 2;
	}
	/* Send a triggered join for each S,G entry, if this is first leaf */
	if (g->leaves == EMPTY_VIF) {
		struct mrtentry *r1;
		for (r1 = g->mrtlink; r1; r1 = r1->grpnext) {
			/* Don't send join if its already being sent */
			if ((r1->outgoing != EMPTY_VIF) || (r1 == mrtptr))
				continue;
			if (r1->flags & MRTF_SG) {	/* only send joins for pure S,G */
				event_join_prune(r1, TRUE);
			}
		}
	}
	/* keep track of old leaves and update kernel if changed */
	changed_leaves = 0;
	/* Add the leaf */
	g->timers[vifi] = 0;
	if (debug > 2)
		log(LOG_DEBUG, 0, "add_leaf:check %d on outgo for (*,%s)",
		    vifi,
		    inet_fmt(g->group, s1));

	if (!VIFM_ISSET(vifi, g->leaves)) {
		oldLeaves = g->leaves;
		VIFM_SET(vifi, g->leaves);
		changed_leaves = 1;
	}
	if (changed_leaves) {
		if (debug > 2)
			log(LOG_DEBUG, 0, "add_leaf: %d on outgo for (*,%s)",
			    vifi,
			    inet_fmt(g->group, s1));

		if (mrtptr->flags & MRTF_WC)
			update_kernel_leaves(oldLeaves, g);
		dirty_bits |= 2;
	}
	return;
}

/* function name: update_kernel_oifs
   input: mrtentry pointer, old outgoing map, and old vif mask off map.
   output: int (TRUE if found and updated kernel cache(s), o.w. FALSE).
   operation: if the passed mrtentry pointer is a wild card entry, then
   update all the pertinent (S,G) kernel caches after checking
   the change in the oiflist. If the entry is not a wild card,
   simply check the oiflist and update the kernel, if the entry
   is a cache mirror.
   comments: called by add_outgo
 */
int update_kernel_oifs(struct mrtentry *mrtPtr, vifbitmap_t oldOutgoing,
		       vifbitmap_t oldVifmaskoff)
{
	int updated = 0;
	struct mrtentry *mrtptr, *rteptr;
	vifbitmap_t tempOif = 0, oldOif = 0, newOif = 0;

	rteptr = mrtPtr->group->rp_entry;

	if (debug > 3)
		log(LOG_DEBUG, 0, "update kernel oifs, oldoif %x, oldmsk %x",
		    oldOutgoing, oldVifmaskoff);

	if ((mrtPtr->flags & MRTF_RP) && (mrtPtr->flags & MRTF_WC)) {
		if (!rteptr)	/* if there is no rp entry */
			return (FALSE);

		/* browse thru the corresponding (S,G) caches */
		for (mrtptr = mrtPtr->group->mrtlink; mrtptr != NULL;
		     mrtptr = mrtptr->grpnext) {
			oldOif = 0;
			newOif = 0;
			tempOif = 0;

			if (mrtptr->flags & MRTF_KERNEL_CACHE) {
				/* check if there is a change in the oifs */
				tempOif = (mrtptr->outgoing & ~nomrt_vif);
				tempOif |= mrtptr->group->leaves;

				oldOif = tempOif |
					((oldOutgoing & ~(mrtptr->vifmaskoff)) & ~nomrt_vif);
				newOif = tempOif |
					((rteptr->outgoing & ~(mrtptr->vifmaskoff)) & ~nomrt_vif);
				if (oldOif != newOif) {
					updated = TRUE;
					if (mrtptr->flags & MRTF_NULL_OIF)
						mrtptr->flags &= ~MRTF_NULL_OIF;
					/* tell the kernel.. */
					k_chg_mrtf(mrtptr);
					dirty_bits |= 2;
				}	/* if changed */
			}
		}
	} else if (mrtPtr->flags & MRTF_KERNEL_CACHE) {
		/* not a wild card entry, so check if kernel cache mirror */
		/* check if change and update kernel */
		tempOif = mrtPtr->group->leaves;
		if (rteptr) {
			oldOif = tempOif |
				((rteptr->outgoing & ~oldVifmaskoff) & ~nomrt_vif);
			newOif = tempOif |
				((rteptr->outgoing & ~(mrtPtr->vifmaskoff)) & ~nomrt_vif);
		}
		oldOif |= (oldOutgoing & ~nomrt_vif);
		newOif |= (mrtPtr->outgoing & ~nomrt_vif);
		if (oldOif != newOif) {
			updated = TRUE;
			if (debug > 4)
				log(LOG_DEBUG, 0, "updating oifs in kernel");

			if (mrtPtr->flags & MRTF_NULL_OIF)
				mrtPtr->flags &= ~MRTF_NULL_OIF;
			/* tell the kernel */
			k_chg_mrtf(mrtPtr);
			dirty_bits |= 2;
		}		/* if changed */
	}			/* if kernel cache mirror */
	return (updated);
}

/* function name: add_outgo
   input: vifi and mrtentry pointer
   output: none
   operation: if the entry passed was newly created, or if the 
   oiflist has changed from null to non-null, trigger a join/prune
   message. Reset the oif timer, and update the oiflist,
   if a change occured, call update_kernel_oifs.
   comments: called by accept_join_prune
   calls update_kernel_oifs
 */
void add_outgo(vifi_t vifi, struct mrtentry *r)
{
	u_int change_vif = 0;	/* change flag */
	vifbitmap_t oldOutgoing = 0, oldVifmaskoff = 0;
	if (debug > 2)
		log(LOG_DEBUG, 0, "add_outgo: vifi=%d, r->outgoing=%08x ",
		    vifi,
		    r->outgoing);
	if ((r->flags & MRTF_CRT) || ((r->outgoing | r->group->leaves)
				      == EMPTY_VIF)) {
		event_join_prune(r, TRUE);
		dirty_bits |= 2;
		r->flags &= ~MRTF_CRT;
	}
	((u_char *) (r->timers))[vifi] = 0;	/* Reset the timers */
	if (!VIFM_ISSET(vifi, r->outgoing)) {
		oldOutgoing = r->outgoing;
		VIFM_SET(vifi, r->outgoing);
		change_vif = 1;
	}
	if (VIFM_ISSET(vifi, r->vifmaskoff)) {
		oldVifmaskoff = r->vifmaskoff;
		VIFM_CLR(vifi, r->vifmaskoff);
		change_vif = 1;
	}
	if (change_vif) {
		update_kernel_oifs(r, oldOutgoing, oldVifmaskoff);
		dirty_bits |= 2;
	}
}

/* function name: del_outgo
   input: vifi and mrtentry pointer
   output: none
   operation: if the vif deleted is set in the oiflist, it is cleared,
   else, it is added to the vifmaskoff. If a change
   occured, update_kerenl_oifs is called.
   coments: called by accept_join_prune and accept_assert
 */
static void del_outgo(vifi_t vifi, struct mrtentry *r)
{
	vifbitmap_t oldOutgoing = 0, oldVifmaskoff = 0;
	u_int change_vif = 0;	/* change flag */
	if (debug > 2)
		log(LOG_DEBUG, 0, "del_outgo: Clears %d on (%s,%s)",
		    vifi,
		    inet_fmt(r->source->source, s1),
		    inet_fmt(r->group->group, s2));
	if (VIFM_ISSET(vifi, r->outgoing)) {
		oldOutgoing = r->outgoing;
		VIFM_CLR(vifi, r->outgoing);
		change_vif = 1;
	} else if (r->group->rp_entry != NULL) {
		if (VIFM_ISSET(vifi, r->group->rp_entry->outgoing)) {
			oldVifmaskoff = r->vifmaskoff;
			VIFM_SET(vifi, r->vifmaskoff);
			change_vif = 1;
		}
	}
	if (change_vif) {
		update_kernel_oifs(r, oldOutgoing, oldVifmaskoff);
		dirty_bits |= 2;
	}
}

static int event_join_prune(struct mrtentry *r, char join)
{
	struct srcentry *s;
	struct grpentry *g;
	struct vifentry *v;
	struct nbrentry *n;
	__u32 group;

	s = r->source;

	if (local_address(s->source) != MAXVIFS) {
		/* No need to send join as the source is local */
		if (debug > 3)
			log(LOG_DEBUG, 0, "event_join_prune: Source is local");
		return (FALSE);
	}
	g = r->group;
	group = g->group;

	if (r->flags & MRTF_RP) {
		n = g->rp_entry->source->upstream;
		if (debug > 3) {
			log(LOG_DEBUG, 0, "event_join_prune:source is %s,RP entry nbr is %s, on vif[%d]",
			    inet_fmt(g->rp_entry->source->source, s1),
			    n ? inet_fmt(n->address, s2) : "0.0.0.0",
			    g->rp_entry->source->incoming);
		}
		v = &viflist[g->rp_entry->source->incoming];
	} else if (r->flags & MRTF_ASSERTED) {
		n = r->upstream;
		if (n)
			v = &viflist[n->vif];
	} else {
		n = s->upstream;
		v = &viflist[s->incoming];
	}

	if (n == NULL) {	/* if the neighbor is not PIM don't send Join/Prune messages */
		if (debug > 3)
			log(LOG_DEBUG, 0, "event_join_prune: non-PIM next hop..fail");
		return (FALSE);
	}

	if (!v->v1neighnum || v->v2neighnum)
		event_join_prune_v2(r, join, n, g, v);
	if (v->v1neighnum)
		event_join_prune_v1(r, join, n, g, v);

	return (TRUE);
}


void do_accept_join(vifi_t vifi, __u32 source, __u32 group, u_short flags)
{
	struct mrtentry *r;

	if (!(flags & MRTF_WC))		/* not wild card */
		flags |= MRTF_SG;	/* set the (S,G) bit */

	if (debug > 3)
		log(LOG_DEBUG, 0, "Joining (%s,%s), flags %x",
		    inet_fmt(source, s1), inet_fmt(group, s2), flags);

	if ((r = find_route_r(source, group, flags, TRUE)) != NULL) {
		if (r->source->incoming != vifi ||
		    local_address(source) != MAXVIFS) {
			add_outgo(vifi, r);
			r->flags |= flags;
		}
		if ((r->flags&(MRTF_WC|MRTF_RP|MRTF_IIF_REGISTER))
		    == (MRTF_WC|MRTF_RP|MRTF_IIF_REGISTER) &&
		    local_address(r->group->active_rp->rpaddr) != MAXVIFS) {
			if (r->rpreachtimer == 0) {
				if (debug)
					log(LOG_DEBUG, 0, "Start RPReach timer (%s,%s)",
					    inet_fmt(source, s1), inet_fmt(group, s2));
				r->rpreachtimer = 1;
			}
		}
	}
}

void do_accept_prune(vifi_t vifi, __u32 source, __u32 group, u_short flags)
{
	struct mrtentry *r;

	if (debug > 3)
		log(LOG_DEBUG, 0, "Pruning (%s,%s), flags %x",
		    inet_fmt(source, s1), inet_fmt(group, s2), flags);

	if ((r = find_route_r(source, group, flags, ((flags & MRTF_RP) ?
						     TRUE : FALSE))) != NULL) {
		if (flags & MRTF_RP) {
			if (r->group->rp_entry)
				/* if I am not the RP */
				if (local_address(r->group->rp_entry->source->source) == MAXVIFS) {
					if (r->group->rp_entry->source->incoming != vifi)
						del_outgo(vifi, r);
				} else
					del_outgo(vifi, r);
		} else if (r->source->incoming != vifi)
			del_outgo(vifi, r);
	}
}


static void send_assert(struct mrtentry *mrtPtr, struct vifentry *v)
{
	if (!v->v1neighnum || v->v2neighnum)
		send_assert_v2(mrtPtr, v);
	if (v->v1neighnum)
		send_assert_v1(mrtPtr, v);
}


/* function name: trigger_assert
   input: mrtentry pointer, vif number [the iif on which the 
   wrong pkt was received]
   output: none
   operation: gets the viflist entry for the corresponding iif,
   and calls assert... 
   Optional Assert rate limiting is currently not implemented.
   comments: called by process_wrongiif
   calls assert
 */
static void trigger_assert(struct mrtentry *mrtPtr, vifi_t iif)
{
	struct vifentry *v;
	u_long source, group;

	source = mrtPtr->source->source;
	group = mrtPtr->group->group;
	v = &viflist[iif];

	if (debug > 3)
		log(LOG_DEBUG, 0, "triggering assert for (%s,%s) on %s",
		    inet_fmt(source, s1), inet_fmt(group, s2), v->name);

	send_assert(mrtPtr, v);
}


/* function name: compare_metric
   input: mrt pointer, source address, vifi, metric, preference
   output: int (TRUE if the metric given is less than that of the
   mrt given, FALSE o.w.)
   operation: compares the preference values, if this is the same
   then compare the metric, if this is the same, then check
   the source address vs the local vifi address
   comments: called by accept_assert
 */
static int
compare_metric(struct mrtentry *mrtPtr, __u32 candidate, __u32 oldnh,
	       __u32 metric, __u32 preference)
{
	__u32 mrt_pref = ((mrtPtr->flags & MRTF_RP) ? 1 << 31 : 0);
	__u32 mrt_metric;

	if (mrtPtr->flags & MRTF_ASSERTED)
		mrt_pref |= mrtPtr->preference;
	else
		mrt_pref |= mrtPtr->source->preference;

	if (mrtPtr->flags & MRTF_ASSERTED)
		mrt_metric = mrtPtr->metric;
	else
		mrt_metric = mrtPtr->source->metric;


	if (preference < mrt_pref)
		return (TRUE);
	if (preference > mrt_pref)
		return (FALSE);

	/* the preference value is the same */
	if (metric < mrt_metric)
		return (TRUE);
	if (metric > mrt_metric)
		return (FALSE);

	/* metric & pref. are same, so check addresses */
	return ((ntohl(candidate) > ntohl(oldnh)));
}

int accept_assert_gen(vifi_t vifi, struct nbrentry *n,
		      __u32 source, __u32 group,
		      __u32 preference, __u32 metric)
{
	struct mrtentry *r;
	struct srcentry *s;
	struct vifentry *v = &viflist[vifi];
	__u32 oldnh;

	s = find_source_r(source, FALSE);
	if (s) {
		if (s->upstream == n) {
			s->asserted = 1;
			s->preference = preference & 0x7fffffff;
			s->metric = metric + v->metric;
		} else if (s->upstream)
			tracef(TRF_PIM, "accept_assert: no upstream for %s", inet_fmt(source, s1));
		else
			tracef(TRF_PIM, "accept_assert: upstream for %s is %s", inet_fmt(source, s1), inet_fmt(s->upstream->address, s2));
	} else
		tracef(TRF_PIM, "accept_assert: no source for %s", inet_fmt(source, s1));

	/* for the upstream routers we preform longest match */
	if ((r = longest_match_r(source, group)) == NULL) {
		tracef(TRF_PIM, "accept_assert: no longest match for (%s,%s)",
		       inet_fmt(source, s1),
		       inet_fmt(group, s2));
		return FALSE;
	}

	if (VIFM_ISSET(vifi, r->outgoing)) {
		if (compare_metric(r, n->address, v->prefsrc, metric, preference)) {
			del_outgo(vifi, r);
			return (FALSE);
		} else {
			/* I won! So assert with the winning metric */
			send_assert(r, v);
			return (TRUE);
		}
	}

	/* for downstream routers we perform an exact match */
	if ((r = find_route_r(source, group, EMPTY_FLAG, FALSE)) == NULL) {
		tracef(TRF_PIM, "accept_assert: no exact match for (%s,%s)",
		       inet_fmt(source, s1),
		       inet_fmt(group, s2));
		return FALSE;
	}

	if (vifi != r->source->incoming)
		return FALSE;

	if (r->flags & MRTF_ASSERTED) {
		oldnh = r->upstream->address;
		if (r->upstream == n) {
			r->assert_rpf_timer = 0;
			r->metric = metric;
			r->preference = preference & 0x7FFFFFFF;
			return TRUE;
		}
	} else {
		if (r->source->upstream)
			oldnh = r->source->upstream->address;
		else
			return FALSE;
		if (r->source->upstream == n)
			return TRUE;
	}

	if (compare_metric(r, n->address, oldnh, metric, preference)) {
		r->upstream = n;
		r->flags |= MRTF_ASSERTED;
		r->assert_rpf_timer = 0;
		r->metric = metric;
		r->preference = preference & 0x7FFFFFFF;
		return TRUE;
	}
	return FALSE;
}


/* function name: send_register
   input: pkt pointer
   output: int (TRUE if successful, FALSE o.w.)
   operation: If the matching entry has the REGISTER bit set, encapsulates 
   the data packet in a Register message, and unicast it
   to the RP for the group.
   comments: called by process_wholepkt
   calls find_route, and send_pim_unicast
 */

int send_register(char *pkt)
{
	struct ip *ip;
	struct mrtentry *r;
	struct grpentry *g;
	__u32 source, group;
	__u32 dest;
	int pktlen = 0;

	ip = (struct ip *) (pkt);
	source = ip->ip_src.s_addr;
	group = ip->ip_dst.s_addr;

	/* Check the matching entry, do not modify flags,
	 * and do not create if non-existent.
	 */
	if (!(r = find_route_r(source, group, EMPTY_FLAG, FALSE))) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "\nSend Register, no route %s -> %s",
			    inet_fmt(source, s1), inet_fmt(group, s2));
		return (FALSE);
	}

	/* if the register bit is off, don't send registers */
	if (!(r->flags & MRTF_REGISTER)) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "\nSend Register, but MRT has Reg-bit cleared");
		return (FALSE);
	}
	g = r->group;

	/* recompute the checksum; as checksum may fail when decapsulating
	 * at the RP if we don't do this.
	 */
	ip->ip_csum = 0;
	ip->ip_csum = inet_cksum((u_short *) ip, ip->ip_hl << 2);

	if (debug > 2) {
		log(LOG_DEBUG, 0, "\nRegistering for (%s,%s), RP=%s",
		    inet_fmt(source, s1), inet_fmt(group, s2),
		    inet_fmt(g->active_rp->rpaddr, s3));
	}
	dest = g->active_rp->rpaddr;

#if 0
	if (!(g->active_rp->pimversion&2))
#else
	if (g->active_rp->pimversion&1)
#endif
		send_register_v1(dest, ip, pktlen);
	else
		send_register_v2(dest, ip, pktlen);
	return (TRUE);
}


void send_null_register(struct mrtentry *r)
{
	int pimvers = r->group->active_rp->pimversion;

#if 0
	if (pimvers&2)
#else
	if (!(pimvers&1))
#endif
		send_null_register_v2(r);
	else
		send_null_register_v1(r);
}


int accept_register_gen(__u32 src, __u32 dst, char *pkt,
			int bb, int null_reg, void (*srs)(__u32, __u32, __u32, __u32, __u32))
{
	struct ip * ip = (struct ip *) pkt;
	__u32 source = ip->ip_src.s_addr;
	__u32 group = ip->ip_dst.s_addr;
	struct mrtentry *r;

	/* if I could not find a route, or the SPT is set,
	 * trigger register stop and return
	 */
	if (!(r = find_route_r(source, group, EMPTY_FLAG, FALSE))) {
		if (!(r = longest_match_r(source, group))) {
			if (debug > 3)
				log(LOG_DEBUG, 0, "\nNo longest match, Sending Reg-Stop");

			srs(source, group, dst, src, dst);
			return FALSE;
		}
		/* found a longest match, so create a new entry with iif register 
		 * flag set 
		 */
		if (!(r = find_route_r(source, group, MRTF_IIF_REGISTER, TRUE))) {
			if (debug > 4)
				log(LOG_DEBUG, 0, "\nCouldn't create IIF REG entry");
			return FALSE;
		}
	} else if ((r->flags & MRTF_SPT) || !computeOifs(r)) {
		/* I found an entry, so check the spt bit and the oiflist */
		if (debug > 3)
			log(LOG_DEBUG, 0, "\nSPT set or null oif, Sending Reg-Stop");

		srs(source, group, dst, src, dst);
		return FALSE;
	}
	/* if the entry is just created, tell the kernel */
	if (r->flags & MRTF_CRT) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "\nCreating kernel mfc... ");
		numberOfCaches++;
		k_chg_mrtf(r);
		r->flags &= ~MRTF_CRT;
		return TRUE;
	}
	/* if the route is not set up for receiving registers,
	 * modify it so that it is. Note that the oiflist
	 * is copied from other wild card entries (if exist)
	 * in kern.c when created the fwdg cache.
	 */
	if (!(r->flags & MRTF_IIF_REGISTER)) {
		r->flags |= MRTF_IIF_REGISTER;
		r->flags &= ~MRTF_NULL_OIF;
		/* tell the kernel to set the iif right */
		if (debug > 3)
			log(LOG_DEBUG, 0, "\nModifying kernel mfc... ");
		k_chg_mrtf(r);
	}
	r->reg_cnt++;
	return TRUE;
}

void accept_reg_stop_gen(__u32 source, __u32 group)
{
	struct mrtentry *r;

	/* get the route, do not check the flags, do not create ! */
	if (!(r = find_route_r(source, group, EMPTY_FLAG, FALSE)))
		return;

	/* Reset the Register bit to stop registering */
	r->flags &= ~MRTF_REGISTER;
	r->flags |= MRTF_REG_SUPP;	/* to keep the entry around, even if
					 * the oifs are null.
					 */
	/* restart the Register-bit timer */
	r->registerBitTimer = REGISTER_BIT_TIMEOUT;
	/* remove the register vif from kernel */
	k_chg_mrtf(r);
}


static int prune_rpt(struct mrtentry *mrtPtr)
{
	struct mrtentry *r;
	struct srcentry *sourcep;
	struct grpentry *groupp;
	struct srcentry *s;
	register struct vifentry *v;
	struct nbrentry *n;

	sourcep = mrtPtr->source;
	groupp = mrtPtr->group;

	if (!(r = find_route_r(sourcep->source, groupp->group, EMPTY_FLAG, FALSE)))
		return (FALSE);

	s = r->group->rp_entry->source;	/*sourcep; as the prune needs to be sent */
	if (local_address(s->source) != MAXVIFS)
		return FALSE;		/* I am RP, no prune needed */
	v = &viflist[s->incoming];
	n = s->upstream;

	if (n == NULL)
		return (FALSE);

	if (!v->v1neighnum || v->v2neighnum)
		prune_rpt_v2(v, n, mrtPtr);
	if (v->v1neighnum)
		prune_rpt_v1(v, n, mrtPtr);
	return TRUE;
}


/* function name: dr
   inputs: src; the source address & vif the virtual interface.
   outputs: int, 1 on success 0 on failure.
   operation: checks to see if the local router is the DR for the source
   specified by src, on the interface specified by vif; by checking the 
   netmasks and the VIFF_DR flag. 
 */
static int dr(__u32 src, vifi_t vif)
{
	struct srcentry *s;

	/* if not the DR for that interface, then return */
	if (debug > 4)
		log(LOG_DEBUG, 0, "dr: Checking if I'm the DR for src: %s, on vif: %d",
		    inet_fmt(src, s1), vif);

	if (!(viflist[vif].flags & VIFF_DR)) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "dr: vif: %d, not DR", vif);
		return 0;
	}

	s = find_source_r(src, TRUE);
	if (s && (s->direct == 2 || (s->direct && s->incoming == vif)))
		return 1;
	return 0;
}


/*************** Kernel Calls manipulations ********************************/

/* function name: process_cacheMiss
   input: igmp message pointer
   output: none
   operation: Check if I am the DR for the source, if so, and I am not
   the RP, then install an mfc in the kernel to pull up
   the packets and Register to the RP. If I am the DR, and 
   the RP, then just install entry to forward the packets
   according to the matching entries down the shared tree.
   If I am not the DR, then perform a longest match and install
   the entry accordingly.
   comments: called by process_kernelCall
 */
static void process_cacheMiss(struct igmpmsg *igmpctl)
{
	__u32 src, grp;
	vifi_t iif;
	struct vifentry *v;
	struct mrtentry *r;
	struct grpentry *g;

	/* when we get a cache miss or wrong iif, we expect only the 
	 * header 
	 */

	grp = igmpctl->im_dst.s_addr;
	src = igmpctl->im_src.s_addr;
	iif = igmpctl->im_vif;
	v = &viflist[iif];

	if (dr(src, iif)) {
		/* I am the DR for the src */
		if (!(g = find_group_r(grp, TRUE)))
			return;

		if (g->active_rp) {
			if (local_address(g->active_rp->rpaddr) == MAXVIFS) {
				/* if I am not the RP */
				/* create an entry with the REGISTER flag set */
				if ((r = find_route_r(src, grp, MRTF_REGISTER, TRUE)) != NULL) {
					r->flags &= ~MRTF_CRT;
					if (debug > 3)
						log(LOG_DEBUG, 0, "process_cacheMiss:Installing entry in kernel cache with reg vif");
					numberOfCaches++;
					k_chg_mrtf(r);
				}	/* if find_route */
				return;
			} else {
				/* if RP, should forward down (*,G), or (S,G,RP-bit) */

				if ((r = longest_match_r(src, grp)) != NULL) {
					if (r->flags & MRTF_WC) {	/* if the match was (*,G), create (S,G) */

						if ((r = find_route_r(src, grp, EMPTY_FLAG, TRUE)) != NULL) {
							numberOfCaches++;
							if (debug > 3)
								log(LOG_DEBUG, 0, "process_cacheMiss:RP: Creating (S,G) off WC in the kernel");
							k_chg_mrtf(r);
						}
						return;
					} else {
						/* if the match was not (*,G), just install in kernel */
						numberOfCaches++;
						if (debug > 3)
							log(LOG_DEBUG, 0, "process_cacheMiss:RP: Creating (S,G) in the kernel");
						k_chg_mrtf(r);
						return;
					}	/* match is not (*,G) */
				}	/* if there was a longest match */
			}
		}
	} else {
		/* I am not the DR, so perform longest match */
		if ((r = longest_match_r(src, grp)) != NULL) {
			/* If the call to kernel fails we will have an entry
			 * in daemon but not the kernel...
			 * The coming is now taken care of in accept register.., 
			 * but just in case we fail there, or we get a race
			 * condition between raw_input and socket_send ...
			 */
			if (iif == reg_vif_num) {
				if (!(r = find_route_r(src, grp, MRTF_IIF_REGISTER, TRUE)))
					return;

				r->flags &= ~MRTF_CRT;

				if (debug > 4)
					log(LOG_DEBUG, 0, "process_cacheMiss: Creating Register kernel cache.. ");
				numberOfCaches++;
				k_chg_mrtf(r);
				return;
			}	/* if the iif is the reg_vif */
			if (r->flags & MRTF_WC) {
				if ((r = find_route_r(src, grp, EMPTY_FLAG, TRUE)) != NULL) {
					if (debug > 3)
						log(LOG_DEBUG, 0, "process_cacheMiss:Creating (S,G) off WC in the kernel");
					numberOfCaches++;
					k_chg_mrtf(r);
				}
				return;
			} else {
				/* not (*,G) so just install kernel cache */
				if (debug > 3)
					log(LOG_DEBUG, 0, "process_cacheMiss: Creating (S,G) in the kernel");
				numberOfCaches++;
				k_chg_mrtf(r);
				return;
			}
		}
	}

	/* 
	 * Default to installing null oif cache in the kernel
	 * to avoid further cache misses on the same entry in the kernel.
	 * Just an implementation optimization, not a requirement.
	 */

	if (debug > 5)
		log(LOG_DEBUG, 0, "process_cacheMiss: Installing MRTF_NULL_OIF entry");

	if (!(r = find_route_r(src, grp, MRTF_NULL_OIF, TRUE)))
		return;
	numberOfCaches++;
	k_chg_mrtf(r);
	return;
}

/* function name: computeOifs
   input: mrtentry pointer
   output: vifbitmap_t the oifs computed
   operation: computes the outgoing interface list,
   if the entry is (S,G) the oiflist is copied from
   (*,G) entry if exists. For (S,G) RPT-bit, the prune
   bitmask (vifmaskoff) is cleared from the oiflist.
   comments: called by process_wrongiif
 */
static vifbitmap_t computeOifs(struct mrtentry *mrtPtr)
{
	struct mrtentry *rpPtr;	/* a pointer to the rp (*,G) entry */
	vifbitmap_t oifs;


	oifs = mrtPtr->group->leaves;
	oifs |= mrtPtr->outgoing & ~nomrt_vif;

	if (mrtPtr->flags & MRTF_WC) {	/* this is a (*,G) entry */
		if (debug > 4)
			log(LOG_DEBUG, 0, "computeOifs: OutMask: %d, (*,G)", oifs);
		return (oifs);
	}
	if (!(rpPtr = mrtPtr->group->rp_entry)) {
		/* this is a (S,G) entry and there is no (*,G) entry */
		if (debug > 4)
			log(LOG_DEBUG, 0, "computeOifs: OutMask: %d, (S,G) without (*,G)", oifs);
		return (oifs);
	}
	/* if there is a (*,G) then use the vifmaskoff */
	if (debug > 4)
		log(LOG_DEBUG, 0, "computeOifs: OutMask: %d, (S,G) with (*,G)", oifs | (rpPtr->outgoing & ~(mrtPtr->vifmaskoff) & ~nomrt_vif));

	return (oifs | (rpPtr->outgoing & ~(mrtPtr->vifmaskoff) & ~nomrt_vif));
}

/* function name: process_wrongiif
   input: igmp message pointer
   output: none
   operation: Checks the matching entry. If I am the RP for the group,
   had an IIF_REGISTER entry, and the iif is the one towards
   the source, then set the SPT bit; thus completing the switch
   to the SPT. If I am not the RP, the iif is that towards the src,
   and iif(S,G) != iif(*,G), trigger prune off RP tree, and set
   the SPT; thus completing the switch to SPT.
   Else, if the packet's iif is in the oiflist of the entry,
   then trigger an Assert.
   comments: called by process_kernelCall
 */
static void process_wrongiif(struct igmpmsg *igmpctl)
{
	struct mrtentry *r;
	__u32 src, grp;
	vifi_t iif;

	/* when we get a cache miss or wrong iif, we expect only the
	 * header. 
	 */

	grp = igmpctl->im_dst.s_addr;
	src = igmpctl->im_src.s_addr;
	iif = igmpctl->im_vif;

	if (!(r = find_route_r(src, grp, MRTF_KERNEL_CACHE, FALSE))) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "Wrong IIF, but no cache mirror !");
		return;		/* should not happen */
	}
	if (r->group->active_rp) {
		if (local_address(r->group->active_rp->rpaddr) != MAXVIFS) {	/* I am the RP for the group */
			if ((iif == r->source->incoming) &&
			    (r->flags & MRTF_IIF_REGISTER)) {
				if (r->flags & MRTF_SPT)
					return;		/* this should not happen */
				else {
					/* set the spt bit */
					r->flags |= MRTF_SPT;
					/* clear the iif register flag */
					r->flags &= ~MRTF_IIF_REGISTER;
					/* update the kernel */
					if (debug > 4)
						log(LOG_DEBUG, 0, "WrongIIF: setting SPT & updating kernel");
					k_chg_mrtf(r);
					return;
				}	/* if the spt is cleared */
			}	/* if iif=r->source->incoming... */
		} else {
			/* if I am not the RP, check the switch to SPT, and asserts */
			/* if iif towards source is not same as that towards RP */
			if (iif == r->source->incoming &&
			    iif != r->group->rp_entry->source->incoming) {
				r->flags |= MRTF_SPT;	/* set the SPT bit */

				/* trigger a prune upstream towards the RP */
				if (debug > 4)
					log(LOG_DEBUG, 0, "WrongIIF: Pruning RPT");
				prune_rpt(r);
				k_chg_mrtf(r);
				/* update the kernel with the new iif */
				return;
			} else {
				/* redundant, for clarity */
				/* check for asserts */
				if (VIFM_ISSET(iif, computeOifs(r)))
					trigger_assert(r, iif);
				return;
			}
		}		/* if I am not the RP */
	}			/* if grp->active_rp */
}

/* function name: process_kernelCall
   inputs: none
   outputs: none
   operation: takes the recvbuf, and processes the kernelCall from the kernel
   accdg to msg_type. 
   Currently there are 3 types for kernelCalls: NOCACHE, WRONGVIF, and 
   WHOLEPKT (used for registering).     
   comments: called by accept_igmp
   calls process_cacheMiss, process_wrongiif and process_wholepkt
 */
char *msg_kind(int t)
{
	switch (t) {
	case IGMPMSG_NOCACHE:
		return ("IGMPMSG_NO_CACHE");
	case IGMPMSG_WRONGVIF:
		return ("IGMPMSG_WRONGVIF");
	case IGMPMSG_WHOLEPKT:
		return ("IGMPMSG_WHOLEPKT");
	default:
		return ("UNKNWON IGMPMSG");
	}

}

void process_kernelCall(char *data)
{
	struct igmpmsg *igmpctl;	/* igmpmsg control struct */
	igmpctl = (struct igmpmsg *) data;

	tracef(TRF_MRTSOCK, "RX MRTSOCK: %s, im_mbz: %d, Vif: %d (%s -> %s)",
		    msg_kind(igmpctl->im_msgtype),
		    igmpctl->im_mbz,
		    igmpctl->im_vif,
		    inet_fmt(igmpctl->im_src.s_addr, s1),
		    inet_fmt(igmpctl->im_dst.s_addr, s2));

	switch (igmpctl->im_msgtype) {	/* switch on the message type */
	case IGMPMSG_NOCACHE:
		process_cacheMiss(igmpctl);
		break;

	case IGMPMSG_WRONGVIF:
		process_wrongiif(igmpctl);
		break;

	case IGMPMSG_WHOLEPKT:
		send_register(data + sizeof(struct ip));
		break;

	default:
		if (!(trace_flags&TRF_MRTSOCK))
			tracef(TRF_UNUSUAL, "RX MRTSOCK: bad type=%d, im_mbz: %d, Vif: %d, im_src: %s, im_dst: %s",
			       igmpctl->im_msgtype,
			       igmpctl->im_mbz,
			       igmpctl->im_vif,
			       inet_fmt(igmpctl->im_src.s_addr, s1),
			       inet_fmt(igmpctl->im_dst.s_addr, s2));
	}
}
/* end kernelCall */


/********************** Debugging Dumps *******************************/

static void dump_sources(FILE *fp)
{
	struct srcentry *s;

	fprintf(fp, "                    Source List\n"
		"S               NH                    Metric\n\n");

	forall_sources(s) {
		fprintf(fp, "%-15s %-5s %-15s %x/%x\n",
			inet_fmt(s->source, s1),
			s->incoming < numvifs ? viflist[s->incoming].name : "none",
			s->upstream ? inet_fmt(s->upstream->address, s2) : "direct",
			s->preference, s->metric);
	} endforall_sources(s);
	fprintf(fp, "\n");
}

static void dump_groups(FILE *fp)
{
	struct grpentry *g;

	fprintf(fp, "                    Group List\n"
		"G               RP                WC     Leaves\n\n");

	forall_groups(g) {
		vifi_t vifi;

		fprintf(fp, "%-15s %-15s %-6s",
			inet_fmt(g->group, s1),
			inet_fmt(g->active_rp->rpaddr, s2),
			g->rp_entry ? "exists" : "none");

		for (vifi = 0; vifi < numvifs; ++vifi)
			if (VIFM_ISSET(vifi, g->leaves))
				fprintf(fp, " %-4s", viflist[vifi].name);
		fprintf(fp, "\n");
	} endforall_groups(g);
	fprintf(fp, "\n");
}

/* For debugging purposes only */
void dump_routes(FILE *fp)
{
	struct grpentry *g;
	register struct mrtentry *r;
	register vifi_t vifi;
	u_int numberOfCacheMirrors = 0;

	dump_sources(fp);
	dump_groups(fp);

	fprintf(fp, "                    Multicast Routing Table\n"
		"(S,G)                           RP              Flags\n\n");

	forall_groups(g) {
		for (r = g->mrtlink; r != NULL; r = r->grpnext) {
			if (r->flags & MRTF_CRT)
				continue;

			if (!(r->flags & MRTF_WC)) {
				char buf[64];
				sprintf(buf, "(%s,%s)",
					inet_fmt(r->source->source, s1),
					inet_fmt(r->group->group, s2));
				fprintf(fp, "%-31s ", buf);
			} else if (r->source->source != r->group->active_rp->rpaddr) {
				char buf[64];
				sprintf(buf, "(*%s,%s)",
					inet_fmt(r->source->source, s1),
					inet_fmt(r->group->group, s2));
				fprintf(fp, "%-31s ", buf);
			} else {
				char buf[64];
				sprintf(buf, "(*,%s)",
					inet_fmt(r->group->group, s1));
				fprintf(fp, "%-31s ", buf);
			}

			if (!(r->flags&MRTF_DENSE))
				fprintf(fp, "%-15s", inet_fmt(g->active_rp->rpaddr, s1));
			else
				fprintf(fp, "%-15s", "dense");

			/* fprintf(fp," %8u ", r->pktrate_prev ); */
			if (r->flags & MRTF_KERNEL_CACHE) {
				fprintf(fp, "K");
				numberOfCacheMirrors++;
			}
			if (r->flags & MRTF_DENSE)
				fprintf(fp, "D");
			if (r->flags & MRTF_SPT)
				fprintf(fp, "T");
			if (r->flags & MRTF_RP)
				fprintf(fp, "R");
			if (r->flags & MRTF_REGISTER)
				fprintf(fp, "F");

			if (r->flags & MRTF_SG)
				fprintf(fp, " SG");

			fprintf(fp, "\n");

			fprintf(fp, "\tIncoming: %-6s", viflist[r->source->incoming].name);

			if (r->flags & MRTF_IIF_REGISTER)
				fprintf(fp, " IIF_REG");

			if (r->flags & MRTF_ASSERTED) {
				fprintf(fp, " Asserted to ");
				if (r->upstream) {
					fprintf(fp, "%s Metric %x/%x\n",
						inet_fmt(r->upstream->address, s1),
						r->preference, r->metric);
				} else
					fprintf(fp, "????\n");
			} else
				fprintf(fp, "\n");

			fprintf(fp, "\tOutgoing: ");
			for (vifi = 0; vifi < numvifs; ++vifi)
				if (VIFM_ISSET(vifi, r->outgoing))
					fprintf(fp, "%-4s ",
						viflist[vifi].name);

			if (r->flags & MRTF_NULL_OIF)
				fprintf(fp, "NULL_OIF ");

			if (r->flags & MRTF_REGISTER)
				fprintf(fp, "register ");
			if (r->flags & MRTF_REG_SUPP)
				fprintf(fp, "register(suppressed) ");

			fprintf(fp, "\n");
		}
	} endforall_groups(g);
	fprintf(fp, "\n");
	fprintf(fp, "Number of Cache MIRRORs is %u\n",
		numberOfCacheMirrors);
}

