/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: timer.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */

#include "defs.h"

/**************** House-Keeping and Timer Periodic Processing **************/

/* function name: spt_switch
   input: source and destination addresses, igmp msg (char *),
   and length of the pim packet
   output: none
   operation
   comments: called from check_reg_rate, and check_data_rate
 */
static void spt_switch(struct mrtentry *mrtPtr)
{
	__u32 source, group;

	source = mrtPtr->source->source;
	group = mrtPtr->group->group;

	if (debug > 3)
		log(LOG_DEBUG, 0, "PIM: spt_switch: (%s,%s) is now pure SG",
		    inet_fmt(source, s1),
		    inet_fmt(group, s2));

	/* now this is a pure S,G entry that triggers joins upstream */
	mrtPtr->flags |= MRTF_SG;
	dirty_bits |= 2;
}


/* function name: del_empty_vifs
   input: a pointer to the mrt, and a pointer to the previous
   mrt, in case we are going to delete the former.
   output: int (TRUE, if going to delete, o.w. FALSE
   operation:
   comments: called by clean_entry, 
 */
static int del_empty_vifs(struct mrtentry *r)
{
	/* Check if this is a problem for the (*,G) entry in the RP */
	/* for (*,G) 'not at the RP' delete all the other pertinent entries,
	   as well,... check spec., this should happen automatically... check */

	/* if this is S,G [not RP] and there is a *,G then should chk oifs for *,G */
	if ((r->group->rp_entry) && (!((r->flags & MRTF_WC) | (r->flags & MRTF_RP))))
		if (r->group->rp_entry->outgoing != EMPTY_VIF)
			return (FALSE);

	/* for register suppression, leave the entry til it times out */
	if (r->flags & MRTF_REG_SUPP)
		return (FALSE);

	/* for register entries, the oif = reg_vif */
	if (((r->group->leaves | r->outgoing | r->vifmaskoff) == EMPTY_VIF) &&
	    !(r->flags & MRTF_REGISTER)) {
		log(LOG_DEBUG, 0, "No oifs Deleting entry (%s,%s)(FIX?)",
		    inet_fmt(r->source->source, s1),
		    inet_fmt(r->group->group, s2));
		return TRUE;
	}
	return (FALSE);
}


/* function name: check_sg_cnt
   input: mrtentry pointer
   output: int (TRUE, if the last count matches the current count,
   i.e. the entry is idle, FALSE o.w.).
   operation: calls the kernel to check for the packet count for the
   given (S,G) entry, then compares the value obtained from
   the kernel with the previous value of the packet count
   for that entry to determine the activity of the entry.
   comments: called by timeout_entry_timer
   calls k_get_sg
 */
static int check_sg_cnt(struct mrtentry *r)
{
	u_long newCnt;

	if (!(r->flags & MRTF_KERNEL_CACHE))
		return -1;	/* this should not happen */

	newCnt = k_get_sg(r);

	if (newCnt == 0xffffffff)
		log(LOG_DEBUG, 0, "**ERR** newCnt is %x", newCnt);

	if (debug > 4) {
		log(LOG_DEBUG, 0, "NewCnt is %x, old is %x", newCnt,
		    r->idlecnt_prev);

		log(LOG_DEBUG, 0, "*** Pkt Cnt Diff for (%s,%s) is %d",
		    inet_fmt(r->source->source, s1), inet_fmt(r->group->group, s2),
		    (int) (newCnt - r->idlecnt_prev));
	}
	if (r->idlecnt_prev == newCnt)
		return (TRUE);

	r->idlecnt_prev = newCnt;

	return (FALSE);
}

/* function name: timeout_entry_timer
   input: mrt entry pointer, and a pointer to the previous mrt entry, just in
   case we decided to delete the former.
   output: int (TRUE, if the entry timed out, FALSE if not)
   operation: For entries that are mirrors for kernel cache, age
   the timer, and if expired, check if there is any
   activity (i.e. data packets forwarded) accdg to the 
   entry in the kernel.
   comments: called by clean_entry
   calls check_sg_cnt
 */
static int timeout_entry_timer(struct mrtentry *r)
{
	r->timer += TIMER_INTERVAL;
	if (r->timer < ROUTE_TIMEOUT)
		return (FALSE);

	/* have to check the S,G count from the kernel if this is a kernel cache */
	if (r->flags & MRTF_KERNEL_CACHE)
		if (check_sg_cnt(r) == TRUE)
			return TRUE;
		else {
			r->timer = 0;
			return (FALSE);
		}
	/* reset the timer */ 
	else			/* not cache, i.e. (*,G) w/ null oif, so delete */
		return TRUE;
}

/* function name: clean_entry
   input: pointer to the entry to be cleaned, and a pointer to the 
   entry previous to that (that is needed when/if the entry is
   deleted).
   output: int (TRUE if the entry is deleted, o.w. FALSE)
   operation
   comments: called by age_routes
 */
static int clean_entry(struct mrtentry *r)
{
	if ((r->flags & MRTF_WC) && !(r->group->leaves) && !(r->outgoing))
		/*this is a (*,G) with null oif and leaves, then check the 
		   timer of the entry [this case occurs when receiving
		   registers without any relevant group entry existing; chk spec]
		 */
		return timeout_entry_timer(r);

	if (del_empty_vifs(r))
		return (TRUE);

	/* if the entry still isn't deleted, then age the cache */
	if (r->flags & MRTF_KERNEL_CACHE)
		return timeout_entry_timer(r);

	return (FALSE);
}

/* function name: timeout_assert_rpf
   input: mrt pointer
   output: int (TRUE if timed out, FLASE o.w.)
   operation: ages out the timer for the assert_rpf, if expired
   resets the asserted flag.
   comments: called by age_routes
 */
static int timeout_assert_rpf(struct mrtentry *mrtPtr)
{
	mrtPtr->assert_rpf_timer += TIMER_INTERVAL;

	if (mrtPtr->assert_rpf_timer < ASSERT_RPF_TIMER)
		return (FALSE);

	mrtPtr->flags &= ~MRTF_ASSERTED;
	mrtPtr->upstream = NULL;
	dirty_bits |= 2;
	return TRUE;
}

/* function name: update_src_routes_oifs
   input: srcentry pointer, vifi [the old iif], and the old nbr address.
   output: int
   operation: updates the oifs accdg to the new unicast routing,
   comments: called by unicast_routing_changes
 */
static int update_src_routes_oifs(struct srcentry *src_ent,
				  vifi_t oldiif,
				  __u32 oldnbr)
{
	struct mrtentry *mrtPtr;

	if (oldiif != src_ent->incoming)
		/* a change in the iif occured, so change the oifs accdgly */
		for (mrtPtr = src_ent->mrtlink; mrtPtr; mrtPtr = mrtPtr->srcnext) {
			/* remove the new iif from the oifs */
			if (VIFM_ISSET(src_ent->incoming, mrtPtr->outgoing))
				VIFM_CLR(src_ent->incoming, mrtPtr->outgoing);
		}
	/* FIX optionally, can trigger a join to the new iif, and prune off of
	   old iif. Also, if the nbr only changed, can trigger prune off
	   of the old nbr and join to the new nbr; but this is only 
	   optimization, as this would happen periodically. */
	return (TRUE);
}

/* function name: unicast_routing_changes
   input: srcentry pointer
   output: int (TRUE if a change in unicast routing happens,
   FALSE o.w.)
   operation: checks the timer for the src entry passed, if the
   timer expired then checks the unicast routing tables
   (currently implemented thru an interface to the kernel)
   and updates the iif and rpf neighbor accordingly.
   comments: called by clean_srclist
   calls set_src_iif
 */
int unicast_routing_check(struct srcentry *src_ent)
{
	vifi_t oldiif;
	__u32 oldnbr, newnbr;
	__u32 oldpref, oldmetric;
	int oldasserted;

	/* maintain the old info to check for the change */
	oldiif = src_ent->incoming;
	oldpref = src_ent->preference;
	oldmetric = src_ent->metric;
	oldasserted = src_ent->asserted;

	if (src_ent->upstream != NULL)
		oldnbr = src_ent->upstream->address;
	else
		oldnbr = 0;

	src_ent->timer = 0;	/* reset the timer */

	if (!set_src_iif(src_ent)) {
		/* No route to source! */
		log(LOG_WARNING, 0, "route to %s is lost!!!!", inet_fmt(src_ent->source, s1));
		src_ent->upstream = NULL;
		src_ent->incoming = update_src_routes_oifs(src_ent, oldiif, oldnbr);
		return (TRUE);
	}

	if (src_ent->upstream != NULL)
		newnbr = src_ent->upstream->address;
	else
		newnbr = 0;

	if (oldnbr != newnbr || oldiif != src_ent->incoming) {
		update_src_routes_oifs(src_ent, oldiif, oldnbr);
		return (TRUE);
	}

	if (oldasserted && !src_ent->asserted) {
		src_ent->preference = oldpref;
		src_ent->metric = oldmetric;
		src_ent->asserted = 1;
	}
	return (FALSE);
}

/* function name: clean_srclist
   input: none
   output: int (TRUE if a source entry (or more) was deleted, FALSE o.w.)
   operation: browses thru the srclist and checks the mrtlink, if a source 
   entry does not have an mrtlink, the entry is deleted.
   comments: called by age_routes
 */
static int clean_srclist(void)
{
	int cleaned = 0;
	struct srcentry *s, **sp;
	int h;

	for (h = 0; h < SRCHASH_SIZE; h++) {
		for (sp = &srchash[h]; (s=*sp) != NULL; ) {
			struct mrtentry *r;

			s->timer += TIMER_INTERVAL;
			if (s->timer < AssertTimeout) {
				sp = &s->next;
				continue;
			}

			r = s->mrtlink;
			if (r == NULL) {
				if (debug > 3)
					log(LOG_DEBUG, 0, "Removing entry for source %s",
					    inet_fmt(s->source, s1));
				*sp = s->next;
				free(s);
				cleaned++;
				continue;
			}
			unicast_routing_check(s);
			sp = &s->next;
		}
	}
	return (cleaned);
}

/**************** Periodic Join/Prune manipulations *******************/

/*Function name: join_or_prune
   input: nbr entry pointer, virtual interface, and mrt entry ptr
   output: int (join is 1, prune is 2, or none is 3)
   operation
   comments: called by periodic_join_prune
 */
int join_or_prune(struct nbrentry *n, vifi_t vifi, struct mrtentry *r)
{
	struct grpentry *g;
	struct srcentry *src;
	vifi_t iif, rp_iif;
	struct nbrentry *src_upstream, *rp_upstream;

	g = r->group;
	src = r->source;

	/* if not (*,G) or (S,G)RPbit, or pure (S,G) return */
	if (!(r->flags & MRTF_WC) && !(r->flags & MRTF_RP) &&
	    !(r->flags & MRTF_SG))
		return NONE;

	/* if the entry was asserted, use assert winner as rpf */
	if (r->flags & MRTF_ASSERTED) {
		src_upstream = r->upstream;
		iif = src_upstream ? src_upstream->vif : MAXVIFS;
	} else {
		src_upstream = src->upstream;
		iif = src->incoming;
	}

	if (g->rp_entry) {
		if (g->rp_entry->flags & MRTF_ASSERTED) {
			rp_upstream = g->rp_entry->upstream;
			rp_iif = rp_upstream ? rp_upstream->vif : MAXVIFS;
		} else {
			rp_upstream = g->rp_entry->source->upstream;
			rp_iif = g->rp_entry->source->incoming;
		}
	}

	/* it is a (*,G) entry */
	if (g->rp_entry == r) {
		/* check if the iif and rpf nbr match, then send join */
		if (iif == vifi && src_upstream == n)
			return JOIN;
		else
			return NONE;
	}

	/* There is no (*,G) entry so this has to be a join */
	if (!g->rp_entry) {
		/* check if the iif and rpf nbr match, then send join */
		if (iif == vifi && src_upstream == n)
			return JOIN;
		else
			return NONE;
	}

	/* Now for the case when there is a (*,G) entry */

	/* all the nodes downstream on RP_Tree have pruned off S */
	if (r->vifmaskoff == g->rp_entry->outgoing) {
		if ((r->flags & MRTF_RP) && !(r->flags & MRTF_WC))
			/* if it is an RP entry then send prunes */
			if (rp_iif == vifi && rp_upstream == n)
				return PRUNE;
			else
				return NONE;
	}

	/* if the SPT bit is set, and the rpf nbrs differ, then send prune
	   off of RP */
	if ((r->flags & MRTF_SPT)) {
		/* check if the rpf neighbors towards the src and RP differ */
		if (iif != rp_iif && rp_iif == vifi && rp_upstream == n)
				return PRUNE;
	}

	/* if the iif and the rpf nbr match, send join */
	if (iif == vifi && src_upstream == n)
		return JOIN;
	else
		return NONE;
}


/* function name: periodic_vif_join_prune
   input: none
   output: int; number of vifs for which join/prune messages were sent
   operation: browses thru the viflist and sends join/prune messages
   on vifs as appropriate.
   comments: called by age_routes
   calls periodic_join_prune
 */
static int periodic_vif_join_prune(void)
{
	int numVifsJoinedPruned = 0, numGrpsJoinedPruned = 0, newGrps;
	vifi_t vifi;
	register struct vifentry *v;
	register struct nbrentry *n;

	/* browse thru the vif list and send periodic join prune messages */
	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		if (v->flags & (VIFF_DOWN | VIFF_DISABLED))
			continue;
		for (n = v->neighbors; n != NULL; n = n->next) {
			if (!v->v1neighnum || v->v2neighnum)
				newGrps = periodic_join_prune_v2(n, vifi);
			if (v->v1neighnum)
				newGrps = periodic_join_prune_v1(n, vifi);
			numGrpsJoinedPruned += newGrps;
			numVifsJoinedPruned += (newGrps ? 1 : 0);
		}
	}
	if (debug > 3)
		log(LOG_DEBUG, 0, "joined/pruned %d grps, on %d vifs",
		    numGrpsJoinedPruned, numVifsJoinedPruned);

	return (numVifsJoinedPruned);
}


/* function name: check_reg_rate
   input: mrt entry pointer
   output: int (TRUE if rate exceeded, o.w. FALSE)
   operation: ages the timer for checking the register rate,
   then checks the register counter [if the timer
   had expired], if the limit is exceeded, triggers
   a switch to the shortest path.
   comments: called by age_routes
 */
static int check_reg_rate(struct mrtentry *mrtPtr)
{
	mrtPtr->reg_rate_timer += TIMER_INTERVAL;	/* age the timer */

	if (mrtPtr->reg_rate_timer < reg_rate_timer)
		return (FALSE);

	if (mrtPtr->reg_cnt < reg_rate_limit) {		/* if the rate is not exceeded, then reset all and return */
		mrtPtr->reg_cnt = 0;
		mrtPtr->reg_rate_timer = 0;
		return (FALSE);
	}
	/* switch to the SPT by sending annotated joins towards the src,
	   resetting the SPT bit, and changing the flag of the mrt */
	if (debug)
		log(LOG_DEBUG, 0, "check_reg_rate: Changing to SPT");
	spt_switch(mrtPtr);
	return TRUE;
}

/* function name: check_data_rate
   input: mrt entry pointer
   output: int (TRUE if switch to SPT, o.w. FALSE)
   operation: ages the timer for checking the data rate,
   if timer expired, then check the data pkt count,
   if exceeded the limit, then trigger switch to SPT.
   comments: called by age_routes
 */
static int check_data_rate(struct mrtentry *mrtPtr)
{
	u_long newCnt, pktdiff = 0;

	/* only check for those entries for which we have leaves */
	if (mrtPtr->group->leaves == EMPTY_VIF)
		return (FALSE);

	/* if this is not a cache mirror, then there is no activity
	   in the kernel for that entry... hence no need to check */
	if (!(mrtPtr->flags & MRTF_KERNEL_CACHE))
		return (FALSE);

	mrtPtr->data_rate_timer += TIMER_INTERVAL;	/* age the timer */

	if (mrtPtr->data_rate_timer < data_rate_timer)
		return (FALSE);

	newCnt = k_get_sg(mrtPtr);

	if (newCnt == 0xffffffff) {	/* should not happen */
		if (debug > 3)
			log(LOG_DEBUG, 0, "**ERR** can't get Cnt for S %s, G %s",
			    inet_fmt(mrtPtr->source->source, s1),
			    inet_fmt(mrtPtr->group->group, s2));
		return (FALSE);
	}
	if (mrtPtr->pktrate_prev <= newCnt)
		pktdiff = newCnt - mrtPtr->pktrate_prev;
	else			/* we wrapped around.. ! */
		pktdiff = newCnt + (0xffffffff - mrtPtr->pktrate_prev);

	if (pktdiff < data_rate_limit) {	/* if the rate is not exceeded, then reset all and return */
		/* keep the count for the next time around */
		mrtPtr->pktrate_prev = newCnt;
		mrtPtr->data_rate_timer = 0;
		return (FALSE);
	}
	/* switch to the SPT by sending annotated joins towards the src,
	   resetting the SPT bit, and changing the flag of the mrt */
	if (debug)
		log(LOG_DEBUG, 0, "check_reg_rate: Changing to SPT");
	spt_switch(mrtPtr);
	return TRUE;
}

/* function name: timeout_leaf
   input: group entry
   output: none
   operation: increases the timers for the leaves oifs for the given
   group. If a timer exceeds the upper limit the corresponding
   oif is cleared from the leaves.
   comments: called by age_routes
 */
static void timeout_leaf(struct grpentry *g)
{
	vifbitmap_t leaves;
	register vifi_t vifi;
	register u_char *t;

	leaves = g->leaves;

	for (vifi = 0, t = g->timers; vifi < numvifs; ++vifi) {
		if (VIFM_ISSET(vifi, g->leaves)) {
			t[vifi] = t[vifi] + TIMER_INTERVAL;
			if ((t[vifi] > LEAF_TIMEOUT) ||
			    (!(&viflist[vifi])->flags & VIFF_DR)) {
				/* or it is no longer a DR,
				 * might need to put 
				 * extra check when put 
				 * forwarder bit 
				 */
				if (debug > 4)
					log(LOG_DEBUG, 0, "Clearing: %d on (-,%s)",
					    vifi,
					    inet_fmt(g->group, s1));
				VIFM_CLR(vifi, g->leaves);

				t[vifi] = 0;
			}
		}
	}
	if (leaves != g->leaves)
		if (debug > 3)
			log(LOG_DEBUG, 0, "Timed out leaf for group %s",
			    inet_fmt(g->group, s1));
	return;
}

/* function name: timeout_outgo
   input: mrt entry pointer
   output: none
   operation: For the given entry, age out the oif timers. If the timer
   exceeds the timeout for the given vif, the corresponding
   vif is cleared in the oiflist.
   comments: called by age_routes
 */
static void timeout_outgo(struct mrtentry *r)
{
	vifbitmap_t outgoing;
	register vifi_t vifi;
	register u_char *t;

	outgoing = r->outgoing;
	for (vifi = 0, t = r->timers; (vifi < numvifs) && (t != NULL); ++vifi) {
		t[vifi] += TIMER_INTERVAL;
		if (VIFM_ISSET(vifi, outgoing) && ((t[vifi]) > oifTimeout[vifi])) {
			t[vifi] = 0;
			if (debug > 4)
				log(LOG_DEBUG, 0, "Clearing vif:%d on (%s,%s) oif",
				    vifi,
				    inet_fmt(r->source->source, s1),
				    inet_fmt(r->group->group, s2));
			VIFM_CLR(vifi, r->outgoing);
		}
	}
}


/* function name: ageRegisterBitTimer
   input: none
   output: int (TRUE if RP-Set timer expired, False o.w.)
   operation: decreases the Register Bit timer, if expired, resets the 
   REGISTER bit, and installs the register vif  
   comments: called by age_routes
 */
static int ageRegisterBitTimer(struct mrtentry *r)
{
	/* decrease the timer */
	r->registerBitTimer -= TIMER_INTERVAL;
	if (r->registerBitTimer == TIMER_INTERVAL) {	/* if about to expire,
							   * send null register.
							 */
		send_null_register(r);
		return (FALSE);
	}
	if (r->registerBitTimer > 0)
		return (FALSE);

	if (debug > 2)
		log(LOG_DEBUG, 0, "\nRegister bit timer timed out");

	/* we may have to check more things, e.g. if the RP changes and I am
	   the RP now,.. */
	r->flags |= MRTF_REGISTER;
	r->flags &= ~MRTF_REG_SUPP;	/* turn off register suppression */
	dirty_bits |= 2;
	k_chg_mrtf(r);
	return TRUE;
}

/* function name: delete_mrt
   input: a pointer to the mrt to be deleted, and a pointer 
   to the previous entry to that.
   output: int (TRUE if successful)
 */

static void delete_mrt(struct mrtentry *r)
{
	if (debug > 3)
		log(LOG_DEBUG, 0, "\ndeleting the entry (%s,%s)",
		    inet_fmt(r->source->source, s1),
		    inet_fmt(r->group->group, s2));

	if ((r->flags & MRTF_RP) && (r->flags & MRTF_WC))
		r->group->rp_entry = NULL;

	unlink_route(r);

	/* if this is a mirror entry in the kernel... delete it from
	   kernel as well */

	if (r->flags & MRTF_KERNEL_CACHE) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "\ndeleting entry in kernel");
		k_del_mrt(r);
	}

	free(r);
}


/* function name: age_routes
   input: none
   output: none
   operation: is triggered every 5 sec.s by the timer
   interrupt to perform house cleaning
   comments:
 */
void age_routes()
{
	int h;
	struct mrtentry *r;
	struct grpentry *g, **gp;

	/* browse thru the group list */
	for (h = 0; h < GRPHASH_SIZE; h++) {
		for (gp = &grphash[h]; (g=*gp) != NULL; ) {
			timeout_leaf(g);

			/* browse thru the mcast routing tables (mrt) */
			for (r = g->mrtlink; r != NULL;) {
				timeout_outgo(r);

				if (r->registerBitTimer)
					ageRegisterBitTimer(r);

				if (r->flags & MRTF_IIF_REGISTER)
					check_reg_rate(r);
				/* Rate limitting the registers sent */
				r->reg_snd_cnt = 0;

				check_data_rate(r);

				if (clean_entry(r)) {
					struct mrtentry *tmp = r;
					tmp = r;
					r = r->grpnext;
					delete_mrt(tmp);
					continue;
				}
				if (r->flags & MRTF_ASSERTED)
					timeout_assert_rpf(r);

				if (r->rpreachtimer) {
					if ((r->flags&(MRTF_WC|MRTF_RP|MRTF_IIF_REGISTER))
					    != (MRTF_WC|MRTF_RP|MRTF_IIF_REGISTER) ||
					    local_address(g->active_rp->rpaddr) == MAXVIFS)
						r->rpreachtimer = 0;
					else if (r->rpreachtimer > TIMER_INTERVAL)
						r->rpreachtimer -= TIMER_INTERVAL;
					else {
						r->rpreachtimer = TIMER_INTERVAL*18;
						send_rpreach_v1(r, TIMER_INTERVAL*36);
					}
				}

				r = r->grpnext;
			}

			if (g->leaves == EMPTY_VIF && g->mrtlink == NULL) {
				*gp = g->next;
				free(g);
				continue;
			}
			gp = &g->next;
		}
	}

	clean_srclist();

	periodic_vif_join_prune();

	/* check the candidate RP flag and if set, call periodic
	   candidate RP advertisement function */
	if (CRPFlag)
		periodic_C_RP_Adv();

	ageRPSetTimer();
}
