

/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: vif.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"

struct vifentry viflist[MAXVIFS];	/* array of all the virtual interfaces */
vifi_t numvifs;			/* Number of vifs */
vifi_t reg_vif_num;		/* dummy interface for registers */
vifbitmap_t nomrt_vif;

static void start_vif(vifi_t vifi);
static void stop_vif(vifi_t vifi);
static int add_reg_vif(void);

/* function name: start_vifs
   input: none
   output: none
   operation: starts all UP vifs
   comments: called by init_vifs
   calls start_vif
 */
void start_vifs(void)
{
	vifi_t vifi;
	struct vifentry *v;

	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		if (v->flags & VIFF_DISABLED)
			continue;
		/* Start vif if not DISABLED or DOWN */
		if (!(v->flags & VIFF_DOWN))
			start_vif(vifi);
		else
			log(LOG_INFO, 0, "%s is not up; vif #%u out of service",
			    v->name, vifi);
	}
}

/* function name: max_local_address
   input: none
   output: the address of the maximum local vif that's UP
   operation: browses thru the viflist and checks the flags and
   addresses
   comment: 
 */
__u32 max_local_address(void)
{
	vifi_t vifi;
	struct vifentry *v;
	u_long max_address = 0;

	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		if (v->flags & (VIFF_DISABLED | VIFF_DOWN))
			continue;
		if (v->prefsrc > max_address)
			max_address = v->prefsrc;
	}
	return (max_address);
}

/* function name: init_vifs
   input:none
   output: none
   operation: inits vifs from kernel and from file, starts vifs in kernel,
   inits bootstrap rp and adds reg vif
   comments: called by main [main.c]
   calls config_vifs_from_kernel, config_vifs_from_file,
   initBSR, start_vifs, add_reg_vif
   and k_add_vif.
 */
void init_vifs()
{
	struct ifreq ifr;
	register vifi_t vifi;
	register struct vifentry *v;
	int en_vif, en_phyint;
	char *c;

	numvifs = 0;
	nomrt_vif = ALL_VIF;

	/* Read the information about of the interfaces 
	 * from the kernel and configure. 
	 */
	config_vifs_from_kernel();

	/* Read the information about of the interfaces 
	 * from the configuration file and configure. 
	 */
	if ((c = alloc_mem(sizeof(struct rplist))) == NULL)
		 log(LOG_ERR, 0, "ran out of memory");

	config_vifs_from_file();

	initBSR();

	en_vif = 0;
	en_phyint = 0;

	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		/* Initialize the outgoing timeout for each vif.
		 * Currently, only use a fixed time 'OUTGOING_TIMEOUT'.
		 * Later, may add a configurable array to feed these
		 * parameters, or compute them as function of the i/f
		 * bandwidth and the overall connectivity...etc.
		 */
		oifTimeout[vifi] = OUTGOING_TIMEOUT;
		if (v->flags & VIFF_DISABLED)
			continue;
		en_vif++;
		en_phyint++;
	}

	if (en_vif < 2)
		log(LOG_ERR, 0, "can't forward: %d enabled vif", en_vif);

	if (en_phyint == 0)
		log(LOG_WARNING, 0, "No forwarding as no physical interface");

	k_init_pim();		/* Call to kernel to initiliaze structures */

	start_vifs();

	/* Add a dummy virtual interface to support Registers in the kernel. 
	 * In order for this to work, the kernel has to have been modified
	 * with the PIM patches to ip_mroute.{c,h}.
	 */
	if (debug > 2)
		log(LOG_INFO, 0, "Adding reg_vif");
	add_reg_vif();
	k_add_vif(reg_vif_num);
	strcpy(ifr.ifr_name, "pimreg");
	if (ioctl(igmp_socket, SIOGIFINDEX, (char *) &ifr) < 0) {
		log(LOG_ERR, errno, "ioctl SIOGIFINDEX for %s",
		    ifr.ifr_name);
		/* Not reached */
		return;
	}
	viflist[reg_vif_num].ifindex = ifr.ifr_ifindex;
	if (debug > 2)
		log(LOG_INFO, 0, "reg_vif is up (%d)", reg_vif_num);
}


static void start_vif(vifi_t vifi)
{
	register struct vifentry *v;

	v = &viflist[vifi];
	/* Initialy no router on any vif */
	v->flags = (v->flags | VIFF_DR | VIFF_NOMRT) & ~VIFF_DOWN;
	v->timer = 0;
	v->gq_timer = 0;
	v->query_holdtime = HelloHoldtime;
	v->neighbors = NULL;
	if (v->flags & VIFF_NOMRT)
		VIFM_SET(vifi, nomrt_vif);
	k_joinleave(igmp_socket, v, htonl(INADDR_ALLROUTERS_GROUP), 1);
	k_joinleave(igmp_socket, v, htonl(INADDR_ALL_PIM_ROUTERS), 1);
	k_add_vif(vifi);	/* Tell kernel to start this vif */
	log(LOG_INFO, 0, "%s comes up; vif #%u now in service", v->name, vifi);
	query_neighbors_v1(v);
	query_neighbors_v2(v);

	/* Query for groups to check if there are any directly connected hosts, 
	 * initially assume I am the DR.
	 */
	query_groups(v);
	autorp_vif(v->ifindex, 1);
}

static int add_reg_vif(void)
{
	struct vifentry *v;

	v = &viflist[numvifs];

	v->flags = 0;		/* initialization */

	if ((numvifs + 1) == MAXVIFS) {
		log(LOG_ERR, 0, "can't install reg_vif... too many vifs");
		return -1;
	}
	reg_vif_num = numvifs;


	/* set the REGISTER flag */
	v->flags = VIFF_REGISTER;
	strncpy(v->name, "register_vif", IFNAMSIZ);
	v->prefsrc = viflist[0].prefsrc;
	v->threshold = MINTTL;
	return 0;
}

static void neigh_destroy(struct nbrentry *n)
{
	int i;
	struct srcentry *s, **sp;
	struct mrtentry *r;

	for (i = 0; i < SRCHASH_SIZE; i++) {
		for (sp = &srchash[i]; (s = *sp) != NULL; ) {
			for (r = s->mrtlink; r != NULL; r = r->next) {
				if (r->flags & MRTF_ASSERTED) {
					if (r->upstream == n) {
						r->flags &= ~MRTF_ASSERTED;
						r->upstream = NULL;
					}
				}
			}
			if (s->upstream != n) {
				sp = &s->next;
				continue;
			}
			unicast_routing_check(s);
			sp = &s->next;
		}
	}
}

struct nbrentry * neigh_lookup(vifi_t vifi, __u32 src, int version)
{
	struct nbrentry *n;

	for (n = viflist[vifi].neighbors; n; n = n->next) {
		if (n->address == src) {
			if (n->pimversion&version)
				return n;
			return NULL;
		}
	}
	return NULL;
}

static void stop_vif(vifi_t vifi)
{
	struct vifentry *v;
	register struct nbrentry *n;

	/* Here make sure that the kernel viftable is 
	 * consistent with the daemon table
	 */
	v = &viflist[vifi];
	autorp_vif(v->ifindex, 0);
	k_joinleave(igmp_socket, v, htonl(INADDR_ALLROUTERS_GROUP), 0);
	k_joinleave(igmp_socket, v, htonl(INADDR_ALL_PIM_ROUTERS), 0);
	k_del_vif(vifi);

	v->flags = (v->flags & ~VIFF_DR & ~VIFF_NOMRT) | VIFF_DOWN;
	v->timer = 0;
	v->gq_timer = 0;
	while ((n = v->neighbors) != NULL) {
		v->neighbors = n->next;
		neigh_destroy(n);
	}
	log(LOG_INFO, 0, "%s goes down; vif #%u out of service", v->name, vifi);
}

#ifdef NO_ANK_FIX
/* Mama! What is it? --ANK */

void set_vif_default(char *s)
{
	vifi_t vifi;
	register struct vifentry *v;

	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		if (v->flags & (VIFF_DOWN | VIFF_DISABLED))
			continue;
		if (debug > 4)
			log(LOG_DEBUG, 0, "set_vif_default: Looking for %s in %s",
			    s,
			    v->name);
		if (!strcmp(s, v->name)) {
			v->flags |= VIFF_DEFAULT;
			break;
		}
	}
}

#endif

vifi_t vifi_by_index(int ifindex)
{
	vifi_t vifi;
	register struct vifentry *v;

	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		if (v->ifindex != ifindex)
			continue;
		if (v->flags & (VIFF_DOWN | VIFF_DISABLED))
			return MAXVIFS;
		return vifi;
	}
	return MAXVIFS;
}

void query_groups(struct vifentry *v)
{
	v->gq_timer = 0;	/* Send a query to the groups if you are a DR */
	if (v->flags & VIFF_DR) {
		send_igmp(igmp_send_buf, v->prefsrc, allhosts_group,
			  IGMP_HOST_MEMBERSHIP_QUERY, 0, 0, v->ifindex, 1, 1);
	}
}


void accept_group_report(__u32 src, __u32 dst, __u32 group)
{
	vifi_t vifi;
	register struct vifentry *v;

	/* IGMPv1 specs propose to ignore report if dst != group;
	   it was silly restriction and seems IGMPv2 deprecated it.
	 */
	if (dst != group)
		tracef(TRF_UNUSUAL, "UNUSUAL accept_group_report: %s is reported from %s to %s",
		       inet_fmt(group, s1), inet_fmt(src, s2), inet_fmt(dst, s3));

	vifi = vifi_by_index(igmp_ifindex);
	v = &viflist[vifi];

	if (vifi == MAXVIFS) {
		tracef(TRF_UNUSUAL, "accept_group_report: ignoring group %s report from host %s",
		       inet_fmt(group, s1), inet_fmt(src, s2));
		return;
	}

	tracef(TRF_IGMP, "RX IGMP: accepting group report from host %s for group %s",
	       inet_fmt(src, s1), inet_fmt(group, s2));

	/* If I am a DR add the leaf */
	if (v->flags & VIFF_DR)
		add_leaf(vifi, dst);
	/* Might need to add a check if I am the forwarder */
}

static struct nbrentry *neigh_create(vifi_t vifi, __u32 src, int mode, int vers)
{
	struct nbrentry *n;
	struct vifentry *v = &viflist[vifi];

	n = (struct nbrentry *) alloc_mem(sizeof(struct nbrentry));
	n->address = src;
	n->mode = mode;
	n->pimversion = vers;
	n->next = v->neighbors;
	n->uptime = wall_time;
	v->neighbors = n;
	v->neighnum++;
	n->vif = vifi;
	return n;
}

static void event_neigh_up(vifi_t vifi, struct nbrentry *n)
{
	struct vifentry *v = &viflist[vifi];
	struct srcentry *s;

	if (ntohl(v->prefsrc) < ntohl(n->address) && (v->flags & VIFF_DR)) {
		v->flags &= ~VIFF_DR;
		if (debug)
			log(LOG_DEBUG, 0, "PIM: New DR on vif %d is %s",
			    vifi,
			    inet_fmt(n->address, s1));

		/* If a change in DR status occurs, may want to check
		 * if REGISTER entries exist then should delete.
		 */
		forall_sources(s) {
			struct mrtentry *r;

			if (s->incoming != vifi)
				continue;

			for (r = s->mrtlink; r != NULL; r = r->srcnext) {
				if ((r->flags & (MRTF_REGISTER|MRTF_REG_SUPP))) {
					if (debug)
						log(LOG_DEBUG, 0, "Deleting REGISTER information for (%s,%s)",
						    inet_fmt(s->source, s1),
						    inet_fmt(r->group->group, s2));
					r->flags &= ~(MRTF_REGISTER|MRTF_REG_SUPP);
					dirty_bits |= 2;
					k_chg_mrtf(r);
				}
			}
		} endforall_sources(s);
	};
	/* FIX all the other cases are not treated */

	if (v->flags & VIFF_NOMRT) {
		v->flags &= ~VIFF_NOMRT;
		VIFM_CLR(vifi, nomrt_vif);
	}
}


/* function name: parse_query
   input: pkt pointer, holdtime address, and bootstrap address
   output: int (TURE if parse successful)
   operationg: parses the pkt given as PIM query, and assigns the holdtime
   and bootstrap accordignly
   comments: called by accept_query
 */
int parse_query(char *p, __u16 *holdtime)
{
	__u16 *ptr = (void*)p;
	if (ntohs(ptr[0]) != 1 || ntohs(ptr[1]) != 2)
		return FALSE;
	*holdtime = ntohs(ptr[2]);
	return (TRUE);
}

/* 
   function name: accept_query
   input: src, & dst addresses, pointer to the pkt, and the datalength
   output: none
   operation: processes the PIM Query messages
   comments: called by accept_pim
 */
void accept_query_v2(__u32 src, __u32 dst, char *p, int datalen)
{
	vifi_t vifi;
	struct vifentry *v;
	struct nbrentry *n;
	__u16 holdtime;
	char *pkt;
	u_int length;

	if ((vifi = vifi_by_index(pim_ifindex)) == MAXVIFS) {
		log(LOG_INFO, 0, "ignoring neighbor query from router %s on if%d",
		    inet_fmt(src, s1), pim_ifindex);
		return;
	}
	v = &viflist[vifi];
	parse_query(p, &holdtime);

	n = neigh_lookup(vifi, src, 3);
	if (n) {
		if (n->pimversion&2) {
			n->timer2 = 0;
			n->holdtime2 = holdtime;
			return;
		}
		n->pimversion |= 2;
		if (n->pimversion&1)
			v->v1neighnum--;
	} else {
		if (dst != allpimrouters_group) {
			log(LOG_INFO, 0, "ignoring neighbor query from router %s on %s. Dst=%s",
			    inet_fmt(src, s1), v->name, inet_fmt(dst, s2));
			return;
		}
		n = neigh_create(vifi, src, 2, 2);
		n->holdtime2 = holdtime;
		n->timer2 = 0;
		v->v2neighnum++;
		event_neigh_up(vifi, n);
	}

	/* Since new neighbour has come up, let it know your existence */
	query_neighbors_v2(v);

	/* 
	 * If I am the DR on that interface, and this is a new nbr, then
	 * should trigger an RP-Set message to it.
	 */
	if (v->flags & VIFF_DR) {
		pkt = pim_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN;
		if ((length = construct_c_rp_set(pkt)) != 0)
			send_c_rp_set(pim_send_buf, length, v->prefsrc, src);
	}

	dirty_bits |= 1;
}

/* function name: accept_query
   input: src, & dst addresses, pointer to the pkt, and the datalength
   output: none
   operation: processes the PIM Query messages
   comments: called by accept_pim
 */
void accept_query_v1(__u32 src, __u32 dst, char *p, int datalen)
{
	vifi_t vifi;
	struct vifentry *v;
	struct nbrentry *n;
	struct pimquery *query;
	u_short holdtime;

	if ((vifi = vifi_by_index(igmp_ifindex)) == MAXVIFS) {
		log(LOG_INFO, 0, "ignoring neighbor query from router %s on if%d",
		    inet_fmt(src, s1), igmp_ifindex);
		return;
	}
	query = (struct pimquery *) p;
	holdtime = ntohs(query->holdtime);

	tracef(TRF_PIM1, "RX PIM: Query[Mode:%d | Holdtime=%u", query->mode>>4, holdtime);

	v = &viflist[vifi];
	n = neigh_lookup(vifi, src, 3);

	if (n) {
		if (n->pimversion&1) {
			n->timer = 0;
			n->holdtime = holdtime;
			return;
		} else {
			n->pimversion |= 1;
			if (n->pimversion&2)
				v->v2neighnum--;
		}
	} else {
		if (dst != allrouters_group) {
			log(LOG_INFO, 0, "ignoring neighbor query from router %s on %s. Dst=%s",
			    inet_fmt(src, s1), v->name, inet_fmt(dst, s2));
			return;
		}
		n = neigh_create(vifi, src, query->mode>>4, 1);
		n->holdtime = holdtime;
		n->timer = 0;
		v->v1neighnum++;
		event_neigh_up(vifi, n);
	}

	dirty_bits |= 1;

	/* Since new neighbour has come up, let it know your existence */
	query_neighbors_v1(v);
}

void age_vifs()
{
	struct ifreq ifr;
	vifi_t vifi;
	register struct vifentry *v;
	register struct nbrentry *n, **np;
	u_char flags;

	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		int we_are_dr = 1;

		if (v->flags & VIFF_DISABLED)
			continue;

		strncpy(ifr.ifr_name, v->name, IFNAMSIZ);
		/* get the interface flags */
		if (ioctl(igmp_socket, SIOCGIFFLAGS, (char *) &ifr) < 0)
			log(LOG_ERR, errno, "age_vifs: ioctl SIOCGIFFLAGS for %s", ifr.ifr_name);

		if (v->flags & VIFF_DOWN) {
			if (ifr.ifr_flags & IFF_UP) {
				dirty_bits |= 1;
				start_vif(vifi);
			}
		} else if (!(ifr.ifr_flags & IFF_UP)) {
			dirty_bits |= 1;
			stop_vif(vifi);
		}

		if (v->flags & VIFF_DOWN)
			continue;

		for (np = &v->neighbors; (n=*np) != NULL; ) {
			if ((n->pimversion&1) && (n->timer += TIMER_INTERVAL) > n->holdtime) {
				n->pimversion &= ~1;
				if (n->pimversion&2)
					v->v2neighnum++;
				dirty_bits |= 1;
			}
			if ((n->pimversion&2) && (n->timer2 += TIMER_INTERVAL) > n->holdtime2) {
				n->pimversion &= ~2;
				if (n->pimversion&1)
					v->v1neighnum++;
				dirty_bits |= 1;
			}
			if (n->pimversion) {
				if (ntohl(v->prefsrc) < ntohl(n->address))
					we_are_dr  = 0;
				np = &n->next;
				continue;
			}
			log(LOG_DEBUG, 0,
			    "age_vifs: Del nbr %s,  n->timer=%d > n->holdtime %d",
			    inet_fmt(n->address, s1),
			    n->timer,
			    n->holdtime);
			dirty_bits |= 1;
			v->neighnum--;
			*np = n->next;
			neigh_destroy(n);
		}

		flags = v->flags;
		if (we_are_dr)
			v->flags |= VIFF_DR;
		if ((n = v->neighbors) == NULL) {
			v->flags |= (VIFF_NOMRT | VIFF_DR);
			VIFM_SET(vifi, nomrt_vif);
		}

		if ((v->timer += TIMER_INTERVAL) > HelloPeriod) {
			query_neighbors_v1(v);
			query_neighbors_v2(v);
		}
		if ((v->gq_timer += TIMER_INTERVAL) > GROUP_QUERY_INTERVAL)
			query_groups(v);
	}
}

int inet_mask_len(__u32 mask)
{
	int i = 0;
	mask = ntohl(mask);
	while (mask & 0xFFFFFFFF) {
		mask <<= 1;
		i++;
	}
	return i;
}

void dump_vifs(FILE *fp)
{
	vifi_t vifi;
	register struct vifentry *v;
	struct nbrentry *n;
	int exist_neigh = 0;

	fprintf(fp, "\nVirtual Interface Table\n");
	fprintf(fp, "Flags: Di-Disabled, Dn-Down, Dr-Designated Router, Nn-No Neighbor Df-Default Vif\n\n");
	fprintf(fp, "Dev    Vif Local-Address                   Ttl  Flags\n");
	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		struct vifaddr *vifa;
		fprintf(fp, "%-6s %-2u  %-15s %-15s %-3u ", v->name, vifi,
			inet_fmt(v->prefsrc, s1), "", v->threshold);
		if (v->flags & VIFF_DISABLED)
			fprintf(fp, " Di");
		if (v->flags & VIFF_DOWN)
			fprintf(fp, " Dn");
		if (v->flags & VIFF_DR)
			fprintf(fp, " Dr");
		if (v->flags & VIFF_NOMRT)
			fprintf(fp, " Nn");
#ifdef NO_ANK_FIX
		if (v->flags & VIFF_DEFAULT)
			fprintf(fp, " Df");
#endif
		fprintf(fp, "\n");
		for (vifa = v->ifa; vifa; vifa = vifa->next) {
			fprintf(fp, "%-6s %-2s  %-15s %s/%d\n", "", "",
				inet_fmt(vifa->local, s1),
				inet_fmt(vifa->prefix, s1),
				inet_mask_len(vifa->mask));
		}
		if (v->neighbors != NULL)
			exist_neigh++;
	}
	fprintf(fp, "\n");
	if (exist_neigh) {
		fprintf(fp, "\nPIM Neighbor Table\n");
		fprintf(fp, "Neighbor Address  Vif Uptime    Expires   Vers  Mode\n\n");

		for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
			for (n = v->neighbors; n != NULL; n = n->next) {
				fprintf(fp, "%-15s   %-3d %-9s %-9s ",
					inet_fmt(n->address, s1),
					n->vif,
					inet_time(n->uptime, s2),
					inet_time(n->holdtime + wall_time - n->timer, s3));
				switch (n->pimversion) {
				case 0:	
					fprintf(fp, "%-5s", "None");
					break;
				case 1:	
					fprintf(fp, "%-5s", "PIMv1");
					break;
				case 2:	
					fprintf(fp, "%-5s", "PIMv2");
					break;
				case 3:	
					fprintf(fp, "%-5s", "Dual");
					break;
				default:	
					fprintf(fp, "%-5u", n->pimversion);
					break;
				}

				switch (n->mode) {
				case 2:	
					fprintf(fp, " Sparse-Dense");
					break;
				case 1:	
					fprintf(fp, " Sparse");
					break;
				case 0:	
					fprintf(fp, " Dense");
					break;
				default:	
					fprintf(fp, " Unknown(%d)", n->mode);
					break;
				}
				fprintf(fp, "\n");
			}
		}
		fprintf(fp, "\n");
	}
}

